/*
 ķ
   Program..: _DCXBROW.PRG                                 
   Author...:  Roger J. Donnay                             
   Notice...: (c) DONNAY Software Designs 1987-2002        
   Date.....: August 8, 2018                               
   Notes....: eXPress++ Browse Classes                     
 Ľ
*/

#INCLUDE "appevent.ch"
#INCLUDE "dcdialog.ch"
#INCLUDE "dll.ch"
#INCLUDE "dcbitmap.CH"
#INCLUDE "dcads.CH"
#INCLUDE "dmlb.CH"
#INCLUDE "dcprint.CH"
#INCLUDE "dccursor.CH"

#DEFINE QUERY_SAY          1
#DEFINE QUERY_FIELD        2
#DEFINE QUERY_PICTURE      3
#DEFINE QUERY_INDEX        4
#DEFINE QUERY_ROW          5
#DEFINE QUERY_COL          6
#DEFINE QUERY_GET          7
#DEFINE QUERY_VAR          8

// Drawing modes used with XBPs
#ifndef XBP_DRAW_NORMAL
#define XBP_DRAW_NORMAL                0
#define XBP_DRAW_OWNER                 1
#define XBP_DRAW_OWNERADVANCED         2
#endif

#define FONTBOLD      "Bold"    // change to "Fett" for German versions
#define FONTITALIC    "Italic"  // change to "Kursiv" for German versions
#define CRLF          Chr(13)+Chr(10)

CLASS DC_XbpBrowse FROM DC_XbpBrowseSuper, DC_XbaseParts

EXPORTED:
  VAR oHasFocus
  VAR dataLink
  VAR color
  VAR varBlock
  VAR captionBlock
  VAR picture
  VAR optionArray
  VAR arrayElement
  VAR dataSource
  VAR dataSourceType                      // J. Duijf 26-11-2002
  VAR thumbRecords
  VAR headerLines
  VAR headerDelim
  VAR footerLines
  VAR footerDelim
  VAR headerPres
  VAR footerPres
  VAR browsePres                          // J. Duijf 10-03-2011
  VAR fit
  VAR fitMax
  VAR optimize
  VAR outputBlock
  VAR inputBlock
  VAR mark
  VAR markColorBlock
  VAR rowColorBlock
  VAR isScope
  VAR font
  VAR pointerBlock
  VAR autoLock
  VAR freezeLeft
  VAR freezeRight
  VAR columns
  VAR sortColorSelectedFG
  VAR sortColorSelectedBG
  VAR sortColorUnSelectedFG
  VAR sortColorUnSelectedBG
  VAR sortColorNonselectableFG
  VAR sortColorNonselectableBG
  VAR sortFontSelected
  VAR sortFontUnSelected
  VAR sortFontNonSelectable
  VAR sortButtonObject
  VAR sortButtonConfig
  VAR descending
  VAR descendToggle
  VAR sortImageAscend
  VAR sortImageDescend
  VAR sortColumn
  VAR refreshInterval
  VAR refreshWhenBlock
  VAR refreshBlock
  VAR timer
  VAR rightButtonSelect
  VAR cellEditMode
  VAR cellEditor
  VAR saveRecord
  VAR hasHeader
  VAR hasFooter
  VAR navBlocks
  VAR dataEnabled
  VAR rightButtonDown
  VAR defaultColumnsGetList
  VAR configMessageBlock
  VAR filter
  VAR noRestore
  VAR restoreConfig
  VAR restoreProc
  VAR skipUnique
  VAR lBrow
  VAR nBrow
  VAR resizeColumns
  VAR tagEnable
  VAR tagElement
  VAR tagColorFG
  VAR tagColorBG
  VAR tagMode
  VAR lastTagRec
  VAR lastTagPos
  VAR config
  VAR scrollBarTipBlock
  VAR isItemMarkedCodeBlock
  VAR ownerDrawBlock
  VAR isLeftMouseButton
  VAR isTrapErrors
  VAR evenRow
  VAR CellEditKeyBlock                                                        // J. Duijf 30-03-2013
  VAR isCellEditing
  VAR scrollBarHeight
  VAR navigateBlock
  VAR recordObject
  VAR browseColor
  VAR sqlSessionDict
  VAR sqlSessionFree
  VAR sqlTableName
  VAR sqlStatementOriginal
  VAR rowColor
  VAR columnTagColor
  VAR columnCache
  VAR isColumnCacheAutomatic
  VAR displayedRows
  VAR saveScope0
  VAR saveScope1

  METHOD Init, create, Navigate, FieldBlock, CellEditSetup, ;
         Save, Restore, MoveColumn, Refresh, WhenHideEval, Destroy, ;
         handleEvent, setTimerRefresh, right, itemRbDown, CellEdit, ;
         CellEditMode, RefreshAll, SortClear, HFLbDown, DataDisplay, ;
         Optimize, RestoreDefaults, FitColumns, ForceStable, ;
         RefreshView, RowColor, ResizeColumns, Tag, Resize, SetSize, ;
         ChooseCachedColumn, NextCachedColumn, CachedColumnKey, ;
         CachedColumnPan, StabilizeColumns, DisplayedRows, OrderBy, ;
         FindGoodIndex, IndexInfo

  METHOD XbpBrowsehandleEvent IS handleEvent IN DC_XbpBrowseSuper
  METHOD ZebraColor                                                           // J. Duijf 21-04-2010
  METHOD CustomDrawCell

ENDCLASS

* ---------------------

METHOD DC_XbpBrowse:init( oParent, oOwner, aPos, aSize, aPres, lVisible, oGetList )

LOCAL xOptions, bBlock, aArray, bColorBlock, aColor, nPointer, ;
      aGetList, aGetListItem, xData, aSort, xFont, lThumbLock, ;
      nFound, aConfig, cID, nColumns
LOCAL aEvenColor, aOddColor, bTmpColorBlock
LOCAL aTagedColor,aTagedHiliteColor                                           // J. Duijf 02-08-2018

::DC_XbaseParts:init( oGetList,,self )

aGetList := ::getList:getListArray
aGetListItem := ::getList:getListArray[::getListPointer]

IF Valtype(aGetListItem[oGETLIST_CONFIG]) == 'O'
  ::config := aGetListItem[oGETLIST_CONFIG]
ELSE
  ::config := DC_XbpBrowseDefault()
ENDIF

//
// J. Duijf 11-03-2011
// First get the default presentation parameters (As in DC_BrowPres())
// Then merge the optional presentation parameters with the defaults.
//
::browsePres      := Aclone(::config:presParam)                               // J. Duijf 10-03-2011
DC_MergePresentation(::browsePres,aPres)

::browsePres      := ApplyScaleFactor(::browsePres,oGetList)
//
// J. Duijf, 12-04-2011 Moved here to update presentation parameters
//
xFont             := aGetListItem[cGETLIST_FONT]
IF Empty(xFont)
  DEFAULT xFont   := ::config:font
ENDIF

IF !Empty(xFont)
  IF Valtype(xFont) == 'B'
    xFont := Eval(xFont)
  ENDIF
   DC_MergePresentation(::browsePres,{{XBP_PP_COL_DA_COMPOUNDNAME,xFont}})
   DC_MergePresentation(::browsePres,{{XBP_PP_COL_HA_COMPOUNDNAME,xFont}})
   DC_MergePresentation(::browsePres,{{XBP_PP_COL_FA_COMPOUNDNAME,xFont}})
//  ::setFontCompoundName(Alltrim(xFont))
ENDIF


bBlock := DC_GetBlock(aGetListItem[bGETLIST_REFVAR])
IF Valtype(bBlock) = 'B'
  xData := Eval( bBlock )
  IF !DC_ReadDataCheck( xData, 'ACONI', aGetList, ::getListPointer , .t. )
    RETURN nil
  ENDIF
ELSEIF Valtype(aGetListItem[cGETLIST_VARNAME]) = 'C'
  xData := aGetListItem[cGETLIST_VARNAME]
  IF !DC_ReadDataCheck( xData, 'C', aGetList, ::getListPointer, .t. )
    RETURN nil
  ENDIF
ENDIF

IF Valtype(xData) # 'A' .AND. Empty(xData)
  xData := Alias()
  IF Empty(xData)
    DC_WinAlert('DC_ReadGui() error! Browse has no data source')
    RETURN nil
  ENDIF
ENDIF

xOptions := aGetListItem[xGETLIST_OPTIONS]
ASize(xOptions,61)
aArray := xOptions

::XbpBrowse:init( oParent, oOwner, aPos, aSize, ::browsePres, lVisible )

IF Valtype(aArray[51]) == 'N'
  ::nScrollHeight := aArray[51]
  ::oHScroll:setSize({::oHScroll:currentSize()[1],::nScrollHeight})
  ::oVScroll:setSize({::nScrollHeight,::oVScroll:currentSize()[2]})
ENDIF

::scrollBarHeight := ::nScrollHeight // protected var in XbpBrowse() class.

IF IsMemberVar(self,'useVisualStyle')
  ::useVisualStyle := ::config:useVisualStyleBrowse
ENDIF

::columnTagColor := { GRA_CLR_BLACK, GraMakeRGBColor({230,240,250})}
::rowColor := {nil,nil}
::tabstop := .f.
::dataEnabled := .t.
::noRestore := .f.
::skipUnique := .f.
::nBrow := 1
::lBrow := .f.
::isItemMarkedCodeBlock := .f.
::isTrapErrors := DC_BrowseTrapErrors()
::evenRow := .f.
::isLeftMouseButton := .f.
::configMessageBlock := DC_BrowseConfigMessageBlock()
::isCellEditing := .f.
IF Valtype(DC_BrowseColor()) == 'A'
  ::browseColor := AClone(DC_BrowseColor())
ENDIF
IF Valtype(aGetListItem[xGETLIST_OPTIONS7]) = 'N'
  ::cursorMode := aGetListItem[xGETLIST_OPTIONS7]
ELSE
  ::cursorMode := XBPBRW_CURSOR_CELL
ENDIF
::sqlStatementOriginal := ''
::displayedRows := -1

IF Valtype(xData) = 'A'
  aGetListItem[nGETLIST_SUBTYPE] := BROWSE_ARRAY
ELSEIF Valtype(xData) = 'C'
 // ::displayedRows := (xData)->(RecCount())
  aGetListItem[nGETLIST_SUBTYPE] := BROWSE_DATABASE
  Sleep(5)
  ::sqlTableName := (xData)->(Upper(dbInfo( DBO_FILENAME )))
  ::sqlTableName := Strtran(DC_Path(::sqlTableName,.t.),'.DBF','')
  IF ::sqlTableName = 'SELECT '
    ::sqlStatementOriginal := ::sqlTableName
    ::sqlTableName := Alias()
  ELSEIF ::sqlTableName = '<CURSOR>' .OR. '</CURSOR>' $ ::sqlTablename
    ::sqlStatementOriginal := DC_AdsGetSqlStatement()
    ::sqlTableName := Alias()
  ELSEIF Left(::sqlTableName,1) == '_'
    ::sqlStatementOriginal := 'SELECT * FROM ' + ::sqlTableName
  ENDIF
ELSEIF Valtype(xData) = 'O'                                                   // J. Duijf 27-11-2002
   IF xData:isDerivedFrom("SQLStatement") .OR. xData:isDerivedFrom("SQLDataSet")  // Sql-express
     aGetListItem[nGETLIST_SUBTYPE] := BROWSE_SQLEXPRESS
   ELSEIF xData:isDerivedFrom("dsACEServer")    // ACE Server
     aGetListItem[nGETLIST_SUBTYPE] := BROWSE_ACESERVER
   ELSEIF xData:isDerivedFrom("SpSql")    // SpSql silex
     aGetListItem[nGETLIST_SUBTYPE] := BROWSE_SPSQL //
   ELSEIF xData:isDerivedFrom("MyResult")    // MYSQL
     aGetListItem[nGETLIST_SUBTYPE] := BROWSE_MYSQL
   ELSE                                                             // Database object
     aGetListItem[nGETLIST_SUBTYPE] := BROWSE_DATAOBJECT
   ENDIF
ELSEIF Valtype(xData) = 'N'
  aGetListItem[nGETLIST_SUBTYPE] := BROWSE_ADSHANDLE
ENDIF

::dataSource      := xData
::dataSourceType  := aGetListItem[nGETLIST_SUBTYPE]                           // J. Duijf 27-11-2002

IF Valtype(aGetListItem[aGETLIST_COLOR]) = 'A' .AND. ;
           Valtype(aGetListItem[aGETLIST_COLOR,1]) = 'B'
  bColorBlock := aGetListItem[aGETLIST_COLOR,1]

ELSEIF Valtype( aColor ) = 'A' .AND. Valtype(aGetListItem[aGETLIST_COLOR]) # 'B'
  bColorBlock := &('{|| {' + DC_XtoC(aColor[1]) + ',' + DC_XtoC(aColor[2]) + '}}' )

ELSEIF Valtype(aGetListItem[aGETLIST_COLOR]) = 'B'
  bColorBlock := aGetListItem[aGETLIST_COLOR]

ELSE

  bColorBlock := nil

ENDIF

IF Empty(aColor)
  aColor := { nil,15 }
ENDIF

DEFAULT aArray[9] := .f.,     ; //
        aArray[10] := nil,    ; // Thumb records
        aArray[11] := 1,      ; // Header Lines
        aArray[12] := ';',    ; // Header Delimiter
        aArray[13] := 1,      ; // Footer Lines
        aArray[14] := ';',    ; // Footer Delimiter
        aArray[15] := nil,    ; // Header Presentation Array
        aArray[16] := nil,    ; // Footer Presentation Array
        aArray[17] := .f.,    ; // No vertical scroll
        aArray[18] := .f.,    ; // No horizontal scroll
        aArray[19] := .f.,    ; // No Soft Track
        aArray[20] := .f.,    ; // Fit
        aArray[21] := .f.,    ; // Optimize
        aArray[24] := .f.,    ; // No Size Cols
        aArray[31] := .f.,    ; // Descending
        aArray[32] := 0,      ; // refresh time
        aArray[33] := .f.,    ; // Right Button Select
        aArray[34] := .f.,    ; // Don't disable descending toggle
        aArray[35] := nil,    ; // Code block to evaluate when Right-Button clicked in Data Area
        aArray[36] := nil,    ; // Array Filter
        aArray[37] := .F.,    ; // No Restore if .T.
        aArray[40] := .f.,    ; // Resize Columns
        aArray[41] := .f.,    ; // Tag Mode
        aArray[42] := 0,      ; // Tag Element
        aArray[43] := nil,    ; // Tag Foreground Color
        aArray[44] := GRA_CLR_YELLOW, ; // Tag Background Color
        aArray[45] := 0,      ; // Tag Background Color
        aArray[46] := .f.,    ; // Use Visual Style
        aArray[47] := nil,    ; // Refresh When
        aArray[48] := nil,    ; // Refresh Block
        aArray[49] := nil,    ; // OwnerDraw Block
        aArray[50] := 0,      ; // Max width for auto fit
        aArray[60] := 200,    ; // Column cache (200 columns displayed)
        aArray[61] := .f.     ; // Auto Add Columns

::tagEnable    := aArray[41]
::tagElement   := aArray[42]
::tagColorFG   := aArray[43]
::tagColorBG   := aArray[44]
::tagMode      := aArray[45]
::lastTagRec   := 0
::lastTagPos   := 0
IF IsMemberVar(self,'USEVISUALSTYLE')
  ::useVisualStyle := aArray[46]
ENDIF
::CellEditMode := .f.
::CellEditSetup()
::resizeColumns := aArray[40]
::isColumnCacheAutomatic := aArray[61]

aSort := DC_BrowseSort()
::sortColorSelectedFG := aSort[1]
::sortColorSelectedBG := aSort[2]
::sortColorNonselectableFG := aSort[3]
::sortColorNonselectableBG := aSort[4]
::sortImageAscend := aSort[5]
::sortImageDescend := aSort[6]
::sortColorUnSelectedFG := aSort[7]
::sortColorUnSelectedBG := aSort[8]
::sortFontSelected := aSort[9]
::sortFontUnSelected := aSort[10]
::sortFontNonSelectable := aSort[11]
::sortButtonConfig := aSort[12]

IF !Empty(aArray[25])
  ::sortColorSelectedFG := aArray[25]
ENDIF
IF !Empty(aArray[26])
  ::sortColorSelectedBG := aArray[26]
ENDIF
IF !Empty(aArray[27])
  ::sortColorUnselectedFG := aArray[27]
ENDIF
IF !Empty(aArray[28])
  ::sortColorUnselectedBG := aArray[28]
ENDIF
IF !Empty(aArray[29])
  ::sortImageAscend := aArray[29]
ENDIF
IF !Empty(aArray[30])
  ::sortImageDescend := aArray[30]
ENDIF
IF !Empty(aArray[54])
  ::sortFontSelected := aArray[54]
ENDIF
IF !Empty(aArray[55])
  ::sortFontUnSelected := aArray[55]
ENDIF
IF !Empty(aArray[56])
  ::sortFontNonSelectable := aArray[56]
ENDIF
IF !Empty(aArray[57])
  ::sortButtonConfig := aArray[57]
ENDIF

IF !Empty(aArray[38])
  ::sortColorNonselectableFG := aArray[38]
ELSEIF Empty(::sortColorNonselectableFG)
   //
   // J. Duijf, 10-03-2011
   // Check for defaults first in presentation parameters
   //  ::sortColorNonselectableFG := GRA_CLR_WHITE
   ::sortColorNonselectableFG := DC_GetPresentation(::browsePres,XBP_PP_COL_HA_FGCLR,GRA_CLR_WHITE)
ENDIF
IF !Empty(aArray[39])
  ::sortColorNonselectableBG := aArray[39]
ELSEIF Empty(::sortColorNonselectableBG)
   //
   // J. Duijf, 10-03-2011
   // Check for defaults first in presentation parameters
   // ::sortColorNonselectableBG := GRA_CLR_DARKGRAY
   ::sortColorNonselectableBG := DC_GetPresentation(::browsePres,XBP_PP_COL_HA_BGCLR,GRA_CLR_DARKGRAY)
ENDIF
IF !Empty(aArray[37])
  ::noRestore := aArray[37]
ENDIF
DEFAULT ::sortColorUnselectedFG := ::sortColorNonselectableFG
DEFAULT ::sortColorUnselectedBG := ::sortColorNonselectableBG
//
// J. Duijf 10-03-2011
// If there is NO colour block. Test presentation parameters for
// default zebra striping
//

if ValType(bColorBlock) <> "B"

   if DC_GetPresentation(::browsePres,XBP_PP_COL_ZEBRA_ENABLED,0) > 0          // Zebra enabled
      //
      // If zebra and no zebra colors are defined, use default Black/White + Black/Yellow
      // And Red/Yellow + White/Darkgreen for marked rows.
      // It's better to add these colors to DC_BrowPres() in _dcpres.prg
      //
      //
      aEvenColor           := nil
      aOddColor            := nil
      aTagedColor          := nil                                             // J. Duijf 02-08-2018
      aTagedHiliteColor    := nil                                             // J. Duijf 02-08-2018

      if Ascan(::browsePres,{|a|a[1] == XBP_PP_COL_ZEBRA_ODDROW_FGCLR}) > 0
         aEvenColor        := {Dc_GetPresentation(::browsePres,XBP_PP_COL_ZEBRA_ODDROW_FGCLR,GRA_CLR_WHITE),;
                               Dc_GetPresentation(::browsePres,XBP_PP_COL_ZEBRA_ODDROW_BGCLR,GRA_CLR_YELLOW)}
      endif
      if Ascan(::browsePres,{|a|a[1] == XBP_PP_COL_ZEBRA_EVENROW_FGCLR}) > 0
         aOddColor         := {Dc_GetPresentation(::browsePres,XBP_PP_COL_ZEBRA_EVENROW_FGCLR,GRA_CLR_BLACK),;
                               Dc_GetPresentation(::browsePres,XBP_PP_COL_ZEBRA_EVENROW_BGCLR,GRA_CLR_WHITE)}
      endif

      if Ascan(::browsePres,{|a|a[1] == XBP_PP_COL_ZEBRA_TAGGED_FGCLR}) > 0                                                // J. Duijf 02-08-2018
         aTagedColor       := {Dc_GetPresentation(::browsePres,XBP_PP_COL_ZEBRA_TAGGED_FGCLR,GRA_CLR_RED),;
                               Dc_GetPresentation(::browsePres,XBP_PP_COL_ZEBRA_TAGGED_BGCLR,GRA_CLR_YELLOW)}
      endif
      if Ascan(::browsePres,{|a|a[1] == XBP_PP_COL_ZEBRA_TAGGEDHILITE_BGCLR}) > 0                                          // J. Duijf 02-08-2018
         aTagedHiliteColor := {Dc_GetPresentation(::browsePres,XBP_PP_COL_ZEBRA_TAGGEDHILITE_FGCLR,GRA_CLR_WHITE),;
                               Dc_GetPresentation(::browsePres,XBP_PP_COL_ZEBRA_TAGGEDHILITE_BGCLR,GRA_CLR_DARKGREEN)}
      endif
      bColorBlock          := {||::ZebraColor(aEvenColor,aOddColor,aTagedColor,aTagedHiliteColor)}                         // J. Duijf 02-08-2018
//      bColorBlock    := {||::ZebraColor(aEvenColor,aOddColor)}
   endif
endif

/*
if ValType(bColorBlock) = "B"
   //
   // J. Duijf 11-03-2011
   // Workaround for PDR 6365
   // The colorBlock should return nil or a 4 element array to prevent an incorrect
   // Hilite forground color.
   // {CellFGColor,CellBGColor,CellHiliteFGColor,CellHiliteBGColor}
   // The highlite colors come from the presentation parameters.
   // If undefined, the default BG/FG hilite colors are BLUE/WHITE
   //
   bTmpColorBlock    := bColorBlock
   bColorBlock       := {|a|a := Eval(bTmpColorBlock,self),;
                              iif(ValType(a) = "A" .and. Len(a) = 2,;
                                 (Aadd(a,Dc_GetPresentation(::browsePres,XBP_PP_COL_DA_HILITE_FGCLR,GRA_CLR_WHITE)), ;
                                  Aadd(a,Dc_GetPresentation(::browsePres,XBP_PP_COL_DA_HILITE_BGCLR,GRA_CLR_BLUE))),nil),a}
endif
*/

::sortColumn := 0
::descending := aArray[31]
::descendToggle := !aArray[34]
::oHasFocus := NIL
::dataLink := xData
::outputBlock := DC_GetBlock(aGetListItem[bGETLIST_VAR])
::inputBlock := DC_GetBlock(aGetListItem[bGETLIST_ACTION])
::whenBlock := DC_GetBlock(aGetListItem[bGETLIST_WHEN])
::hideBlock := DC_GetBlock(aGetListItem[bGETLIST_HIDE])
::protectBlock := DC_GetBlock(aGetListItem[bGETLIST_PROTECT])
::mark := aArray[1]
::markColorBlock := aArray[3]
// ::rowColorBlock := _RowColor( self, aArray, aColor, xData, bColorBlock )         // J. Duijf 02-08-2018
::rowColorBlock    := _RowColor( self, aArray, aColor, ::dataSource, bColorBlock )  // J. Duijf 02-08-2018
::freezeLeft := aArray[7]
::freezeRight := aArray[8]
::isScope := aArray[9]
::font := aGetListItem[xGETLIST_OPTIONS2,4]
::pointerBlock := DC_GetBlock(aGetListItem[xGETLIST_OPTIONS6])
::thumbRecords := aArray[10]
::autoLock := !(aGetListItem[xGETLIST_OPTIONS2,6] .OR. Valtype(xData)#'C')
::headerLines := aArray[11]
::headerDelim := aArray[12]
::footerLines := aArray[13]
::footerDelim := aArray[14]
::headerPres := aArray[15]
::footerPres := aArray[16]
::columns := {}
::fit := aArray[20]
::optimize := aArray[21]
::arrayElement := 1
::refreshInterval := aArray[32]
::refreshWhenBlock := aArray[48]
::refreshBlock := aArray[47]
::rightButtonSelect := aArray[33]
::hasHeader := .f.
::hasFooter := .f.
::rightButtonDown := aArray[35]
::filter := aArray[36]
::ownerDrawBlock := aArray[49]
::fitMax := aArray[50]
::sqlSessionDict := aArray[58]
::sqlSessionFree := aArray[59]
::columnCache := aArray[60]

DEFAULT ::columnCache := 0

#if XPPVER > 02000000
IF Valtype(aArray[52]) == 'N'
  ::navigationMode := aArray[52]
ELSE
  ::navigationMode := XBPBRW_NAVIGATION_1XCOMPATIBLE
ENDIF
#endif

::recordObject := aArray[53]

IF Valtype(::recordObject) == 'O'
  ::navigateBlock := {||::recordObject:scatter()}
  Eval(::navigateBlock,::recordObject)
ELSE
  ::navigateBlock := {||nil}
ENDIF

#if XPPVER > 1900331
IF Valtype(::ownerDrawBlock) == 'B'
  ::drawMode := XBP_DRAW_OWNER
ENDIF
#endif

::cargo := { ::getListPointer, ;
             aGetList, ;
             aGetListItem[xGETLIST_CARGO], ;
             ::arrayElement, ;
             ::dataSource }

//
// J. Duijf 12-04-2011 Move up in this code
//
//xFont := aGetListItem[cGETLIST_FONT]
//IF Empty(xFont)
//  DEFAULT xFont := ::config:font
//ENDIF

//IF !Empty(xFont)
//  IF Valtype(xFont) == 'B'
//    xFont := Eval(xFont)
//  ENDIF
//  ::setFontCompoundName(Alltrim(xFont))
//ENDIF

::vScroll   := !aArray[17]
::hScroll   := !aArray[18]
::softTrack := !aArray[19]
::sizeCols  := !aArray[24]

::itemSelected := ;
         { || IIF(Valtype(::dataSource)='B',Eval(::dataSource,self),nil), ;
              IIF(Valtype(::mark)='N',::dataSource[::arrayElement,::mark] := ;
                           !::dataSource[::arrayElement,::mark],nil), ;
              IIF(Valtype(::mark)='B',Eval(::mark,self),nil), ;
              IIF(Valtype(::pointerBlock)=='B',Eval(::pointerBlock,::arrayElement),nil),;
              ::RefreshCurrent() }

IF Valtype(xOptions[23]) = 'B'
  ::itemSelected := DC_MergeBlocks(::itemSelected,xOptions[23])
ELSEIF Valtype(bBlock:=DC_GetBlock(aGetListItem[bGETLIST_ACTION])) = 'B'
  ::itemSelected := DC_MergeBlocks(::itemSelected,bBlock)
ENDIF

IF Valtype(::pointerBlock) == 'B'
  ::itemMarked := {||Eval(::pointerBlock,::arrayElement)}
ENDIF

IF Valtype(xOptions[22]) = 'B'
  ::itemMarked := DC_MergeBlocks(::itemMarked,xOptions[22])
  ::isItemMarkedCodeBlock := .t.
ENDIF

IF DC_GetWhenOnBrowseItemMarked()
  ::isItemMarkedCodeBlock := .t.
ENDIF

IF Valtype(xData) = 'C'
  xData := Alltrim(Upper(xData)) + '->'
ENDIF

IF Valtype(::thumbRecords) == 'B'
  lThumbLock := Eval(::thumbRecords)
ELSEIF Valtype(::dataSource) == 'C'
  DEFAULT ::thumbRecords := 32000
  lThumbLock := ::thumbRecords == 0 .OR. (::dataSource)->(RecCount()) > ::thumbRecords
ELSE
  lThumbLock := .f.
ENDIF

IF aGetListItem[nGETLIST_SUBTYPE] = BROWSE_ARRAY // array browse

  ::skipBlock     := {|n,o,x|x:=DC_SkipArray(n,self),x}

  ::goTopBlock    := {|o| ::arrayElement := ;
                           ArrayGoTop(self), ;
                           ::cargo[4] := ::arrayElement }

  ::goBottomBlock := {|o| ::arrayElement := ;
                           ArrayGoBottom(self), ;
                           ::cargo[4] := ::arrayElement}

  ::posBlock      := {|o,x|x := IIF(::descending,Len(::dataSource)-::arrayElement,::arrayElement), x}
  ::phyPosBlock   := {| x| x := ::arrayElement, x }

  ::lastPosBlock  := {||Len(::dataSource)}
  ::firstPosBlock := {||1}

  nPointer := aGetListItem[xGETLIST_OPTIONS5]
  IF Empty(nPointer) .AND. Valtype(::pointerBlock) == 'B'
    nPointer := Eval(::pointerBlock)
  ENDIF

  IF !Empty(::filter)
    ::arrayElement := ArrayGoTop(self)
  ELSEIF Valtype(nPointer) = 'N' .AND. nPointer > 0 .AND. nPointer <= Len(xData)
    ::arrayElement := nPointer
  ELSEIF ::descending
    ::arrayElement := Len(::dataSource)
  ENDIF

ELSEIF aGetListItem[nGETLIST_SUBTYPE] = BROWSE_SQLEXPRESS // SQLExpress browse

  ::skipBlock     := {|n,o| o:dataSource:Skipper(n) }
  ::goTopBlock    := {|o| o:dataSource:goTop() }
  ::goBottomBlock := {|o| o:dataSource:goBottom() }
  ::phyPosBlock   := {|o| o:dataSource:RecNo() }
  // Navigation blocks for VScroll bar
  ::posBlock      := {|o| o:dataSource:RecNo() }
  ::lastPosBlock  := {|o| o:dataSource:RecCount() }
  ::firstPosBlock := {|| 1 }

ELSEIF aGetListItem[nGETLIST_SUBTYPE] = BROWSE_MYSQL // MySQL browse

  ::skipBlock     := {|n,o| o:dataSource:Skipper(n) }
  ::goTopBlock    := {|o| o:dataSource:goTop() }
  ::goBottomBlock := {|o| o:dataSource:goBottom() }
  ::phyPosBlock   := {|o| o:dataSource:RecNo() }
  // Navigation blocks for VScroll bar
  ::posBlock      := {|o| o:dataSource:RecNo() }
  ::lastPosBlock  := {|o| o:dataSource:RecCount() }
  ::firstPosBlock := {|| 1 }

ELSEIF aGetListItem[nGETLIST_SUBTYPE] = BROWSE_DATAOBJECT // Database object  J.Duijf 27-11-2002

  ::skipBlock     := {|n|::dataSource:DbSkipper(n)    }
  ::goTopBlock    := {| |::dataSource:DbGoTop()       }
  ::goBottomBlock := {| |::dataSource:DbGoBottom()    }
  ::phyPosBlock   := {| |::dataSource:Recno()         }
  ::GoPosBlock    := {|n|::dataSource:DbGoPosition(n) }
  ::posBlock      := {| |::dataSource:DbPosition()    }
  ::lastPosBlock  := {||100}
  ::firstPosBlock := {|| 1 }

ELSEIF aGetListItem[nGETLIST_SUBTYPE] = BROWSE_ACESERVER // ACEServer++

   ::skipBlock     := {|n| ACEDbSkipper(n, ::dataSource) }
   ::goTopBlock    := {| | ::dataSource:GoTop()          }
   ::goBottomBlock := {| | ::dataSource:GoBottom()       }
   ::phyPosBlock   := {| | ::dataSource:Recno()          }
   ::posBlock      := {| | ::dataSource:Position()*10    }
   ::goPosBlock    := {|n| ::dataSource:GoPosition(n/10) }
   ::lastPosBlock  := {| | 1000 }
   ::firstPosBlock := {| | 0    }

ELSEIF aGetListItem[nGETLIST_SUBTYPE] = BROWSE_SPSQL // SPSQL silex

   ::skipBlock     := {|n| ::dataSource:SQLViewSkip(n) }
   ::goTopBlock    := {| | ::dataSource:GoTop()          }
   ::goBottomBlock := {| | ::dataSource:GoBottom()       }
 //  ::phyPosBlock   := {| | ::dataSource:dataarea:Recno()          }
  ::phyPosBlock   := {| |::dataSource:getNum()}
  ::posBlock      := {| |::dataSource:getNum()}

//    ::posBlock      := {| |posicionFisica(::dataSource:hIndex)}
  //  ::phyPosBlock   := {| |posicionFisica(::dataSource:hIndex)}
 //  ::posBlock      := {| | ::dataSource:dataarea:Recno() }

   ::goPosBlock    := {|n| ::dataSource:dataarea:goTo(::dataSource:getNum()) }

   ::lastPosBlock  := {| | ::dataSource:LastRec() }
   ::firstPosBlock := {| | 1 }

ELSEIF IsFunction('AceWBlock')

  // ** NOTE **
  // since macro compiled codeblocks are "late bound", they require the ACE Wrapper
  // functions to be passed exactly as they were defined in ACE.PRG (in most cases
  // with a leading underscore "_" character). (Boris - 2000/06/05)

  xData := StrTran(xData,"->","")
  ::skipBlock     := &( '{|n|AceWBlock("' + xData + '", {|x|_DbSkipper(x)},n)}' )
  ::goTopBlock    := &( '{| |AceWBlock("' + xData + '", {||_DbGoTop()})}' )
  ::goBottomBlock := &( '{| |AceWBlock("' + xData + '", {||_DbGoBottom()})}' )
  ::phyPosBlock   := &( '{| |AceWRecNo("' + xData + '")}' )

  // scrollbar navigation
  if &('AceWBlock("' + xData + '", {||OrdNumber()})') == 0
    ::posBlock      := &( '{| |AceWBlock("' + xData + '", {||_RecNo()})}' )
    ::goPosBlock    := &( '{|n|AceWBlock("' + xData + '", {|x|_DbGoTo(x)},n)}' )
    ::lastPosBlock  := &( '{| |AceWBlock("' + xData + '", {||_LastRec()})}' )
    ::firstPosBlock := {|| 1}
  else
    ::posBlock      := &( '{| |AceWBlock("' + xData + '", {||_DbPosition()})}' )
    ::goPosBlock    := &( '{|n|AceWBlock("' + xData + '", {|x|_DbGoPosition(x)},n)}' )
    ::lastPosBlock  := {|| 100}
    ::firstPosBlock := {|| 0 }
  endif

ELSEIF IsFunction('XPPSETSCOPE')

  ::firstPosBlock := {||1}
  ::LastPosBlock  := {||100}
  ::PosBlock      := {||(::dataSource)->(DbPosition(,::descending))}
  ::GoPosBlock    := {|n|(::dataSource)->(DbGoPosition(n,::descending))}
  ::skipBlock     := {|n|(::dataSource)->(DbSkipper(iif(::descending, n*(-1), n)))}
  ::goTopBlock    := {||(::dataSource)->(IIF(::descending,DbGoBottom(),DbGoTop()))}
  ::goBottomBlock := {||(::dataSource)->(IIF(::descending,DbGoTop(),DbGoBottom()))}
  ::phyPosBlock   := {||(::dataSource)->(Recno())}

ELSEIF ::isScope .AND. DC_ScopeExpress()

  ::skipBlock     := {|n|(::dataSource)->(DC_DbSkipper(n)) }
  ::goTopBlock    := {| |(::dataSource)->(DC_DbGoTop())    }
  ::goBottomBlock := {| |(::dataSource)->(DC_DbGoBottom()) }
  ::phyPosBlock   := {|n|IIF(Valtype(n)='N',(::dataSource)->(dbGoTo(n)), ;
                         (::dataSource)->(Recno()))}
  IF lThumbLock
    ::posBlock      := {||50}
    ::lastPosBlock  := {||100}
  ELSE
    ::posBlock      := {| |(::dataSource)->(DC_KeyNo32k())   }
    ::lastPosBlock  := {| |(::dataSource)->(DC_KeyCount32k())}
  ENDIF
  ::firstPosBlock := {| | 1 }

ELSEIF aGetListItem[nGETLIST_SUBTYPE] = BROWSE_ADSHANDLE // Advantage Server Handle

  ::firstPosBlock := {||1}
  ::lastPosBlock  := {||100}
  ::goTopBlock    := {||AdsGoToTop(::dataSource)}
  ::goBottomBlock := {||AdsGoToBottom(::dataSource)}
  ::skipBlock     := {|n|DC_AdsSkipper(::dataSource,n)}
  ::phyPosBlock   := {|n|n:=0,AdsGetRecordNum(::dataSource,ADS_IGNOREFILTERS,@n),n}
  ::GoPosBlock    := {|n|DC_AdsGoPosition(::dataSource,n)}
  ::posBlock      := {|n|DC_AdsPosition(::dataSource )}

ELSE

  IF lThumbLock
    ::posBlock      := {||50}
  ELSE
    ::PosBlock      := {||(::dataSource)->(DC_DbPosition(,::descending))}
  ENDIF
  ::firstPosBlock := {||1}
  ::lastPosBlock  := {||100}
  ::GoPosBlock    := {|n,x|x := (::dataSource)->(DC_DbGoPosition(n,::descending)),Eval(::navigateBlock),x}
  ::skipBlock     := {|n,x|x := (::dataSource)->(DC_DbSkipper(n,,::descending,::skipUnique)),Eval(::navigateBlock),x}
  ::goTopBlock    := {|  x|x := (::dataSource)->(IIF(::descending,DC_DbGoBottom(),DC_DbGoTop())),Eval(::navigateBlock),x}
  ::goBottomBlock := {|  x|x := (::dataSource)->(IIF(::descending,DC_DbGoTop(),DC_DbGoBottom())),Eval(::navigateBlock),x}
  ::phyPosBlock   := {|n  |IIF(Valtype(n)='N',(::dataSource)->(dbGoTo(n)),(::dataSource)->(Recno()))}

ENDIF

// ::headerRbDown := {|a,n,o|_HeaderRbDown( a,n,o,self )}

IF !Empty(::getList:getOptions[nGETOPT_HILITECOLOR])
  ::killInputFocus := {|x,y,o|  ;
                        DC_HiliteGets(o,2,o:getList:getOptions[nGETOPT_HILITECOLOR])}
  ::setInputFocus := {|x,y,o| ;
                       DC_HiliteGets(o,1,o:getList:getOptions[nGETOPT_HILITECOLOR])}
ENDIF

bBlock := DC_GetBlock(aGetListItem[bGETLIST_GROUP],'O')
IF Valtype( bBlock ) = 'B'
  Eval( bBlock, self )
ENDIF

bBlock := DC_GetBlock(aGetListItem[bGETLIST_PREEVAL])
IF Valtype(bBlock) = 'B'
  Eval( bBlock, self )
ENDIF

cID := aGetListItem[cGETLIST_ID]

IF (!Empty(DC_AutoRestoreBrowse()) .AND. !Empty(cID) .AND. !::noRestore) .OR. ;
      !Empty(::restoreConfig)

  IF Valtype(::restoreConfig) = 'C'
    aConfig := DC_ARestore(::restoreConfig)
  ELSEIF Valtype(::restoreConfig) = 'A'
    aConfig := ::restoreConfig
  ELSE
    IF DC_AutoRestoreBrowse()[3]
      ::restoreProc := ProcName(3)
      IF 'READGUI' $ ::restoreProc
        ::restoreProc := ProcName(4)
      ENDIF
    ELSE
      ::restoreProc := ''
    ENDIF

    ::restoreProc += ':' + DC_XtoC(cID)

    IF Valtype(DC_AutoRestoreBrowse()[1]) == 'B'
      aConfig := Eval(DC_AutoRestoreBrowse()[1],::restoreProc)
    ELSE
      aConfig := DC_RegQuery(DC_AutoRestoreBrowse()[1],DC_AutoRestoreBrowse()[2], ;
                             ::restoreProc)
    ENDIF
  ENDIF

  IF Valtype(aConfig) = 'A'

    nPointer := 1
    nColumns := 0
    ::defaultColumnsGetList := {}
    ASize(aConfig,7)

    DO WHILE aGetList[::getListPointer+nPointer,nGETLIST_TYPE] $ { GETLIST_BROWSECOL, GETLIST_SETPARENT } .AND. ;
             ::getListPointer+nPointer <= Len(aGetList)
      IF aGetList[::getListPointer+nPointer,nGETLIST_TYPE] == GETLIST_BROWSECOL
        nColumns++
      ENDIF
      nPointer++
    ENDDO

    BEGIN SEQUENCE

    IF nColumns # Len(aConfig[3])
      IF Valtype(::configMessageBlock) == 'B'
        Eval(::configMessageBlock,self)
      ELSE
        DCMSGBOX 'Your Saved Browse configuration cannot be restored:', ;
                  DC_AutoRestoreBrowse()[2]+'\'+::restoreProc ;
          TIMEOUT 5

        IF Valtype(DC_AutoRestoreBrowse()[1]) == 'B'
          Eval(DC_AutoRestoreBrowse()[1],::restoreProc)
        ELSE
          DC_RegWrite(DC_AutoRestoreBrowse()[1],DC_AutoRestoreBrowse()[2], ;
                     ::restoreProc, 0)
        ENDIF
      ENDIF
      BREAK
    ENDIF

    nPointer := 1
    DO WHILE aGetList[::getListPointer+nPointer,nGETLIST_TYPE] == GETLIST_BROWSECOL .AND. ;
       ::getListPointer+nPointer <= Len(aGetList)
      nFound := AScan(aConfig[3],{|a|a[1]=nPointer})
      AAdd(::defaultColumnsGetList,AClone(aGetList[::getListPointer+nPointer]))
      aGetList[::getListPointer+nPointer,cGETLIST_LEVEL] := { nFound, nPointer }
      IF nFound > 0
        aGetList[::getListPointer+nPointer,nGETLIST_WIDTH] := aConfig[3,nFound,2,1]
        aGetList[::getListPointer+nPointer,lGETLIST_PIXEL] := .T.
        ASize(aConfig[3,nFound],3)
        IF !Empty(aConfig[3,nFound,3])
          aGetList[::getListPointer+nPointer,cGETLIST_FONT] := Trim(aConfig[3,nFound,3])
        ENDIF
        IF !Empty(aConfig[4]) .OR. !Empty(aConfig[5])
          DEFAULT aGetList[::getListPointer+nPointer,xGETLIST_OPTIONS2] := {}
          ASize(aGetList[::getListPointer+nPointer,xGETLIST_OPTIONS2],4)
          aGetList[::getListPointer+nPointer,xGETLIST_OPTIONS2][3] := aConfig[4]
          aGetList[::getListPointer+nPointer,xGETLIST_OPTIONS2][4] := aConfig[5]
        ENDIF
        IF !Empty(aConfig[6]) .OR. !Empty(aConfig[7])
          DEFAULT aGetList[::getListPointer+nPointer,xGETLIST_OPTIONS8] := {}
          ASize(aGetList[::getListPointer+nPointer,xGETLIST_OPTIONS8],4)
          aGetList[::getListPointer+nPointer,xGETLIST_OPTIONS8][3] := aConfig[6]
          aGetList[::getListPointer+nPointer,xGETLIST_OPTIONS8][4] := aConfig[7]
        ENDIF
      ELSE
        // aGetList[::getListPointer+nPointer,bGETLIST_HIDE] := {||.t.}
      ENDIF
      nPointer++
    ENDDO
    ASort( aGetList, ::getListPointer+1, nPointer-1, ;
            {|a,b|a[cGETLIST_LEVEL,1] < b[cGETLIST_LEVEL,1]} )

    END SEQUENCE
  ENDIF
ENDIF

::scrollBarTipBlock := aGetListItem[xGETLIST_OPTIONS9]
IF Valtype(::scrollBarTipBlock) == 'B'
  ::getList:isDataToolTips := .t.
ENDIF

IF ::tagEnable
  ::keyboard := DC_MergeBlocks(::keyboard,{|n|::Tag(,,n)} )
ENDIF

IF !Empty(::columnCache)
  ::keyboard := DC_MergeBlocks(::keyboard,{|n|::CachedColumnKey(n,aGetList)})
  ::pan := DC_MergeBlocks(::pan,{|n,mp2|::CachedColumnPan(@n,mp2,aGetList)})
ENDIF

RETURN self

* ------------------

METHOD DC_XbpBrowse:create( oParent, oOwner, aPos, aSize, aPres, lVisible )

LOCAL aChildren, i

::XbpBrowse:create( oParent, oOwner, aPos, aSize, aPres, lVisible )

DEFAULT oParent := ::setParent()

IF Valtype(::hideBlock)='B' .AND. Eval(::hideBlock)
  ::hide()
ENDIF

::DC_XbaseParts:setOrig(self)

aChildren := ::childList()
FOR i := 1 TO Len(aChildren)
  IF aChildren[i]:isDerivedFrom("XbpScrollBar")
    IF aChildren[i]:type == XBPSCROLL_VERTICAL .AND. Valtype(::scrollBarTipBlock) == 'B'
      aChildren[i]:cargo :=  {{||.t.},::scrollBarTipBlock,'',0,0,0,nil,::dataSource}
      // :cargo[1] = Enable Code Block
      // :cargo[2] = Return Code Block
      // :cargo[3] = Value to display in tooltip
      // :cargo[4] = Last moused cell record offset
      // :cargo[5] = Last selected record
      // :cargo[6] = Last moused cell record
      // :cargo[7] = Pointer to Browse object
      // :cargo[8] = Alias(), Array or SQLExpress object
    ENDIF
    IF IsMemberVar(aChildren[i],"isVisualStyle")
      aChildren[i]:useVisualStyle := ::config:useVisualStyleScrollBars
    ENDIF
  ENDIF
NEXT

RETURN self

* --------------------

METHOD DC_Xbpbrowse:right()

::XbpBrowse:right()

RETURN self

* --------

METHOD DC_XbpBrowse:displayedRows()

LOCAL nRows := 0

IF ::displayedRows > -1
  nRows := ::displayedRows
ELSEIF Valtype(::dataSource) == 'C'
  IF !Empty((::dataSource)->(dbFilter()))
    nRows := -1
  ELSE
    nRows := (::dataSource)->(DC_KeyCount())
  ENDIF
  IF nRows == -1
    nRows := (::dataSource)->(RecCount())
  ENDIF
ELSEIF Valtype(::dataSource) == 'A'
  nRows := Len(::dataSource)
ENDIF

RETURN nRows

* -------------------

METHOD DC_XbpBrowse:orderBy( nColumn, oColumn )

LOCAL cFieldName, cTagName, lStatus := .f.

cFieldName := oColumn:objectVarName
IF Empty(cFieldName)
  cFieldname := oColumn:sqlField
ENDIF
IF Empty(cFieldname)
  cFieldName := oColumn:headerBlock
ENDIF
IF Valtype(::dataSource) == 'C'
  IF !Empty(oColumn:indexTag)
    cTagName := Upper(::indexTag)
  ELSEIF Valtype(cFieldName) == 'C'
    cTagName := ::findGoodIndex(Alltrim(Upper(cFieldName)))
  ENDIF
  IF !Empty(cTagName)
    (::dataSource)->(OrdSetFocus(cTagName))
    lStatus := (::dataSource)->(OrdSetFocus()) == cTagName
  ENDIF
ELSEIF Valtype(::dataSource) == 'A'
  IF Valtype(::dataSource[1,1]) == 'O' // record object
    IF Valtype(cFieldName) == 'C'
      ASort(::dataSource,,,{|a,b|a[1]:&cFieldName<b[1]:&cFieldName})
      lStatus := .t.
    ENDIF
  ELSEIF Valtype(::dataSource[1]) == 'O' // record object
    IF Valtype(cFieldName) == 'C'
      ASort(::dataSource,,,{|a,b|a:&cFieldname<b:&cFieldName})
      lStatus := .t.
    ENDIF
  ELSE
    ASort(::dataSource,,,{|a,b|a[oColumn:arrayPointer]<b[oColumn:arrayPointer]})
  ENDIF
ENDIF

IF Valtype(::filterRefreshBlock) == 'B'
  Eval(::filterRefreshBlock,self)
ENDIF

RETURN lStatus

* ----------------

METHOD DC_XbpBrowse:FindGoodIndex( cFieldName )

LOCAL i, aIndexInfo := ::IndexInfo(), cIndexKey, aTokens

FOR i := 1 TO Len(aIndexInfo)
  cIndexKey := aIndexInfo[i,2]
  aTokens := DC_TokenArray(Upper(cIndexKey),'+')
  IF Empty(aTokens)
    LOOP
  ENDIF
  aTokens[1] := AllTrim(aTokens[1])
  IF Upper(aTokens[1]) == cFieldName
    RETURN aIndexInfo[i,1]
  ELSEIF Left(Upper(aTokens[1]),5) == 'DTOS('
    aTokens[1] := Strtran(aTokens[1],'DTOS(','')
  ELSEIF Left(Upper(aTokens[1]),6) == 'UPPER('
    aTokens[1] := Strtran(aTokens[1],'UPPER(','')
  ENDIF
  aTokens[1] := Strtran(aTokens[1],'(','')
  aTokens[1] := Strtran(aTokens[1],')','')
  IF Upper(aTokens[1]) == cFieldName
    RETURN aIndexInfo[i,1]
  ENDIF
NEXT

RETURN ''

* --------------------

METHOD DC_XbpBrowse:IndexInfo()

LOCAL aInfo, aIndex[0], i

IF Valtype(::dataSource) == 'C'
  FOR i := 1 TO 100
    IF EMPTY(OrdName(i))
      EXIT
    ENDIF
    aInfo := { (::dataSource)->(OrdName(i)), ;
               (::dataSource)->(IndexKey(i)), ;
               (::dataSource)->(OrdFor(i)), ;
               (::dataSource)->(OrdIsUnique(i)), ;
               (::dataSource)->(OrdIsDescend(i)) }
    AAdd( aIndex, aInfo )
  NEXT
ENDIF

RETURN aIndex

* --------------------

METHOD DC_XbpBrowse:customDrawCell( oPS, aInfo )

IF !Empty(::ownerDrawBlock)
  RETURN Eval( ::ownerDrawBlock, oPS, aInfo, self )
ENDIF

RETURN .t.

* --------------------

METHOD DC_XbpBrowse:rowColor()

LOCAL aColor, aBrowseColor

IF Valtype(::dataSource) == 'C'
  IF ::tagEnable .AND. (::dataSource)->(DC_Tagged())
    RETURN {::tagColorFG, ::tagColorBG}
  ELSEIF (::dataSource)->(RecNo()) <> ::nBRow
    ::lBRow := !::lBRow
    ::nBRow := (::dataSource)->(RecNo())
  ENDIF
ELSEIF Valtype(::dataSource) == 'A' .AND. Len(::dataSource) > 0
  IF ::tagEnable .AND. ::arrayElement > 0 .AND. ::dataSource[::arrayElement,::tagElement]
    RETURN {::tagColorFG, ::tagColorBG}
  ELSEIF (::arrayElement) % 2 == 0
    ::lBRow := .t.
  ELSE
    ::lBrow := .f.
  ENDIF
ELSE
  ::lBrow := .t.
ENDIF

IF Valtype(::browseColor) == 'A'
  IF ::lBRow
    aColor := {::browseColor[1],::browseColor[2]}
  ELSE
    aColor := {::browseColor[3],::browseColor[4]}
  ENDIF
ELSE
  aColor := { nil, nil}
ENDIF

::rowColor := aColor

RETURN aColor

* --------------------

METHOD DC_XbpBrowse:destroy()

LOCAL aConfig

IF Valtype(::timer) == 'O' .AND. ::timer:isDerivedFrom('DC_SetTimerEvent')
  ::timer:destroy()
  Sleep(10)
ENDIF

IF !Empty(DC_AutoRestoreBrowse()) .AND. !Empty(::getListPointer) .AND. ;
   !Empty(::getList:getListArray) .AND. !::noRestore .AND. !Empty(::restoreProc)

  aConfig := ::save()
  IF Valtype(DC_AutoRestoreBrowse()[1]) == 'B'
    Eval(DC_AutoRestoreBrowse()[1],::restoreProc,aConfig)
  ELSE
    DC_RegWrite(DC_AutoRestoreBrowse()[1],DC_AutoRestoreBrowse()[2], ;
                ::restoreProc, aConfig)
  ENDIF
ENDIF

::Xbpbrowse:destroy()

//
// J. Duijf 12-04-2011
// Remove references (if any)
//
/*
::Datalink        := nil
::DataSource      := nil
::Cargo           := nil
::GetList         := nil
::xbp             := nil
::config          := nil
::GoBottomBlock   := nil
::GoPosBlock      := nil
::PhyPosBlock     := nil
::PosBlock        := nil
::LastPosBlock    := nil
::GoTopBlock      := nil
::SkipBlock       := nil
::HitBottomBlock  := nil
::HitTopBlock     := nil
::ItemSelected    := nil
::LastPosBlock    := nil
::StableBlock     := nil
::KillInputFocus  := nil
::SetInputFocus   := nil
::BrowsePres      := nil
::ItemSelected    := nil
::StableBlock     := nil
::KillInputFocus  := nil
::OrigFont        := nil
::Cargo           := nil
::Timer           := nil
*/

RETURN self

* -----------------

METHOD DC_XbpBrowse:restoreDefaults()

LOCAL i, aColumns[::colCount], GetList

IF Empty(::defaultColumnsGetList)
  RETURN self
ENDIF

::lockUpdate(.t.)

FOR i := 1 TO Len(aColumns)
  ::delColumn(1)
NEXT

GetList := ::defaultColumnsGetList

DCREAD GUI PARENT self EXIT SAVE

::lockUpdate(.f.)

RETURN self

* --------------------

METHOD DC_XbpBrowse:fitColumns()

LOCAL nBrowseWidth, aChildList, i, nDelta, nColWidth, oPS, oColumn

oPS := ::lockPs()
nBrowseWidth := ::currentSize()[1]
aChildList := ::childList()
FOR i := 1 TO Len(aChildList)
  IF aChildList[i]:isDerivedFrom('XbpScrollBar') .AND. ;
    aChildList[i]:type == XBPSCROLL_VERTICAL
    nBrowseWidth -= aChildList[i]:currentSize()[1]
  ENDIF
NEXT

nColWidth := 0
FOR i := 1 TO ::colCount
  nColWidth += ::getColumn(i):currentSize()[1]
NEXT

nDelta := (nBrowseWidth-nColWidth)/::colCount
FOR i := 1 TO ::colCount
  oColumn := ::getColumn(i)
  oColumn:setSize({oColumn:currentSize()[1]+nDelta,oColumn:currentSize()[2]})
NEXT
::unLockPs(oPS)
::setSize(::currentSize())

RETURN nil

* --------------------

METHOD DC_XbpBrowse:resizeColumns( aOldSize, aNewSize )

LOCAL i, nDelta, oPS, oColumn, aColumns[0]

IF ::xbpBrowse:status() = 0
  RETURN nil
ENDIF

oPS := ::lockPs()

FOR i := 1 TO ::colCount
  oColumn := ::getColumn(i)
  IF !oColumn:noAutoResizeColumn
    AAdd(aColumns,oColumn)
  ENDIF
NEXT

nDelta := Int((aNewSize[1]-aOldSize[1])/Len(aColumns))
FOR i := 1 TO Len(aColumns)
  oColumn := aColumns[i]
  oColumn:setSize({oColumn:currentSize()[1]+nDelta,oColumn:currentSize()[2]})
NEXT
::unLockPs(oPS)
::setSize(::currentSize())

RETURN nil

* --------------------

METHOD DC_XbpBrowse:refreshView()

LOCAL nRows := ::rowPos-1, i

FOR i := 1 TO nRows
  Eval(::skipBlock,-1)
NEXT
::refreshAll()
FOR i := 1 TO nRows
  Eval(::skipBlock,1)
NEXT

RETURN nil

* --------------------

METHOD DC_XbpBrowse:handleEvent( nEvent, mp1, mp2 )

IF nEvent == DCGUI_EVENT_BROWSE_REFRESH
  IF ::isVisible()
    IF Empty(::refreshWhenBlock) .OR. Eval(::refreshWhenBlock)
      IF Valtype(::refreshBlock) == 'B'
        Eval(::refreshBlock,self)
      ELSE
        ::Refresh()
      ENDIF
    ENDIF
  ENDIF
ELSE
  RETURN ::XbpbrowseHandleEvent( nEvent, mp1, mp2 )
ENDIF
RETURN self

* --------------------

METHOD DC_XbpBrowse:setTimerRefresh()

DEFAULT ::refreshInterval := 0

IF ::refreshInterval > 0
  ::timer := DC_SetTimerEvent():new(::refreshInterval, ;
              {||PostAppEvent(DCGUI_EVENT_BROWSE_REFRESH,,,self)})
ENDIF

RETURN self

* --------------------

METHOD DC_XbpBrowse:optimize()

LOCAL i, oHeading

FOR i := 1 TO ::colCount
  oHeading := ::getColumn(i):heading
  _Optimize(nil,nil,oHeading)
NEXT

RETURN self

* --------------------

METHOD DC_XbpBrowse:dataDisplay( lEnable )

LOCAL nRecNo

IF Empty(::navBlocks)
  ::navBlocks := { ;
    ::posBlock, ;
    ::firstPosBlock, ;
    ::lastPosBlock, ;
    ::goPosBlock, ;
    ::skipBlock, ;
    ::goTopBlock, ;
    ::goBottomBlock, ;
    ::phyPosBlock, ;
    ::arrayElement }
ENDIF

IF lEnable
  ::posBlock := ::navBlocks[1]
  ::firstPosBlock := ::navBlocks[2]
  ::lastPosBlock := ::navBlocks[3]
  ::goPosBlock := ::navBlocks[4]
  ::skipBlock := ::navBlocks[5]
  ::goTopBlock := ::navBlocks[6]
  ::goBottomBlock := ::navBlocks[7]
  ::phyPosBlock := ::navBlocks[8]
  ::arrayElement := ::navBlocks[9]
  ::dataEnabled := .t.
  ::refreshAll()
ELSE
  ::posBlock := {||0}
  ::firstPosBlock := {||0}
  ::lastPosBlock  := {||0}
  ::GoPosBlock    := {||0}
  ::skipBlock     := {||0}
  ::goTopBlock    := {||0}
  ::goBottomBlock := {||0}
  ::phyPosBlock   := {||0}
  IF Valtype(::dataSource) == 'C'
    nRecNo := (::dataSource)->(RecNo())
    (::dataSource)->(dbGoBottom())
    (::dataSource)->(dbSkip())
    ::refreshAll()
    (::dataSource)->(dbGoto(nRecNo))
  ELSEIF Valtype(::dataSource) == 'A'
    ::navBlocks[9] := ::arrayElement
    ::arrayElement := 0
    ::refreshAll()
  ENDIF
  ::dataEnabled := .f.
ENDIF

RETURN self

* --------------------

METHOD DC_XbpBrowse:refresh()

LOCAL i, aEvents[0], nEvent, mp1, mp2, oXbp, nRecNo

IF !::dataEnabled
  RETURN self
ENDIF

if ::dataSourceType = BROWSE_DATABASE                             // J. Duijf 27-11-2002
   nRecNo   := (::dataSource)->(RecNo())                          // J. Duijf 27-11-2002
elseif ::dataSourceType = BROWSE_DATAOBJECT                       // J. Duijf 27-11-2002
   nRecno   := ::dataSource:Recno()                               // J. Duijf 27-11-2002
endif                                                             // J. Duijf 27-11-2002

PostAppEvent(xbeP_User+10,12,13,self)

IF Valtype(::freezeLeft) = 'A' .AND. Empty(::setLeftFrozen())
  ::setLeftFrozen( ::freezeLeft )
ENDIF
IF Valtype(::freezeRight) = 'A' .AND. Empty(::setRightFrozen())
  ::setRightFrozen( ::freezeRight )
ENDIF

FOR i := 1 TO ::colCount
  ::getColumn(i):refresh()
NEXT

::RefreshAll()

nEvent := -1
DO WHILE nEvent # 0
  nEvent := AppEvent(@mp1,@mp2,@oXbp,1)
  IF nEvent # 0 .AND. nEvent # xbeBRW_ItemMarked
    AAdd(aEvents,{nEvent,mp1,mp2,oXbp})
  ENDIF
ENDDO
FOR i := 1 TO Len(aEvents)
  PostAppEvent( aEvents[i,1], aEvents[i,2], aEvents[i,3], aEvents[i,4] )
NEXT

IF Valtype(nRecNo) == 'N'
   if ::dataSourceType = BROWSE_DATABASE                             // J. Duijf 27-11-2002
      (::dataSource)->(dbGoTo(nRecNo))
   elseif ::dataSourceType = BROWSE_DATAOBJECT                       // J. Duijf 27-11-2002
      ::dataSource:DbGoto(nRecno)                                    // J. Duijf 27-11-2002
   endif                                                             // J. Duijf 27-11-2002
ENDIF

RETURN self

* ---------------------------

METHOD DC_XbpBrowse:Resize( aOldSize, aNewSize )

::RefreshAll()
::InvalidateRect ( { 0, 0, aNewSize[1], aNewSize[2] } )

RETURN ::XbpStatic:Resize( aOldSize, aNewSize )

* ---------------------------

METHOD DC_XbpBrowse:Setsize( aSize )

LOCAL lStatus, bError := ErrorBlock({|e|Break(e)})

BEGIN SEQUENCE
lStatus := ::XbpBrowse:setSize(aSize)
END SEQUENCE

ErrorBlock(bError)

RETURN lStatus

* ---------------------------

METHOD DC_XbpBrowse:RefreshAll( aNewArray )

LOCAL nRecno := 0, lThumbLock

IF !::dataEnabled .or. ::Status <> XBP_STAT_CREATE
  RETURN self
ENDIF

if ::dataSourceType = BROWSE_DATABASE                             // J. Duijf 27-11-2002
   IF Empty((::dataSource)->(dbScope(1))) .AND. !Empty((::dataSource)->(DC_SetScope(0)))
     nRecno := (::dataSource)->(Recno())             // Save current record
     (::dataSource)->(DC_DbSkip())                   // Next
     (::dataSource)->(DC_DbSkip(-1))                 // Previous
     if (::dataSource)->(Bof()) .or. (::dataSource)->(Eof()) .OR. ;
       (::dataSource)->(RecNo()) != nRecno           // Current record NOT within scope, or deleted
       (::dataSource)->(DC_DbGoTop())                // To first record
       ::GoTop()                                     // Same goes for xbpBrowse
     endif
   ENDIF
elseif ::dataSourceType = BROWSE_DATAOBJECT                          // J. Duijf 27-11-2002
   nRecno := ::dataSource:Recno()                                    // J. Duijf 27-11-2002
   ::dataSource:DbSkip()                                             // J. Duijf 27-11-2002
   ::dataSource:DbSkip(-1)                                           // J. Duijf 27-11-2002
   if ::dataSource:Bof() .or. ::dataSource:Eof() .or. ;              // J. Duijf 27-11-2002
      ::datasource:Recno()  != nRecno                                // J. Duijf 27-11-2002
//      ::dataSource:DbGoTop()                                         // J. Duijf 27-11-2002 ::Gotop() does same
      ::GoTop()                                                      // J. Duijf 27-11-2002
   endif
elseif ::dataSourceType = BROWSE_ARRAY                               // J. Duijf 26-03-2011 Force initialisation of nPointer
   IF Valtype(aNewArray) == 'A'
     ::dataSource := aNewArray
   ENDIF
   ::arrayElement       := Min(::arrayElement,Len(::dataSource))     // J. Duijf 19-04-2015 Keep ::arrayElement in range

   If Len(::dataSource) > 0 .and. ::arrayElement = 0                 // J. Duijf 19-04-2015 Set nPointer to 1 if data availeble
     IF Empty(::filter)
       ::arrayElement    := 1
     ENDIF
   Endif
   IIF(Valtype(::pointerBlock)=='B',Eval(::pointerBlock,::arrayElement),nil)
endif

IF Valtype(::thumbRecords) == 'B' .AND. Valtype(::dataSource) = 'C'
  lThumbLock := Eval(::thumbRecords)
  IF ::isScope .AND. DC_ScopeExpress()
    IF lThumbLock
      ::posBlock      := {||50}
      ::lastPosBlock  := {||100}
    ELSE
      ::posBlock      := {| |(::dataSource)->(DC_KeyNo32k())   }
      ::lastPosBlock  := {| |(::dataSource)->(DC_KeyCount32k())}
    ENDIF
  ELSE
    IF lThumbLock
      ::posBlock      := {||50}
    ELSE
      ::PosBlock      := {||(::dataSource)->(DC_DbPosition(,::descending))}
    ENDIF
  ENDIF
ENDIF

Return ::xbpBrowse:RefreshAll()

* ---------------------------

METHOD DC_XbpBrowse:WhenHideEval( lEnable, lShow, lProtect )

IF Valtype(::whenBlock) = 'B'
  lEnable:= Eval(::whenBlock,self)
  IF lEnable
    IF !::isEnabled()
      ::enable()
    ENDIF
  ELSE
    IF ::isEnabled()
      ::disable()
    ENDIF
  ENDIF
ENDIF

IF Valtype(::hideBlock) = 'B'
  lShow := !Eval(::hideBlock,self)
  IF lShow
   //
   // J. Duijf 15-08-2016 Prevent needless flickering and invisible rows !.
   // Only show/hide when needed
   //
    If !::IsVisible()
      ::show()
    Endif
  ELSEIF !lShow
    If ::IsVisible()
      ::hide()
    Endif
  ENDIF
ENDIF

IF Valtype(lProtect) == 'L' .AND. Valtype(::protectBlock) == 'B'
  lProtect := Eval(::protectBlock)
ENDIF

RETURN self

* --------------------

METHOD DC_XbpBrowse:Navigate( mp1, mp2 )
IF Valtype(::navigateBlock) == 'B'
  Eval(::navigateBlock)
ENDIF
IF mp1 == 12
  ::XbpBrowse:forceStable()
ENDIF
RETURN ::XbpBrowse:navigate( mp1, mp2 )

* --------------------

METHOD DC_XbpBrowse:forceStable()

LOCAL bErrorBlock

IF ::isTrapErrors
  bErrorBlock := ErrorBlock({|e|_TrapErrors(e)})
ENDIF

BEGIN SEQUENCE
::XbpBrowse:forceStable()
END SEQUENCE
IF ::isTrapErrors
  ErrorBlock(bErrorBlock)
ENDIF


RETURN self

* -----------

STATIC FUNCTION _TrapErrors(e)

BREAK
RETURN nil

* --------------------

METHOD DC_XbpBrowse:FieldBlock( nField )

RETURN {|x| iif(PCount()==0, ::dataSource:fieldGet(nField), ;
                             ::dataSource:fieldPut(nField, x))}

* ------------------

METHOD DC_XbpBrowse:save( cFileName, nStatus )

LOCAL aConfig[3], i, oColumn, nScale

nScale := ::getList:scaleFactor[3]
IF Empty(nScale)
  nScale := 1
ENDIF

nStatus := 1

aConfig[1] := ::currentPos()
aConfig[2] := ::currentSize()
aConfig[3] := {}

FOR i := 1 TO ::colCount
  oColumn := ::getColumn(i)
  Aadd( aConfig[3], { oColumn:columnPos, ;
                      { oColumn:currentSize()[1]/nScale, oColumn:currentSize()[2] }, ;
                      oColumn:origFonts[1], ;
                      oColumn:origFonts[2], ;
                      oColumn:heading:currentSize()[2], ;
                      oColumn:origFonts[3], ;
                      oColumn:footing:currentSize()[2] } )
NEXT

IF !Empty(cFileName)
  IF !DC_ASave(aConfig,cFileName)
    nStatus := 0
  ENDIF
ENDIF

RETURN aConfig

* ------------------

METHOD DC_XbpBrowse:restore( acFileName )

LOCAL aConfig, i, j, oColumn, aColumns[::colCount]

IF Valtype(acFileName) == 'C'
  aConfig := DC_ARestore(acFileName)
  IF Valtype(aConfig) # 'A'
    RETURN .f.
  ENDIF
ELSEIF Valtype(acFileName) == 'A'
  aConfig := acFileName
ELSE
  RETURN .f.
ENDIF

::lockUpdate(.t.)
FOR i := 1 TO ::colCount
  aColumns[i] := ::getColumn(i)
NEXT

FOR i := 1 TO Len(aColumns)
  ::delColumn(1)
NEXT
FOR i := 1 TO Len(aConfig[3])
  oColumn := NIL
  FOR j := 1 TO Len(aColumns)
    oColumn := aColumns[j]
    IF aConfig[3,i,1] == oColumn:columnPos
      EXIT
    ENDIF
  NEXT
  IF !oColumn==NIL
    ::addColumn(oColumn )
    oColumn:setSize(aConfig[3,i,2])
    oColumn:setHeaderCaption()
    oColumn:setHeaderColor()
    oColumn:setFooterCaption()
    oColumn:setFooterColor()
  ENDIF
NEXT
::setPos(aConfig[1])
::setSize(aConfig[2])
::lockUpdate(.f.)
::forceStable()
::xbpBrowse:refreshAll()

RETURN .t.

* -------------

METHOD DC_XbpBrowse:CachedColumnKey(nKey,aGetList,lAll)

IF ::isColumnCacheAutomatic
  IF nKey == xbeK_RIGHT .AND. ::colPos == ::colCount
    IF ::nextCachedColumn(aGetList,lAll)
      ::right()
      ::forceStable()
      DC_CompleteEvents()
    ENDIF
  ELSEIF nKey == xbeK_END
    IF ::nextCachedColumn(aGetList,.t.)
     ::lastcol()
     ::forceStable()
     DC_CompleteEvents()
    ENDIF
  ENDIF
ENDIF

/*
IF ::isColumnCacheAutomatic .AND. nKey == xbeK_RIGHT .AND. ::colPos == ::colCount
  IF ::nextCachedColumn(aGetList,lAll)
    ::right()
    ::forceStable()
    DC_CompleteEvents()
  ENDIF
ENDIF
*/

RETURN nil

* -------------

METHOD DC_XbpBrowse:CachedColumnPan(nPan,mp2,aGetList)

IF ::isColumnCacheAutomatic .AND. (nPan == XBPBRW_Pan_Right .OR. nPan == XBPBRW_Pan_LastCol) .AND. ;
   ::hScrollObject:getData() >= ::hScrollObject:setRange()[2]
  IF ::nextCachedColumn(aGetList)
    ::panRight()
    DC_CompleteEvents()
  ENDIF
ELSE
  ::Pan( nPan, mp2 )
ENDIF

RETURN nil

* -------------

METHOD DC_XbpBrowse:NextCachedColumn( aGetList, lAll )

LOCAL i, aPickedColumns[0], oGetList, bEval, cSubClass, oColumn, ;
      lStatus := .t.

DEFAULT lAll := .f.

IF Empty(aGetList)
  aGetList := ::cargo[2]
ENDIF

FOR i := 1 TO Len(aGetList)
  IF aGetList[i,nGETLIST_TYPE] == GETLIST_BROWSECOL .AND. ;
      !Empty(aGetList[i,bGETLIST_PARENT]) .AND. ;
      Eval(aGetList[i,bGETLIST_PARENT]) == self .AND. ;
      Valtype( aGetList[i,oGETLIST_OBJECT] ) # 'O'
    AAdd(aPickedColumns, ;
        {aGetlist[i,cGETLIST_CAPTION],aGetList[i,cGETLIST_SUBCLASS],i,0,.f.,aGetList[i,bGETLIST_EVAL],1})
    IF !lAll
      EXIT
    ENDIF
  ENDIF
NEXT

IF Empty(aPickedColumns)
  lStatus := .f.
ENDIF

IF lStatus
  oGetList := DC_GetDataStore(aGetList)
  IF lAll
    ::lockUpdate(.t.)
  ENDIF
  FOR i := 1 TO Len(aPickedColumns)
    oGetList:getListPointer := aPickedColumns[i,3]
    bEval := aPickedColumns[i,7]
    cSubClass := aPickedColumns[i,2]
    DEFAULT cSubClass := 'DC_XbpColumn()'
    oColumn := &cSubClass
    oColumn := oColumn:new( self, oGetList, 0 )
    aGetList[oGetList:getListPointer,oGETLIST_OBJECT] := oColumn
    IF Valtype(bEval) == 'B'
      Eval(bEval,oColumn)
    ENDIF
    // IF lAll .OR. i == Len(aPickedColumns)
    IF i == Len(aPickedColumns) .AND. !lAll
      ::refreshAll()
      ::stabilizeColumns()
      ::lastCol()
    ENDIF
  NEXT
  IF lAll
    ::lockUpdate(.f.)
    ::stabilizeColumns()
    ::invalidateRect()
    ::setSize(::currentSize())
    ::lastCol()
    ::refreshAll()
  ENDIF
ENDIF

RETURN lStatus

* -------------

METHOD DC_XbpBrowse:ChooseCachedColumn( aGetList, oOwner, lNextCol, aColumns )

LOCAL oGetList, GetList[0], oBrowse, i, nColumn, aFrozenColumns, ;
      oXbp, lStatus := .t., cSubClass, aPickedColumns[0], aAvailColumns[0], ;
      aOptions, oStatic, GetOptions, lGoToLast, nInsertedColumns, ;
      nLockedColumns, bEval, oColumn

DEFAULT aColumns := {}

IF Empty(aGetList)
  aGetList := ::cargo[2]
ENDIF

FOR i := 1 TO Len(aGetList)
  IF aGetList[i,nGETLIST_TYPE] == GETLIST_BROWSECOL .AND. ;
      !Empty(aGetList[i,bGETLIST_PARENT]) .AND. ;
      Eval(aGetList[i,bGETLIST_PARENT]) == self
    IF Valtype( oXbp := aGetList[i,oGETLIST_OBJECT] ) # 'O'
      AAdd(aAvailColumns, ;
         {aGetlist[i,cGETLIST_CAPTION],aGetList[i,cGETLIST_SUBCLASS],i,0,.f.,aGetList[i,bGETLIST_EVAL],Len(aAvailColumns)+1})
    ELSEIF !oXbp:DisplayIndex $ ::setLeftFrozen() .AND. !oXbp:DisplayIndex $ ::setRightFrozen()
      AAdd(aPickedColumns, ;
         {aGetlist[i,cGETLIST_CAPTION],aGetList[i,cGETLIST_SUBCLASS],i,oXbp:DisplayIndex,.f.,aGetList[i,bGETLIST_EVAL],Len(aPickedColumns)+1})
    ENDIF
  ENDIF
NEXT

ASort(aPickedColumns,,,{|a,b|a[4]<b[4]})
ASort(aAvailColumns,,,{|a,b|a[1]<b[1]})

@ 0,0 DCSTATIC OBJECT oStatic TYPE XBPSTATIC_TYPE_TEXT SIZE 98,29

::isColumnCacheAutomatic := .f.
aOptions := {aPickedColumns,aAvailColumns,0,0,90,26,,,,,,,,,,,,,,,,,,,,,,,,,,}

PickBrowse( GetList, @oStatic, aOptions, GetOptions, @::isColumnCacheAutomatic )

DCREAD GUI EXPRESS FIT ;
    MODAL TITLE 'Choose Column(s) to Add' ;
    TO lStatus NOAUTORESTORE SETFOCUS @oBrowse OWNER oOwner

IF Empty(aPickedColumns)
  lStatus := .f.
ENDIF

IF lStatus
  nLockedColumns := Len(::setLeftFrozen())
  aFrozenColumns := ::setLeftFrozen({})
  oGetList := DC_GetDataStore(aGetList)
  IF !lNextCol
    DC_PointerWait()
  ENDIF

  FOR i := 1 TO Len(aAvailColumns)
    oGetList:getListPointer := aAvailColumns[i,3]
    oXbp := aGetList[oGetList:getListPointer,oGETLIST_OBJECT]
    IF Valtype(oXbp) == 'O'
      nColumn := oXbp:displayIndex
      ::delColumn(nColumn)
      aGetList[oGetList:getListPointer,oGETLIST_OBJECT] := nil
    ENDIF
  NEXT
  lGoToLast := .f.
  FOR i := 1 TO Len(aPickedColumns)
    oGetList:getListPointer := aPickedColumns[i,3]
    bEval := aPickedColumns[i,7]
    IF Empty( oXbp := aGetList[oGetList:getListPointer,oGETLIST_OBJECT] )
      cSubClass := aPickedColumns[i,2]
      DEFAULT cSubClass := 'DC_XbpColumn()'
      oXbp := &cSubClass
      nColumn := i+nLockedColumns
      oXbp := oXbp:new( self, oGetList, nColumn )
      aGetList[oGetList:getListPointer,oGETLIST_OBJECT] := oXbp
      IF Valtype(bEval) == 'B'
        Eval(bEval,oXbp)
      ENDIF
    ELSE
      IF oXbp:index # i+nLockedColumns
        nColumn := i+nLockedColumns
        ::delColumn(oXbp:displayIndex)
        oXbp := oXbp:new( self, oGetList, nColumn )
        IF Valtype(bEval) == 'B'
          Eval(bEval,oXbp)
        ENDIF
      ENDIF
    ENDIF
    AAdd(aColumns,oXbp)
  NEXT
  ::setLeftFrozen(aFrozenColumns)
  IF !lNextCol
    DC_PointerArrow()
  ENDIF
  IF lGoToLast
    ::lastCol()
  ENDIF
  ::forceStable()
  ::refreshAll()
//  ::stabilizeColumns()
ENDIF
IF Empty(aColumns)
  lStatus := .f.
ENDIF

RETURN lStatus

* -------------

FUNCTION PickBrowse( GetList, oParent, aOptions, GetOptions, lColumnCacheAutomatic )

LOCAL nRow, nCol, nWidth, nHeight, oObject, aSizePixels, oList, oPick, ;
      i, aPres, nPointerList := 1, nPointerPick := 1, ;
      nCol1, nCol2, aList, aPick, lPixel, oMenu, nFound, cKey

nRow := aOptions[3]
nCol := aOptions[4]
nWidth := aOptions[5]
nHeight := aOptions[6]
lPixel := aOptions[15]
aPres := aOptions[22]
aPick := aOptions[1]
aList := aOptions[2]

DEFAULT lPixel := .f.

IF lPixel
  nRow := 3
  nCol1 := 3
  nCol2 := (nWidth/2)
  nWidth := (nWidth/2) - 6
  nHeight := nHeight - 6
ELSE
  nRow := 3/20
  nCol1 := 3/7
  nCol2 := (nWidth/2) + 8
  nWidth := (nWidth/2) - (6/7)
  nHeight := nHeight - (6/20)
ENDIF

@ nRow, nCol1 DCBROWSE oList SIZE nWidth, nHeight ;
   DATA aList ;
   FONT '10.Lucida Console' ;
   PRESENTATION aPres ;
   NOHSCROLL ;
   RESIZE DCGUI_RESIZE_RESIZEONLY_Y ;
   PARENT oParent ;
   EVAL {|o|o:keyboard := {|n,b,o|PickBrowseKeyboard(n,o)}} ;
   POINTER nPointerList ;
   ITEMMARKED {||DC_GetWhen(GetList)} ;
   COLOR {|x|x:=DC_GetColArray(5,oList),IIF(Empty(x),{nil,GRA_CLR_WHITE},{GRA_CLR_WHITE,GRA_CLR_BLUE})} ;
   ITEMSELECTED {||ListSelected(aList,aPick,oList,oPick,nPointerList,nPointerPick), ;
                   DC_GetWhen(GetList)}

DCBROWSECOL ELEMENT 1 WIDTH nWidth PARENT oList ;
   HEADER 'Available Columns' ;
   DRAG {|o|o} ;
     DIALOG {|mp1,mp2,oXbp|DragDialog(mp1,mp2,oXbp)} ;
   DROP {|a,b,c|Drop(a,b,c,oList,oPick)} CURSOR POINTER_DATADROP_MOVE

@ nRow, nCol2 DCBROWSE oPick SIZE nWidth, nHeight ;
   DATA aPick ;
   PRESENTATION aPres ;
   FONT '10.Lucida Console' ;
   NOHSCROLL ;
   RESIZE DCGUI_RESIZE_RESIZEONLY_Y ;
   PARENT oParent ;
   EVAL {|o|o:keyboard := {|n,b,o|PickBrowseKeyboard(n,o)}} ;
   RBSELECT ;
   POINTER nPointerPick ;
   ITEMMARKED {||DC_GetWhen(GetList)} ;
   COLOR {|x|x:=DC_GetColArray(5,oPick),IIF(Empty(x),{nil,GRA_CLR_WHITE},{GRA_CLR_WHITE,GRA_CLR_BLUE})} ;
   ITEMSELECTED {||PickSelected(aList,aPick,oList,oPick,nPointerList,nPointerPick), ;
                   DC_GetWhen(GetList)}

DCBROWSECOL ELEMENT 1 WIDTH nWidth + 10 PARENT oPick ;
   HEADER 'Selected Columns' ;
   DRAG {|o|o} ;
     DIALOG {|mp1,mp2,oXbp|DragDialog(mp1,mp2,oXbp)} ;
     DROP {|a,b,c|Drop(a,b,c,oPick,oList)} CURSOR POINTER_DATADROP_MOVE

nRow += .5

@ nRow + nHeight, nCol1 DCPUSHBUTTON CAPTION 'Up' SIZE 7,1 ;
           ACTION {|a|a := aList[nPointerList], ;
                      ARemove(aList,nPointerList), ;
                      ASize(aList,Len(aList)+1), ;
                      AIns(aList,nPointerList-1,a), ;
                      oList:up(), ;
                      oList:forceStable(), ;
                      oList:refreshAll(),;
                      DC_GetWhen(GetList)} ;
           TOOLTIP 'Move this item up' ;
           WHEN {||nPointerList > 1} ;
           PARENT oParent ;
           FONT '10.Lucida Console'

@ DCGUI_ROW, DCGUI_COL + 10 DCPUSHBUTTON CAPTION 'Dn' SIZE 7,1 ;
           ACTION {|a|a := aList[nPointerList], ;
                      ARemove(aList,nPointerList), ;
                      ASize(aList,Len(aList)+1), ;
                      AIns(aList,nPointerList+1,a), ;
                      oList:down(), ;
                      oList:forceStable(), ;
                      oList:refreshAll(), ;
                      DC_GetWhen(GetList)} ;
           TOOLTIP 'Move this item Down' ;
           WHEN {||nPointerList < Len(aList)} ;
           PARENT oParent ;
           FONT '10.Lucida Console'

@ DCGUI_ROW, DCGUI_COL DCPUSHBUTTON CAPTION 'Sort' SIZE 7,1 ;
    ACTION {||ASort(aList,,,{|a,b|a[1]<b[1]}), ;
              oList:refreshAll()} ;
    PARENT oParent ;
    FONT '10.Lucida Console'

@ DCGUI_ROW, DCGUI_COL + 30 DCSAY 'Use mouse to drag fields' ;
   SAYSIZE 0 COLOR GRA_CLR_BLUE FONT '10.Arial Bold' PARENT oParent

@ nRow + nHeight, nCol2 DCPUSHBUTTON CAPTION 'Up' SIZE 7,1 ;
           ACTION {|a|a := aPick[nPointerPick], ;
                      ARemove(aPick,nPointerPick), ;
                      ASize(aPick,Len(aPick)+1), ;
                      AIns(aPick,nPointerPick-1,a), ;
                      oPick:up(), ;
                      oPick:forceStable(), ;
                      oPick:refreshAll(), ;
                      DC_GetWhen(GetList)} ;
           TOOLTIP 'Move this item up' ;
           WHEN {||nPointerPick > 1} ;
           PARENT oParent ;
           FONT '10.Lucida Console'

@ DCGUI_ROW, DCGUI_COL + 10 DCPUSHBUTTON CAPTION 'Dn' SIZE 7,1 ;
           ACTION {|a|a := aPick[nPointerPick], ;
                      ARemove(aPick,nPointerPick), ;
                      ASize(aPick,Len(aPick)+1), ;
                      AIns(aPick,nPointerPick+1,a), ;
                      oPick:down(), ;
                      oPick:forceStable(), ;
                      oPick:refreshAll(), ;
                      DC_GetWhen(GetList)} ;
           TOOLTIP 'Move this item Down' ;
           WHEN {||nPointerPick < Len(aPick)} ;
           PARENT oParent ;
           FONT '10.Lucida Console'

@ nRow + nHeight + 1, 2 DCCHECKBOX lColumnCacheAutomatic ;
   PROMPT 'Enable Automatic Cache Column' PARENT oParent  ;
   TOOLTIP 'Columns added automatically with right arrow at end of browse'

@ nRow + nHeight + 3, 36 DCPUSHBUTTON CAPTION 'Ok' SIZE 9, 1.2  ;
   ACTION {||DC_ReadGuiEvent(DCGUI_EXIT_OK,GetList)}

@ DCGUI_ROW, DCGUI_COL + 10 DCPUSHBUTTON CAPTION 'Cancel' SIZE 9, 1.2  ;
   ACTION {||DC_ReadGuiEvent(DCGUI_EXIT_ABORT,GetList)}

nRow := 9
nCol2 -= 7.8

@ nRow + 0, nCol2 DCPUSHBUTTON CAPTION '>' SIZE 7,1 ;
    ACTION {||ListSelected(aList,aPick,oList,oPick,nPointerList,nPointerPick)} ;
    PARENT oParent ;
    FONT '10.Lucida Console'

@ nRow + 2, nCol2 DCPUSHBUTTON CAPTION '>>' SIZE 7,1 ;
    ACTION {||ListSelected(aList,aPick,oList,oPick,-1,nPointerPick)} ;
    PARENT oParent ;
    FONT '10.Lucida Console'

@ nRow + 4, nCol2 DCPUSHBUTTON CAPTION '<<' SIZE 7,1 ;
    ACTION {||PickSelected(aList,aPick,oList,oPick,nPointerList,-1)} ;
    PARENT oParent ;
    FONT '10.Lucida Console'

@ nRow + 6, nCol2 DCPUSHBUTTON CAPTION '<' SIZE 7,1 ;
    ACTION {||PickSelected(aList,aPick,oList,oPick,nPointerList,nPointerPick)} ;
    PARENT oParent ;
    FONT '10.Lucida Console'

RETURN oObject

* -------------

STATIC FUNCTION Drop( oDrop, oColumn, mp1, oBrowse1, oBrowse2 )

LOCAL nOffset, aData1, aData2, nElement1, nElement2, lSame := .f., aValue

IF oDrop:parent == oColumn:parent // dragging within same browse
  oBrowse2 := oBrowse1
  lSame := .t.
ENDIF

aData1 := oBrowse1:dataSource
aData2 := oBrowse2:dataSource

nOffset := oDrop:dataArea:CellFromPos(mp1) - oBrowse1:RowPos
nElement1 := oBrowse1:arrayElement+nOffset
nElement2 := oBrowse2:arrayElement

IF nElement1 > Len(aData1)
  AAdd(aData1,aData2[nElement2])
  IF lSame
    ARemove( aData2, nElement2 )
  ENDIF
ELSEIF lSame
  aValue := aData1[nElement2]
  ARemove( aData2, nElement2 )
  ASize( aData1,Len(aData1)+1 )
  AIns(aData1,nElement1,aValue)
ELSE
  ASize(aData1,Len(aData1)+1)
  AIns(aData1,nElement1,aData2[nElement2])
ENDIF

oBrowse1:refreshAll()
IF !lSame
  ARemove( aData2, nElement2 )
  oBrowse2:refreshAll()
ENDIF

RETURN nil

* -------------

STATIC FUNCTION lbUp( aArray, nPointer, oBrowse )

IF AppKeyState(xbeK_CTRL) == 1
  aArray[nPointer,2] := !aArray[nPointer,2]
  oBrowse:refreshCurrent()
ENDIF

RETURN nil

* -------------

STATIC FUNCTION ListSelected( aList, aPick, oList, oPick, nPointerList, nPointerPick )

LOCAL i, bError, nTagElement

IF nPointerPick == 0
  nPointerPick := 1
ENDIF

IF nPointerList == 0 .OR. Empty(aList)
  RETURN nil
ENDIF

nTagelement := 5 // Len(aList[1])
IF nPointerList > 0
  aList[nPointerList,nTagElement] := .T.
ENDIF
FOR i := 1 TO Len(aList)
  IF nPointerList == -1
    aList[i,nTagElement] := .T.
  ENDIF
  IF aList[i,nTagElement]
    aList[i,nTagElement] := .F.
    IF .t. // Empty(aPick)
      AAdd( aPick, aList[i])
    ELSE
      ASize(aPick,Len(aPick)+1)
      AIns(aPick,nPointerPick,aList[i])
    ENDIF
    ARemove(aList,i)
    i--
  ENDIF
NEXT
oList:refreshAll()
oPick:refreshAll()

RETURN nil

* -------------

STATIC FUNCTION PickSelected( aList, aPick, oList, oPick, nPointerList, nPointerPick )

LOCAL i, nTagElement

IF nPointerList == 0
  nPointerList := 1
ENDIF

IF nPointerPick == 0 .OR. Empty(aPick)
  RETURN nil
ENDIF

nTagElement := 5 // Len(aPick[1])

IF nPointerPick > 0
  aPick[nPointerPick,nTagElement] := .T.
ENDIF
FOR i := 1 TO Len(aPick)
  IF nPointerPick == -1
    aPick[i,nTagElement] := .T.
  ENDIF
  IF aPick[i,nTagElement]
    aPick[i,nTagElement] := .F.
    IF Empty(aList)
      AAdd( aList, aPick[i])
    ELSE
      ASize(aList,Len(aList)+1)
      AIns(aList,nPointerList,aPick[i])
    ENDIF
    ARemove(aPick,i)
    i--
  ENDIF
NEXT

oList:refreshAll()
oPick:refreshAll()

RETURN nil

* -----------

STATIC FUNCTION DragDialog( mp1, mp2, oXbp )

LOCAL GetList[0], GetOptions, oDialog, aPos, cText

IF oXbp:isDerivedFrom('XbpColumn')
  cText := Trim(Eval(oXbp:dataLink))
ELSEIF oXbp:isDerivedFrom('XbpStatic')
  cText := Trim(oXbp:caption)
ELSEIF oXbp:isDerivedFrom('XbpTreeView')
  cText := Trim(oXbp:getData():caption)
ENDIF

@ 2,0 DCSTATIC TYPE XBPSTATIC_TYPE_BITMAP SIZE 10,1.2 ;
   CAPTION "..\..\bitmaps\user_female.Bmp"

@ DCGUI_ROW, DCGUI_COL DCSAY cText ;
   COLOR GRA_CLR_BLACK, GRA_CLR_YELLOW ;
   FONT '10.Courier' SAYSIZE 0,1.2 ;
   OPTION XBPSTATIC_TEXT_LEFT + XBPSTATIC_TEXT_VCENTER

DCGETOPTIONS NOTITLEBAR ORIGIN XBPDLG_ORIGIN_MOUSE ;
   FITPAD 0 COLOR { 240, 240, 255 } ;
   BORDER XBPDLG_NO_BORDER ;
   HIDE ;
   NOTASKLIST

DCREAD GUI FIT OPTIONS GetOptions PARENT @oDialog EXIT NOAUTORESTORE

aPos := DC_CalcAbsolutePosition(mp1,oXbp)
aPos[2] -= (oDialog:currentSize()[2] + 30)
oDialog:setPos( aPos )
oDialog:show()

RETURN oDialog

* -------------

STATIC FUNCTION PickBrowseKeyboard( nKey, oBrowse )

LOCAL cKey, nFound, aData

cKey := Upper(Chr(nKey))
IF !(cKey $ 'ABCDEFGHIJKLMNOPQRSTUVWXYZ')
  RETURN .f.
ENDIF

aData := oBrowse:dataSource

nFound := AScan(aData,{|a|Left(a[1],1)==cKey},oBrowse:arrayElement+1)

IF nFound > 0
  oBrowse:arrayElement := nFound
ELSE
  nFound := AScan(aData,{|a|Left(a[1],1)==cKey},1)
  IF nFound > 0
    oBrowse:arrayElement := nFound
  ELSE
    oBrowse:arrayElement := 1
  ENDIF
ENDIF

oBrowse:refreshAll()

RETURN .t.

* -------------

METHOD DC_XbpBrowse:stabilizeColumns()

LOCAL i, oColumn

FOR i := 1 TO ::colCount
  oColumn := ::getColumn(i)
  IF oColumn:displayIndex # i
    oColumn:displayIndex := i
  ENDIF
  oColumn:columnPos := i
NEXT

RETURN self

* -------------

METHOD DC_XbpBrowse:CellEditSetup()

LOCAL aGetList, aGetListItem

aGetList := ::getList:getListArray
aGetListItem := ::getList:getListArray[::getListPointer]

DEFAULT aGetListItem[xGETLIST_OPTIONS2] := {}    // Cell Editing
//ASize(aGetListItem[xGETLIST_OPTIONS2],7)
ASize(aGetListItem[xGETLIST_OPTIONS2],8)                                      // J. Duijf 30-03-2013 Custom cell-edit keyblock
IF Valtype(aGetListItem[xGETLIST_OPTIONS2,1]) = 'N'
  aGetListItem[xGETLIST_OPTIONS2,1] := { aGetListItem[xGETLIST_OPTIONS2,1], nil }
ENDIF
DEFAULT aGetListItem[xGETLIST_OPTIONS2,1] := {0,nil}   // event
DEFAULT aGetListItem[xGETLIST_OPTIONS2,6] := ;
   ( aGetListItem[nGETLIST_SUBTYPE] == BROWSE_DATABASE )

IF aGetListItem[xGETLIST_OPTIONS2,1,1] > 0 // Edit Mode
  ::CellEditKeyBlock := aGetListItem[xGETLIST_OPTIONS2,8]                     // J. Duijf 30-03-2013
  ::cellEditMode := .t.

  ::getList:AddHotKey( aGetListItem[xGETLIST_OPTIONS2,1,1], ;
              DC_BrowCellBlock( ;
                self, ;
                aGetListItem[xGETLIST_OPTIONS2,3],;     // Edit mode
                aGetListItem[xGETLIST_OPTIONS2,7], ;    // Default value (array editing)
                aGetListItem[xGETLIST_OPTIONS2,2],;     // Edit Action block
                aGetListItem[xGETLIST_OPTIONS8], ;      // Handler block
                aGetListItem[xGETLIST_OPTIONS2,5],;     // Exit block
                aGetListItem[xGETLIST_OPTIONS9],;       // Reference Array
                aGetList),;
            nil, nil, self, self, nil, .t. )
ENDIF

DEFAULT aGetListItem[xGETLIST_OPTIONS3] := {}  // Cell Inserting
ASize(aGetListItem[xGETLIST_OPTIONS3],5)

DEFAULT aGetListItem[xGETLIST_OPTIONS3,1] := 0  // Insert Mode
IF aGetListItem[xGETLIST_OPTIONS3,1] > 0
  ::getList:AddHotKey( aGetListItem[xGETLIST_OPTIONS3,1], ;
               DC_BrowCellBlock( ;
                 self, ;
                 DCGUI_BROWSE_INSERT,;                   // Insert mode
                 aGetListItem[xGETLIST_OPTIONS3,3], ;    // Default value (array editing)
                 aGetListItem[xGETLIST_OPTIONS3,2], ;    // Action Block
                 aGetListItem[xGETLIST_OPTIONS8],   ;    // Handler Block
                 aGetListItem[xGETLIST_OPTIONS3,4], ;    // Exit Block
                 aGetListItem[xGETLIST_OPTIONS9],   ;    // Reference Array
                 aGetList), ;
            nil, nil, self, self, nil, .T.  )
  ::getList:AddHotkey( aGetListItem[xGETLIST_OPTIONS3,1], ;
                       {||SetAppFocus(self), ;
                          PostAppEvent(xbeP_Keyboard,aGetListItem[xGETLIST_OPTIONS3,1],,self)}, ;
                       {|o|o:=SetAppFocus(), ;
                           o:isDerivedFrom('XbpCellGroup') .AND. ;
                           o:parent:parent==self} )
ENDIF

DEFAULT aGetListItem[xGETLIST_OPTIONS3,5] := 0  // Append Mode
IF aGetListItem[xGETLIST_OPTIONS3,5] > 0
  ::getList:AddHotKey( aGetListItem[xGETLIST_OPTIONS3,5], ;
               DC_BrowCellBlock( ;
                 self, ;
                 DCGUI_BROWSE_APPEND,;                   // Append mode
                 aGetListItem[xGETLIST_OPTIONS3,3], ;    // Default value (array editing)
                 aGetListItem[xGETLIST_OPTIONS3,2], ;    // Action Block
                 aGetListItem[xGETLIST_OPTIONS8],   ;    // Handler Block
                 aGetListItem[xGETLIST_OPTIONS3,4], ;    // Exit Block
                 aGetListItem[xGETLIST_OPTIONS9],   ;    // Reference Array
                 aGetList ), ;
            nil, nil, self, self, nil, .T.  )

  ::getList:AddHotkey( aGetListItem[xGETLIST_OPTIONS3,5], ;
                       {||SetAppFocus(self), ;
                          PostAppEvent(xbeP_Keyboard,aGetListItem[xGETLIST_OPTIONS3,5],,self)}, ;
                       {|o|o:=SetAppFocus(), ;
                           o:isDerivedFrom('XbpCellGroup') .AND. ;
                           o:parent:parent==self} )
ENDIF

DEFAULT aGetListItem[xGETLIST_OPTIONS4] := {}  // Cell Deleting
ASize(aGetListItem[xGETLIST_OPTIONS4],3)
DEFAULT aGetListItem[xGETLIST_OPTIONS4,1] := 0

IF aGetListItem[xGETLIST_OPTIONS4,1] > 0  // Delete Mode
  ::getList:AddHotKey( aGetListItem[xGETLIST_OPTIONS4,1], ;
               DC_BrowCellBlock( ;
                 self, ;
                 DCGUI_BROWSE_DELETE,;                 // Delete mode
                 nil,;                                 // nil
                 aGetListItem[xGETLIST_OPTIONS4,2], ;  // Action block
                 aGetListItem[xGETLIST_OPTIONS8],;     // Handler block
                 aGetListItem[xGETLIST_OPTIONS4,3],;   // Exit block
                 aGetListItem[xGETLIST_OPTIONS9],;     // Reference Array
                 aGetList ), ;
            nil, nil, self, self, nil, .T. )

  ::getList:AddHotkey( aGetListItem[xGETLIST_OPTIONS4,1], ;
                       {||SetAppFocus(self), ;
                          PostAppEvent(xbeP_Keyboard,aGetListItem[xGETLIST_OPTIONS4,1],,self)}, ;
                       {|o|o:=SetAppFocus(), ;
                           o:isDerivedFrom('XbpCellGroup') .AND. ;
                           o:parent:parent==self} )
ENDIF

IF Valtype(aGetListItem[xGETLIST_OPTIONS2,1,2]) = 'B'

  ::keyBoard := {|a,b,o|DC_BrowCellEdit(self, ;
                        nil, ;
                        nil, ;
                        aGetListItem[xGETLIST_OPTIONS2,3], ;
                        nil, ;
                        aGetListItem[xGETLIST_OPTIONS2,7], ;
                        aGetListItem[xGETLIST_OPTIONS2,2], ;
                        self, ;
                        aGetListItem[xGETLIST_OPTIONS8], ;
                        aGetListItem[xGETLIST_OPTIONS2,5], ;
                        aGetListItem[xGETLIST_OPTIONS9], ;
                        aGetList, ;
                        .T., ;
                        Eval(aGetListItem[xGETLIST_OPTIONS2,1,2],a)) }
ENDIF

RETURN nil

* --------------------

METHOD DC_XbpBrowse:MoveColumn( nColumns, nPos, lRefresh )

LOCAL oColumn, i, lMoveLeft, bLeftButton, aSize, bRightButton, nNewPos, nSaveColorBG, ;
      bRightDataArea, aHeadingSize, aVisualStyle[3], aDrawMode[3], aCargo

DEFAULT nColumns := 0, ;
        lRefresh := .t., ;
        nPos := ::colPos

lMoveLeft := (nPos<::colCount)
IF nColumns > 0 .AND. nColumns + nPos > ::colCount
  nColumns := ::colCount - nPos
ELSEIF nColumns < 0 .AND. -1*nColumns > nPos
  nColumns := -1*nPos
ENDIF
IF nColumns == 0 .OR. ( nPos=1 .AND. nColumns<0 ) .OR. ;
  ( nPos==::colCount .AND. nColumns > 0 )
  RETURN self
ENDIF

IF nPos == ::sortColumn
  ::sortColumn := nPos + nColumns
ENDIF

::lockUpdate(.t.)
oColumn := ::getColumn(nPos)
nSaveColorBG := oColumn:heading:setColorBG()
bLeftButton := oColumn:heading:lbDown  // <<<<<<< save left sort block
bRightButton := oColumn:heading:rbDown  // <<<<<<< save right sort block
bRightDataArea := oColumn:dataArea:rbDown // <<<<<<< save right dataarea block
#if XPPVER > 1900331
aVisualStyle[1] := oColumn:heading:useVisualStyle
aVisualStyle[2] := oColumn:dataArea:useVisualStyle
aVisualStyle[3] := oColumn:footing:useVisualStyle
aDrawMode[1] := oColumn:heading:drawMode
aDrawMode[2] := oColumn:dataArea:drawMode
aDrawMode[3] := oColumn:footing:drawMode
#endif
aCargo := oColumn:cargo

aSize := oColumn:currentSize()
aHeadingSize := oColumn:heading:currentSize()
::delColumn(nPos)
::lockUpdate(.f.)
nNewPos := nPos+nColumns
::insColumn(nNewPos,oColumn)
oColumn:setHeaderCaption()
oColumn:setFooterCaption()

IF ::cursorMode == XBPBRW_CURSOR_CELL
  IF nColumns > 0
    FOR i := 1 TO nColumns
      ::right()
    NEXT
  ELSE
    nColumns := -1 * nColumns
    IF lMoveLeft
      FOR i := 1 TO nColumns
        ::left()
      NEXT
    ENDIF
  ENDIF
ENDIF

IF !Empty(bLeftButton) .AND. 'oColumn:Sort' $ DC_GetCodeBlockText(bLeftButton)
   oColumn:heading:lbDown := _HeaderDown( self, oColumn, nNewPos )
ENDIF
IF !Empty(bRightButton) .AND. 'oColumn:Sort' $ DC_GetCodeBlockText(bRightButton)
   oColumn:heading:rbDown := _HeaderDown( self, oColumn, nNewPos )
ENDIF
IF !Empty(bRightDataArea)
   oColumn:dataArea:rbDown := bRightDataArea
ENDIF

IF Valtype(nSaveColorBG) == 'N'
  oColumn:heading:setColorBG(nSaveColorBG)
ENDIF
oColumn:setSize(aSize)
oColumn:heading:setSize(aHeadingSize)
oColumn := ::getColumn(nPos)

IF !Empty(oColumn:heading:lbDown) .AND. 'oColumn:Sort' $ DC_GetCodeBlockText(oColumn:heading:lbDown)
   oColumn:heading:lbDown := _HeaderDown( self, oColumn, nPos )
ENDIF
IF !Empty(oColumn:heading:rbdown) .AND. 'oColumn:Sort' $ DC_GetCodeBlockText(oColumn:heading:rbDown)
   oColumn:heading:rbDown := _HeaderDown( self, oColumn, nPos )
ENDIF
#if XPPVER > 1900331
oColumn:heading:useVisualStyle := aVisualStyle[1]
oColumn:dataArea:useVisualStyle := aVisualStyle[2]
oColumn:footing:useVisualStyle := aVisualStyle[3]
oColumn:heading:drawMode := aDrawMode[1]
oColumn:dataArea:drawMode := aDrawMode[2]
oColumn:footing:drawMode := aDrawMode[3]
oColumn:cargo := aCargo
#endif

::setSize(::currentSize())
::forceStable()

IF lRefresh
  ::xbpBrowse:refreshAll()
ENDIF

RETURN self

* --------------------

METHOD DC_XbpBrowse:itemRbDown( aPos, aValues )

LOCAL nRow := aValues[2]
LOCAL nCol

IF ::rightButtonSelect
  #if XPPVER > 02000000
    // nCol := aValues[1]:columnPos
    nCol := aValues[1]:displayIndex
  #else
    nCol := ::ColNum ( aValues[1] )
  #endif
  ::xbpBrowse:gotoItem( nRow, nCol )
  ::colPos := nCol
ENDIF

::XbpBrowse:itemRbDown( aPos, aValues )

RETURN self

* --------------------

METHOD DC_Xbpbrowse:cellEdit()

RETURN self

* --------------------

METHOD DC_XbpBrowse:cellEditMode(lMode)

Grok(lMode)

::cellEditMode := !::cellEditMode

RETURN self

* ----------------------

METHOD DC_XbpBrowse:SortClear()

LOCAL i, j, oColumn, oXbp, aChildList

FOR i := 1 TO ::colCount
  oColumn := ::GetColumn(i)
  IF Valtype(::sortColorNonselectableFG) = 'N'
    oColumn:heading:setColorFG(::sortColorNonselectableFG)
  ENDIF
  IF Valtype(::sortColorNonselectableBG) = 'N'
    oColumn:heading:setColorBG(::sortColorNonselectableBG)
  ENDIF
  IF oColumn:heading:status() > 0 .AND. !Empty(oColumn:heading:childList())
    aChildList := oColumn:heading:childList()
    FOR j := 1 TO Len(aChildList)
      oXbp := aChildList[j]
      IF oXbp:isDerivedFrom('XbpStatic') .AND. Valtype(oXbp:cargo) = 'C' .AND. ;
         oXbp:cargo == 'SortArrow'
        // destroy the arrow
        oXbp:destroy()
      ENDIF
    NEXT
  ENDIF
NEXT

RETURN self

* -----------------------

METHOD DC_XbpBrowse:HFLbDown ( aPos, aValues )

LOCAL i, aColumns[::colCount], nCol
FOR i := 1 TO Len(aColumns)
  aColumns[i] := ::getColumn(i)
  IF aColumns[i] == aValues[1]
    nCol := i
  ENDIF
NEXT

IF nCol > 1 .AND. aColumns[nCol-1]:noResize .AND. aPos[1] < 5
  RETURN nil
ELSEIF aColumns[nCol]:noResize .AND. aPos[1] > 5
  RETURN nil
ELSEIF aColumns[nCol]:noResize .AND. nCol < Len(aColumns) .AND. !aColumns[nCol+1]:noResize
  RETURN nil
ENDIF

RETURN ::XbpBrowse:HFLbDown( aPos, aValues )

* ---------

METHOD DC_XbpBrowse:Tag( oCell, mp1, nKey )

LOCAL nCurrPos, nSaveRec, i, nKeyNo, nElement, nOffset, nSeconds := Seconds(), ;
      lUpdate := .t.

DEFAULT nKey := 0

IF nKey > 0 .AND. !(nKey $ { xbeK_CTRL_DOWN, xbeK_CTRL_UP, xbeK_SPACE, xbeK_CTRL_INS, xbeK_CTRL_DEL, xbeK_SH_DOWN })
  RETURN nil
ENDIF

IF nKey == xbeK_CTRL_DEL
  IF Valtype(::dataSource) == 'A' .AND. ::arrayElement > 0 .AND. ::arrayElement <= Len(::dataSource) .AND. ;
      Len(::dataSource[::arrayElement]) >= ::tagElement .AND. Valtype(::dataSource[::arrayElement,::tagElement]) == 'L'
    FOR i := 1 TO Len(::dataSource)
      ::dataSource[i,::tagElement] := .f.
    NEXT
  ELSEIF Valtype(::dataSource) == 'C'
    (::dataSource)->(DC_RecTagClear())
  ENDIF
  ::refreshAll()
ELSEIF nKey == xbeK_CTRL_INS
  IF Valtype(::dataSource) == 'A' .AND. ::arrayElement > 0 .AND. ::arrayElement <= Len(::dataSource) .AND. ;
      Len(::dataSource[::arrayElement]) >= ::tagElement .AND. Valtype(::dataSource[::arrayElement,::tagElement]) == 'L'
    FOR i := 1 TO Len(::dataSource)
      ::dataSource[i,::tagElement] := .t.
    NEXT
  ELSEIF Valtype(::dataSource) == 'C'
    nSaveRec := (::dataSource)->(RecNo())
    (::dataSource)->(DC_RecTagClear())
    (::dataSource)->(DC_DbGoTop())
    DO WHILE !(::dataSource)->(DC_Eof())
      (::dataSource)->(DC_RecTagToggle())
      (::dataSource)->(DC_DbSkip())
    ENDDO
    (::dataSource)->(dbGoTo(nSaveRec))
  ENDIF
  ::refreshAll()
ELSEIF AppKeyState(xbeK_CTRL) == 1 .OR. nKey $ { xbeK_CTRL_DOWN, xbeK_CTRL_UP, xbeK_SPACE }
  IF Valtype(oCell) == 'O'
    IF ::isLeftMouseButton
      nOffset := oCell:CellFromPos(mp1) - ::RowPos
      nElement := ::arrayElement + nOffset
      IF Valtype(::dataSource) == 'A' .AND. nElement > 0 .AND. nElement <= Len(::dataSource) .AND. ;
        Len(::dataSource[nElement]) >= ::tagElement .AND. Valtype(::dataSource[nElement,::tagElement]) == 'L'
        IF !::dataSource[nElement,::tagElement]
          ::dataSource[nElement,::tagElement] := .t.
          ::lastTagRec := nElement
          ::refreshAll()
        ENDIF
      // ELSEIF Valtype(::dataSource) == 'C'
      //   (::dataSource)->(DC_RecTagToggle())
      //   ::LastTagRec := (::dataSource)->(RecNo())
      ENDIF
    ENDIF
    RETURN nil
  ELSE
    IF Valtype(::dataSource) == 'A' .AND. ::arrayElement > 0 .AND. ::arrayElement <= Len(::dataSource) .AND. ;
      Len(::dataSource[::arrayElement]) >= ::tagElement .AND. Valtype(::dataSource[::arrayElement,::tagElement]) == 'L'
      ::dataSource[::arrayElement,::tagElement] := ;
         !::dataSource[::arrayElement,::tagElement]
      ::lastTagRec := ::arrayElement
    ELSEIF Valtype(::dataSource) == 'C'
      (::dataSource)->(DC_RecTagToggle())
      ::LastTagRec := (::dataSource)->(RecNo())
    ENDIF
    ::refreshCurrent()
    IF nKey == xbeK_CTRL_DOWN
     // PostAppEvent( xbeP_Keyboard, xbeK_DOWN, nil, self )
    ELSEIF nKey == xbeK_CTRL_UP
     // PostAppEvent( xbeP_Keyboard, xbeK_UP, nil, self )
    ENDIF
  ENDIF
  RETURN nil
ELSEIF AppKeyState(xbeK_SHIFT) == 1
  IF ::lastTagRec == 0
    IF Valtype(::dataSource) == 'A' .AND. ::arrayElement > 0 .AND. ::arrayElement <= Len(::dataSource) .AND. ;
      Len(::dataSource[::arrayElement]) >= ::tagElement .AND. Valtype(::dataSource[::arrayElement,::tagElement]) == 'L'
      ::dataSource[::arrayElement,::tagElement] := ;
         !::dataSource[::arrayElement,::tagElement]
    ELSEIF Valtype(::dataSource) == 'C'
      (::dataSource)->(DC_RecTagToggle())
    ENDIF
    ::refreshCurrent()
  ELSE
    IF Valtype(::dataSource) == 'A' .AND. ::arrayElement > 0 .AND. ::arrayElement <= Len(::dataSource) .AND. ;
      Len(::dataSource[::arrayElement]) >= ::tagElement .AND. Valtype(::dataSource[::arrayElement,::tagElement]) == 'L'
      nCurrPos := ::arrayElement
      FOR i := 1 TO Len(::dataSource)
        ::dataSource[i,::tagElement] := .f.
      NEXT
      IF nCurrPos > ::LastTagRec
        FOR i := ::LastTagRec TO nCurrPos
           ::dataSource[i,::tagElement] := .t.
        NEXT
      ELSEIF nCurrPos < ::LastTagRec
        FOR i := nCurrPos TO ::LastTagRec
           ::dataSource[i,::tagElement] := .t.
        NEXT
      ENDIF
    ELSEIF Valtype(::dataSource) == 'C'
      DC_PointerWait()
      nSaveRec := (::dataSource)->(RecNo())
      nCurrPos := (::dataSource)->(DC_KeyNo())
      (::dataSource)->(DC_RecTagClear())
      IF nCurrPos > ::LastTagPos
        (::dataSource)->(DC_dbGoto(::LastTagRec))
        IF ::tagMode == DCGUI_TAGMODE_SET
          (::dataSource)->(DC_DbSkip())
        ENDIF
        DO WHILE (::dataSource)->(DC_KeyNo()) <= nCurrPos .AND. !(::dataSource)->(DC_Eof())
          (::dataSource)->(DC_RecTagToggle())
          (::dataSource)->(DC_DbSkip())
        ENDDO
      ELSEIF nCurrPos < ::LastTagPos
        (::dataSource)->(DC_dbGoto(::LastTagRec))
        IF ::tagMode == DCGUI_TAGMODE_SET
          (::dataSource)->(DC_DbSkip(-1))
        ENDIF
        DO WHILE (nKeyNo := (::dataSource)->(DC_KeyNo())) >= nCurrPos .AND. !(::dataSource)->(DC_Bof())
          (::dataSource)->(DC_RecTagToggle())
          (::dataSource)->(DC_DbSkip(-1))
          IF nKeyNo <= 1
            EXIT
          ENDIF
        ENDDO
      ENDIF
      DC_PointerArrow()
      (::dataSource)->(DC_dbGoto(nSaveRec))
    ENDIF
    ::refreshAll()
  ENDIF
  RETURN nil
ELSEIF Empty(oCell)
  IF Valtype(::dataSource) == 'C' .AND. !Empty((::dataSource)->(DC_RecTagArray()))
    (::dataSource)->(DC_RecTagClear())
    IF ::tagMode == DCGUI_TAGMODE_SET
      (::dataSource)->(DC_RecTagToggle())
    ELSE
      lUpdate := .f.
    ENDIF
  ELSEIF Valtype(::dataSource) = 'A' .AND. !Empty(::dataSource)
    FOR i := 1 TO Len(::dataSource)
      ::dataSource[i,::tagElement] := .f.
    NEXT
    IF ::tagMode == DCGUI_TAGMODE_SET .AND. ::arrayElement > 0 .AND. ::arrayElement <= Len(::dataSource) .AND. ;
      Len(::dataSource[::arrayElement]) >= ::tagElement .AND. Valtype(::dataSource[::arrayElement,::tagElement]) == 'L'
      ::dataSource[::arrayElement,::tagElement] := .t.
    ENDIF
  ELSE
    RETURN nil
  ENDIF
  ::refreshAll()
ELSE
  RETURN nil

ENDIF

IF lUpdate
  IF Valtype(::dataSource) == 'C'
    ::LastTagRec := (::dataSource)->(RecNo())
    ::LastTagPos := (::dataSource)->(DC_KeyNo())
  ELSEIF Valtype(::dataSource) == 'A'
    ::LastTagRec := ::arrayElement
    ::LastTagPos := ::arrayElement
  ENDIF
ENDIF

RETURN nil

//
// J. Duijf  21-04-2010
// Zebra stripes utilisation
//
METHOD DC_XbpBrowse:ZebraColor(uEven,aOdd,aTagged,aTaggedHilite)              // J. Duijf 02-08-2018
// METHOD DC_XbpBrowse:ZebraColor(uEven,aOdd)                                 // J. Duijf 02-08-2018
//
// J. Duijf 11-03-2011 Get default colors from presentation parameters
//
LOCAL aRet           := nil
//
Do Case
   Case ::tagEnable // .and. ::GetSetTag()                                       // Custom color for taged rows
      // J. Duijf 02-08-2018
      If ValType(aTagged) <> "A"
         aTagged     := {::tagColorFG,::tagColorBG}
      Endif
      //
      // And the color for SELECTED taged rows
      //
      aRet           := aTagged
      If ValType(aTaggedHilite) = "A" .and. Len(aTaggedHilite) >= 2
         Aadd(aRet,aTaggedHilite[1])
         Aadd(aRet,aTaggedHilite[2])
      Endif
      //
   case ValType(uEven) = "A"                                            // Array with foreground / background color
      if ::evenRow                                                      // Rows 2,4,6 etc
         aRet := uEven
      else                                                              // Rows 1,3,5 etc
         aRet := aOdd
      endif
   case ValType(uEven) = "B"                                            // User Codeblock
      aRet := Eval(uEven,::EvenRow)
   case ValType(aOdd) = "A"                                             // Array with foreground / background color
      if ::evenRow                                                      // Rows 2,4,6 etc
         aRet := aOdd
      endif
endcase

if ValType(aRet) <> "A"
   aRet              := {DC_GetPresentation(::browsePres,XBP_PP_COL_DA_FGCLR,GRA_CLR_BLACK),;
                         DC_GetPresentation(::browsePres,XBP_PP_COL_DA_BGCLR,GRA_CLR_WHITE)}
   //
   // J. Duijf 11-03-2011
   // Workaround for PDR 6365
   // The colorBlock should return a 4 element array to prevent an incorrect
   // Hilite forground color.
   // {CellFGColor,CellBGColor,CellHiliteFGColor,CellHiliteBGColor}
   // The highlite colors come from the presentation parameters.
   // If undefined, the default BG/FG hilite colors are BLUE/WHITE
   //
   Aadd(aRet,DC_GetPresentation(::browsePres,XBP_PP_COL_DA_HILITE_FGCLR,GRA_CLR_WHITE))
   Aadd(aRet,DC_GetPresentation(::browsePres,XBP_PP_COL_DA_HILITE_BGCLR,GRA_CLR_BLUE))
else
   if Len(aRet) = 2
      //
      // J. Duijf 11-03-2011
      // Workaround for PDR 6365
      // The colorBlock should return a 4 element array to prevent an incorrect
      // Hilite forground color.
      // {CellFGColor,CellBGColor,CellHiliteFGColor,CellHiliteBGColor}
      // The highlite colors come from the presentation parameters.
      // If undefined, the default BG/FG hilite colors are BLUE/WHITE
      //
      Aadd(aRet,DC_GetPresentation(::browsePres,XBP_PP_COL_DA_HILITE_FGCLR,GRA_CLR_WHITE))
      Aadd(aRet,DC_GetPresentation(::browsePres,XBP_PP_COL_DA_HILITE_BGCLR,GRA_CLR_BLUE))
   endif
endif

::rowColor := aRet

Return aRet

* -----------

STATIC FUNCTION _RowColor( oBrowse, aOptions, aColor, xData, bColorBlock )

LOCAL bRowColor

/*
 IF Valtype( bColorBlock ) = 'B' ; RETURN bColorBlock
 ELSEIF Empty(aOptions[1]) .AND. Empty(aOptions[3]) // MARK and MARKEVAL clauses not used
*/
//
// J. Duijf 28-04-2011
// bColorBlock can already be assigned (In case of Zebra stripes), but the MARK and MKEVAL clauses stil need to be evaluated.
//
if Empty(aOptions[1]) .AND. Empty(aOptions[3]) // MARK and MARKEVAL clauses not used
   IF Valtype( bColorBlock ) <> 'B'
      IF Valtype(DC_BrowseColor()) == 'A' .OR. oBrowse:tagEnable
        bRowColor := {||oBrowse:rowColor()}
      ELSE
        bRowColor := nil
      ENDIF
   else
      bRowColor   := bColorBlock
   ENDIF
ELSEIF Valtype(xData) = 'A'
/*
  bRowColor :=  { |x,a,n,f,g | a := oBrowse:dataSource, ;
                             n := oBrowse:arrayElement, ;
                             f := oBrowse:mark, ;
                             g := IIF(a[n,f], ;
                                  { aOptions[5], aOptions[6] }, aColor ) }
*/
//
// Modified 28-04-2011 J. Duijf,
// If the marked color clause is used, and a row is NOT marked, then use bColorBlock if available.
//

  bRowColor :=  { || _BrowseRowColor( oBrowse, aOptions, bColorBlock, aColor ) }

ELSE
/*
  bRowColor :=  { |x,a,n,f,g | a := oBrowse:dataSource, ;
                             n := oBrowse:arrayElement, ;
                             f := IIF(Valtype(oBrowse:markColorBlock)='B', ;
                                  oBrowse:markColorBlock,oBrowse:mark), ;
                             g := IIF(Eval(f), ;
                                  { aOptions[5], aOptions[6] }, aColor ) }
//
// Modified 3-1-2003, J. Duijf DCBROWSE does also support codeblocks as rowcolor
//                [MKCOLOR <nbMarkEval>, <ncMFgC> [,<ncMBgC>] ]               ;
//
// <ncMFgC> can now be a codeblock. It must return {FG,BG}
//
  bRowColor :=  { |x,a,n,f,g | a := oBrowse:dataSource, ;
                             n := oBrowse:arrayElement, ;
                             f := IIF(Valtype(oBrowse:markColorBlock)='B', ;
                                  oBrowse:markColorBlock,oBrowse:mark), ;
                             g := IIF(Eval(f), ;
                                  iif(ValType(aOptions[5]) ='B',Eval(aOptions[5]), ;
                                  { aOptions[5], aOptions[6] }), aColor ) }

 Modified 7-03-2010 J. Duijf DCBROWSE the codeblock for color receives the browse
                             object as first parameter, this for utilisation of zebra patern

  bRowColor :=  { |x,a,n,f,g | a := oBrowse:dataSource, ;
                             n := oBrowse:arrayElement, ;
                             f := IIF(Valtype(oBrowse:markColorBlock)='B', ;
                                  oBrowse:markColorBlock,oBrowse:mark), ;
                             g := IIF(Eval(f), ;
                                  iif(ValType(aOptions[5]) ='B',Eval(aOptions[5],oBrowse), ;
                                  { aOptions[5], aOptions[6] }), aColor ) }

*/
// Modified 28-04-2011 J. Duijf,
// If the MKCOLOR clause is used, and a row is NOT marked, then use bColorBlock if available.
//

  bRowColor :=  { |a,n,f,g | a := oBrowse:dataSource, ;
                             n := oBrowse:arrayElement, ;
                             f := IIF(Valtype(oBrowse:markColorBlock)='B', ;
                                  oBrowse:markColorBlock,oBrowse:mark), ;
                             g := IIF(Eval(f), ;
                                  iif(ValType(aOptions[5]) ='B',Eval(aOptions[5],oBrowse), ;
                                  { aOptions[5], aOptions[6] }), iif(ValType(bColorBlock) = "B",Eval(bColorBlock),aColor)) }

ENDIF

RETURN bRowColor

* ----------

STATIC FUNCTION _BrowseRowColor( oBrowse, aOptions, bColorBlock, aColor )

LOCAL a, nElement, f, g

a := oBrowse:dataSource
nElement := oBrowse:arrayElement
IF nElement == 0 .OR. Empty(a)
  RETURN nil
ENDIF

IF !Empty(oBrowse:markColorBlock)
  IF Eval(oBrowse:markColorBlock,nElement)
    aColor := { aOptions[5], aOptions[6] }
  ELSE
    IF Valtype(bColorBlock) = 'B'
      aColor := Eval(bColorBlock,nElement)
    ELSE
      aColor := aColor
    ENDIF
  ENDIF
ELSEIF !Empty(oBrowse:mark)
  IF oBrowse:dataSource[nElement,oBrowse:mark]
    IF Valtype(bColorBlock) = 'B'
      aColor := Eval(bColorBlock,nElement)
    ELSE
      aColor := aColor
    ENDIF
  ENDIF
ENDIF

RETURN aColor

* ===============================================

CLASS DC_MultiLineCellGroup FROM XbpCellGroup

EXPORTED:
  VAR width
  VAR presentation
  VAR rows
  VAR noResize

  INLINE METHOD Create( oParent, oOwner, aPos, aSize, aPP, lVisible )

  LOCAL i

  IF Len(::presentation) > 0
    IF Valtype(::presentation[2]) = 'A'
      FOR i := 1 TO Len(::presentation[2])
        DC_PresAdd(aPP,::presentation[2,i])
      NEXT
    ELSEIF Valtype(::presentation[1]) = 'A'
      FOR i := 1 TO Len(::presentation[1])
        DC_PresAdd(aPP,::presentation[1,i])
      NEXT
    ENDIF
  ENDIF

  IF ::noResize
    aSize := { ::width, -1 }
  ENDIF
  ::maxRow := ::rows
  if ::MaxRow > 1
     aSize[2] := -1
  endif
  IF IsMemberVar(self,'USEVISUALSTYLE')
    ::xbpCellGroup:useVisualStyle := .f.
  ENDIF

  Return ::XbpCellgroup:create( oParent, oOwner, aPos, aSize, aPP, lVisible )

ENDCLASS

* ===================================================

CLASS DC_XbpColumn FROM DC_XbpColumnSuper, DC_XbaseParts
HIDDEN:
VAR      nOffset                             // J. Duijf 7-03-2010
VAR      LastRow                             // J. Duijf 7-03-2010
EXPORTED:
  VAR validBlock
  VAR sortBlock
  VAR cellEditGetList
  VAR cellEditExitKey
  VAR getsetBlock
  VAR headerBlock
  VAR headerCaption
  VAR headerFont
  VAR headerHeight
  VAR dataSource
  VAR parent
  VAR picture
  VAR cellEditID
  VAR footerBlock
  VAR footerCaption
  VAR footerFont
  VAR footerHeight
  VAR pointer
  VAR presArray
  VAR columnPos
  VAR arrayPointer
  VAR headerColorBlock
  VAR footerColorBlock
  VAR sortDefault
  VAR cellEditor
  VAR noResize
  VAR dataLinkMultiLine
  VAR columnSize
  VAR noAutoResizeColumn
  VAR validateAlways
  VAR origFonts
  VAR visualStyleArray
  VAR drawModeArray
  VAR toolTip
  VAR filterDialog
  VAR varBlock
  VAR sqlVar
  VAR sqlSort
  VAR isTagged
  VAR isCollapsed
  VAR objectVarName
//  VAR displayIndex

  METHOD Init, setHeaderCaption, setFooterCaption, setHeaderColor, setFooterColor, ;
         MoveColumn, Refresh, WhenHideEval, Sort, resizeColumn, Destroy, ;
         BuildSortButton, TagColor, AddColumnFiltered(),  MoveColumnFiltered(), ;
         DeleteColumnFiltered()

  METHOD DrawRow, ScrollUp, ScrollDown       // J. Duijf 7-03-2010

ENDCLASS

* ---------------------

METHOD DC_XbpColumn:Destroy()
::DC_XbpColumnSuper:Destroy()
/*
::datasource      := nil
::parent          := nil
::Footing         := nil
::Heading         := nil
::Cargo           := nil
::datalink        := nil
::getlist         := nil
::xbp             := nil
*/
Return SELF

METHOD DC_XbpColumn:init( oParent, oGetList, nColumn )

LOCAL aHeader, cCaption, xOptions, aArray, aGetList, aGetListItem, ;
      nPointer, aFooter, aColor, lPixel, bBlock, bDataBlock, ;
      bColorBlock, xFont, aPres, lMultiline, ;
      nColorBG, aPP, nFound, nColorHiliteFG, nWidth
LOCAL bColorTmpBlock, aTmpColor

DEFAULT nColumn := 0

::visualStyleArray := Array(3)
::drawModeArray := Array(3)

::nOffset      := 0                          // J. Duijf 7-03-2010

::DC_XbaseParts:init( oGetList,,self )

aGetList := ::getList:getListArray
aGetListItem := ::getList:getListArray[::getListPointer]

DEFAULT aGetListItem[nGETLIST_SUBTYPE] := XBPCOL_TYPE_TEXT
DEFAULT aGetListItem[nGETLIST_WIDTH]   := XBP_AUTOSIZE

xOptions := aGetListItem[xGETLIST_OPTIONS]

aArray   := Aclone(oParent:browsePres)                                        // J. Duijf 10-03-2011
ASize(xOptions,17)

IF Valtype( _Pres(aGetListItem[aGETLIST_PRESENTATION]) ) = 'A'
   //
   // J. Duijf 10-03-2011
   // Instead of replacing the array, properties are merged.
   // aArray := AClone(_Pres(aGetListItem[aGETLIST_PRESENTATION]))
   //
   Dc_MergePresentation(aArray,_Pres(aGetListItem[aGETLIST_PRESENTATION]))
//ELSE
//  aArray := {}
//
ENDIF

::isTagged := .f.
::isCollapsed := .f.

DEFAULT xOptions[5] := xbeK_ENTER
DEFAULT xOptions[6] := .f.  // left button
DEFAULT xOptions[7] := .f.  // no create
DEFAULT xOptions[9] := .f.  // no resize
DEFAULT xOptions[11] := .f. // no resizecolumns
DEFAULT xOptions[12] := .f. // valid always
DEFAULT xOptions[13] := .f. // owner draw
DEFAULT xOptions[17] := ''  // index tag for sorting

::noResize := xOptions[9]
::noAutoResizeColumn := xOptions[11]
::validateAlways := xOptions[12]
::sqlField := xOptions[14]
::sqlSort := xOptions[15]
::indexTag := xOptions[17]

// ::displayIndex := 0

DEFAULT ::sqlSort := .f.

// get pointer to cell editor
::cellEditID := xOptions[4]
IF Valtype(::cellEditID) = 'C' .AND. !Empty(::cellEditID)  // Custom Editor ID
  nPointer := AScan(aGetList,{|a|a[cGETLIST_ID]==::cellEditID})
  IF nPointer > 0
    ::cellEditGetList := { AClone(aGetList[nPointer]) } // , aGetList[nDataStore] }
  ENDIF
ENDIF

// get pointer to array element
nPointer := aGetListItem[nGETLIST_POINTER]
::arrayPointer := nPointer

::toolTip := aGetListItem[cGETLIST_TOOLTIP]

IF Valtype( nPointer ) = 'N' // it's an array element pointer
  bDataBlock := DC_GetColBlock( nPointer, oParent )
  aGetListItem[bGETLIST_VAR] := bDataBlock
ELSE // it's a column code block
  bDataBlock := DC_GetBlock( aGetListItem[bGETLIST_VAR] )
  IF Valtype(bDataBlock) # 'B'
    bDataBlock := DC_GetBlock( aGetListItem[bGETLIST_REFVAR] )
  ENDIF
ENDIF

cCaption := aGetListItem[cGETLIST_CAPTION]
IF Valtype(cCaption) = 'B'
  ::headerBlock := cCaption
  // cCaption := Eval(cCaption)
  cCaption := Chr(255)
ENDIF

IF Valtype(oParent:sortButtonConfig) == 'O'
  ::headerCaption := ''
  aHeader := {Chr(255)}
ELSEIF Valtype(cCaption) = 'C'
  aHeader := DC_TokenArray( cCaption, oParent:headerDelim )
  ::headerCaption := cCaption
ELSEIF Valtype(cCaption) = 'A'
  IF Valtype(cCaption[1]) == 'C'
    aHeader := cCaption
  ELSE
    aHeader := {Chr(255)}
  ENDIF
ELSEIF oParent:hasHeader
  aHeader := {Chr(255)}
ELSE
  aHeader := {nil}
ENDIF

ASize(aHeader,oParent:headerLines)

/*
::resize := {|a,b,o,aChildren|aChildren := o:heading:childList(), ;
              IIF(!Empty(aChildren),;
              aChildren[1]:setPos({o:currentSize()[1]-20,aChildren[1]:currentPos()[2]}),nil)}
*/
//
// J. Duijf 19-02-2012
// If there is more then 1 header line, ignore the fixed header height
// It must be recalculated based on the fontsize (dynamic)
//

if Len(aHeader) > 0
   Dc_MergePresentation(aArray,{{XBP_PP_COL_HA_HEIGHT,-1}})
endif
//

cCaption := xOptions[3]

IF Valtype(cCaption) = 'B'
  cCaption := Eval(cCaption)
ENDIF

IF Valtype(cCaption) = 'C'
  aFooter := DC_TokenArray( cCaption, oParent:footerDelim )
  ::footerCaption := cCaption
//
// J. Duijf 18-09-2013
// Also blank footer line possible
//
  If Len(aFooter) = 0
     aFooter := {cCaption}
  Endif
ELSEIF Valtype(cCaption) = 'A'
  aFooter := cCaption
ELSEIF oParent:hasFooter
  aFooter := {Chr(255)}
ELSE
  aFooter := {nil}
ENDIF

ASize(aFooter,oParent:footerLines)
//
// J. Duijf 19-02-2012
// If there is more then 1 footer line, ignore the fixed footer height
// It must be recalculated based on the fontsize (dynamic)
//
if Len(aFooter) > 0
   Dc_MergePresentation(aArray,{{XBP_PP_COL_FA_HEIGHT,-1}})
endif
//
IF !( aHeader[1] == NIL )
  DC_PresAdd( aArray, { XBP_PP_COL_HA_CAPTION, aHeader[1] } )
  oParent:hasHeader := .t.
ENDIF

lPixel := aGetListItem[lGETLIST_PIXEL]

nWidth := aGetListItem[nGETLIST_WIDTH]

IF lPixel
  nWidth *= ::getList:scaleFactor[3]
  DC_PresAdd( aArray, { XBP_PP_COL_DA_ROWWIDTH, nWidth } )
  DC_PresAdd( aArray, { XBP_PP_COL_DA_CHARWIDTH, nWidth } )
ELSE
  DC_PresAdd( aArray, { XBP_PP_COL_DA_CHARWIDTH, nWidth } )
ENDIF

IF !( aFooter[1] == NIL )
  DC_PresAdd( aArray, { XBP_PP_COL_FA_CAPTION, aFooter[1] } )
  oParent:hasFooter := .t.
ENDIF
//
// J. Duijf 12-03-2011 Optional allignment has prefference
//

::parent := oParent

IF Valtype(aGetListItem[xGETLIST_OPTIONS7]) = 'N'

  if Ascan( aArray, {|a|a[1] = XBP_PP_COL_DA_CELLALIGNMENT}) = 0                  // No default
    DC_PresAdd( aArray, { XBP_PP_COL_DA_CELLALIGNMENT, aGetListItem[xGETLIST_OPTIONS7] })
  Endif

  // J. Duijf 20-06-2012 Thanks to Clifford Wiernick,
  // Only change if alignment is NOT set by previous settings
  //
  If Ascan( aArray, {|a|a[1] = XBP_PP_COL_HA_ALIGNMENT}) = 0                  // No default
     DC_PresAdd( aArray, { XBP_PP_COL_HA_ALIGNMENT,  aGetListItem[xGETLIST_OPTIONS7] } )
  Endif
  If Ascan( aArray, {|a|a[1] = XBP_PP_COL_FA_ALIGNMENT}) = 0                  // No default
     DC_PresAdd( aArray, { XBP_PP_COL_FA_ALIGNMENT,  aGetListItem[xGETLIST_OPTIONS7] } )
  Endif

ELSEIF Valtype(bDataBlock) = 'B' .AND. Valtype(Eval(bDataBlock,::parent:recordObject)) = 'N' // Default handling

  If Ascan( aArray, {|a|a[1] = XBP_PP_COL_DA_CELLALIGNMENT}) = 0                  // No default
    DC_PresAdd( aArray, { XBP_PP_COL_DA_CELLALIGNMENT, XBPALIGN_RIGHT + XBPALIGN_VCENTER } )
  ENDIF
  //
  // J. Duijf 10-03-2011
  // If a data item is numeric, and Right alligned, so should be the header and footer
  //
  // J. Duijf 12-03-2011 Not sure why i do not need to add XBPALIGN_VCENTER
  //                     The header and footer are centered anyway (and do not if XBPALIGN_VCENTER is added)
  //
  // J. Duijf 20-06-2012 Thanks to Clifford Wiernick,
  // Only change if alignment is NOT set by previous settings
  //
  If Ascan( aArray, {|a|a[1] = XBP_PP_COL_HA_ALIGNMENT}) = 0                  // No default
     DC_PresAdd( aArray, { XBP_PP_COL_HA_ALIGNMENT, XBPALIGN_RIGHT}) // + XBPALIGN_VCENTER } )
  Endif
  If Ascan( aArray, {|a|a[1] = XBP_PP_COL_FA_ALIGNMENT}) = 0                  // No default
     DC_PresAdd( aArray, { XBP_PP_COL_FA_ALIGNMENT, XBPALIGN_RIGHT}) // + XBPALIGN_VCENTER } )
  Endif

ENDIF

xFont := aGetListItem[cGETLIST_FONT]
IF !Empty(xFont)
  IF Valtype(xFont) == 'B'
    xFont := Eval(xFont)
  ENDIF
  DC_PresAdd( aArray, { XBP_PP_COMPOUNDNAME, xFont } )
ENDIF

::XbpColumn:init(,,,,aArray, .t. )

#if XPPVER > 1900331
IF xOptions[13]
  ::drawMode := XBP_DRAW_OWNER
ENDIF
#endif

IF !Empty(aGetListItem[nGETLIST_SUBTYPE])
  ::type := aGetListItem[nGETLIST_SUBTYPE]
ENDIF
::picture := aGetListItem[cGETLIST_PICTURE]

aPres := {oParent:headerPres,aGetListItem[xGETLIST_OPTIONS3]}
::Heading := DC_MultilineCellgroup():New( self,,,,,,Len(aHeader) )
::Heading:width := aGetListItem[nGETLIST_WIDTH]
::Heading:presentation := aPres
::Heading:rows := Max(Len(aHeader),oParent:Headerlines)
::Heading:noResize := ::noResize
::Heading:ReferenceString  := "W"

aPres := {oParent:footerPres,aGetListItem[xGETLIST_OPTIONS4]}
::Footing := DC_MultilineCellgroup():New( self,,,,,,Len(aFooter) )
::Footing:width := aGetListItem[nGETLIST_WIDTH]
::Footing:presentation := aPres
::Footing:rows := Max(Len(aFooter),oParent:FooterLines)
::Footing:noResize := ::noResize
::Footing:ReferenceString  := "W"

aColor := aGetListItem[aGETLIST_COLOR]

::sortBlock := DC_GetBlock( xOptions[1] )
::sortDefault := xOptions[8]

IF Valtype(::sortDefault) == 'B'
  ::sortDefault := Eval(::sortDefault,self)
  DEFAULT ::sortDefault := .f.
ENDIF

::hideBlock := DC_GetBlock(aGetListItem[bGETLIST_HIDE])
::validBlock := DC_GetBlock(aGetListItem[bGETLIST_VALID])
::whenBlock := DC_GetBlock(aGetListItem[bGETLIST_WHEN])
::filterBlock := xOptions[14]

::cargo := { ::getListPointer, aGetList, aGetListItem[xGETLIST_CARGO] }

::cellEditExitKey := xOptions[5]
::getsetBlock := {|x|Eval(bDataBlock,x,self)}
::headerBlock := aGetListItem[cGETLIST_CAPTION]
::dataSource := oParent:dataSource
::protectBlock := DC_GetBlock(aGetListItem[bGETLIST_PROTECT])
::footerBlock := xOptions[3]
::pointer := nPointer
::presArray := aArray
IF Valtype(aGetListItem[cGETLIST_LEVEL]) = 'A'
  ::columnPos := aGetListItem[cGETLIST_LEVEL,2]
ELSE
  ::columnPos := oParent:colCount + 1
ENDIF

bColorBlock := _ColorBlock( aColor )

/*
if ValType(bColorBlock) = "B"
   //
   // J. Duijf 11-03-2011
   //
   // Workaround for PDR 6365
   // The colorBlock should return nil, or a 4 element array to prevent an incorrect
   // Hilite forground color.
   // {CellFGColor,CellBGColor,CellHiliteFGColor,CellHiliteBGColor}
   // The highlite colors come from the presentation parameters.
   // If undefined, the default BG/FG hilite colors are BLUE/WHITE
   // If the colorblock returns NIL, the colors come from oParent:rowColorBlock
   //
   bColorTmpBlock    := bColorBlock

   bColorBlock       := {|a|a := Eval(bColorTmpBlock,Eval(::dataLink)),;
                              iif(ValType(a) = "A" .and. Len(a) = 2,;
                                 (Aadd(a,Dc_GetPresentation(::presArray,XBP_PP_COL_DA_HILITE_FGCLR,GRA_CLR_WHITE)), ;
                                  Aadd(a,Dc_GetPresentation(::presArray,XBP_PP_COL_DA_HILITE_BGCLR,GRA_CLR_BLUE))),;
                                  (a := iif(ValType(a) = "A",a,;
                                   iif(ValType(oParent:rowColorBlock) = "B",Eval(oParent:rowColorBlock,oParent),nil)))),a}
endif
*/

IF Valtype(bColorBlock) = 'B'
  ::colorBlock := bColorBlock
  If Valtype(oParent:rowColorBlock) = 'B'                 // Color of parent (Dc_Browse)
      //
      // J. Duijf 01-03-2013
      // In case the COLOR clause of the column return nil, then the default BROWSE color scheme
      // remains active. Before this, zebra pattern gets lost in cells with the color block nil result.
      //
      bColorTmpBlock := bColorBlock
     ::colorblock := {|a,b,c,d,e,f,g,h|aTmpColor := Eval(bColorTmpBlock,a,b,c,d,e,f,g,h),iif(ValType(aTmpColor) = "A",aTmpColor,Eval(oParent:rowColorBlock,oParent))}
  Endif
ELSEIF Valtype(oParent:rowColorBlock) = 'B'                 // Color of parent (Dc_Browse)
//  ::colorBlock := oParent:rowColorBlock
  ::colorBlock := {||Eval(oParent:rowColorBlock,oParent)}   // J. Duijf 7-03-2010
ENDIF

bBlock := DC_GetBlock(aGetListItem[bGETLIST_PREEVAL])
IF Valtype(bBlock) = 'B'
  Eval( bBlock, self )
ENDIF

lMultiline := (Valtype(bDataBlock) = 'A')

IF !Empty(::picture)
  ::dataLink := _ColPicture( bDataBlock, ::picture, self, ::parent:recordObject )
ELSEIF !lMultiline .AND. Valtype(::parent:dataSource) # 'O'
  IF Valtype(::parent:recordObject) == 'O'
    ::dataLink := {||Eval(bDataBlock,::parent:recordObject,self)}
  ELSE
    ::dataLink := {|x|Eval(bDataBlock,x,self)}
  ENDIF
ELSE
  ::dataLink := bDataBlock
ENDIF

IF lMultiLine
  ::dataArea:= DC_XbpCellGroupMultiLine():New( self )
  ::dataLinkMultiLine := ::dataLink
  ::dataLink := ::dataLinkMultiLine[1]
ENDIF

// ::XbpColumn:create( oParent,,,,aArray,.t. )
::presArray := aArray

IF nColumn > 0
  oParent:insColumn( nColumn, self )
ELSE
  oParent:addColumn( self )
ENDIF

IF lMultiline

  nColorHiliteFG := XBPSYSCLR_HILITEBACKGROUND
  aPP := oParent:setPresParam()
  nFound := AScan(aPP,{|a|a[1]==XBP_PP_COL_DA_HILITE_BGCLR})
  IF nFound = 0
    nFound := AScan(aPP,{|a|a[1]==XBP_PP_HILITE_BGCLR})
  ENDIF
  IF nFound > 0
    nColorHiliteFG := aPP[nFound,2]
  ENDIF
  nColorBG := ::dataArea:setColorBG()
  AADD( aArray, { XBP_PP_COL_DA_HILITE_FGCLR, nColorHiliteFG } )

  AADD( aArray, { XBP_PP_COL_DA_FGCLR, nColorBG } )

  ::XbpColumn:configure(,,,,aArray,.t.)
  ::dataLink := {||{}}

  ::dataArea:dataLinkMultiLine := ::dataLinkMultiLine
  ::dataArea:column            := self
  ::dataArea:browse            := ::setParent()

  DO WHILE !::dataArea:browse:isDerivedFrom( "XbpBrowse" )
    ::dataArea:browse := ::dataArea:browse:setParent()
  ENDDO

ENDIF


IF Valtype(::hideBlock)='B' .AND. Eval(::hideBlock)
  ::hide()
ENDIF

IF oParent:optimize // optimize
  ::heading:lbDblClick := {|a,b,o|_Optimize(a,b,o)}
ENDIF

::dataArea:keyboard := oParent:keyboard

// ::dataArea:lbDown := {|a,b,o|PostAppEvent(xbeP_SetInputFocus,,,self)}
IF oParent:tagEnable
   If oParent:IsderivedFrom("Dc_XbpBrowseMultiSelect")                        // Jack Duijf 02-08-2018
      ::dataArea:lbDown  := DC_MergeBlocks(::dataArea:lbDown,{|a|oParent:isLeftMouseButton := .T.,oParent:Tag(::dataArea,a)})
   Else
      ::dataarea:LbClick := DC_MergeBlocks(::dataArea:LbClick,{||oParent:Tag()} )
      ::dataArea:lbDown := DC_MergeBlocks(::dataArea:lbDown,{||oParent:isLeftMouseButton := .t.})
   Endif
   ::dataArea:Motion := {|mp1,b,oCell|oParent:Tag(oCell,mp1,0)}
   ::dataArea:lbUp := {||oParent:isLeftMouseButton := .f.}
ENDIF

::Heading:cargo := aGetListItem[cGETLIST_CAPTION]

IF !::SqlSort
  ::setHeaderCaption(::heading:cargo)
ENDIF

::Footing:cargo := xOptions[3]

::setFooterCaption(::footing:cargo)

IF Valtype(oParent:sortColorUnselectedBG)=='N' .AND. Valtype(::sortBlock)=='B'
  ::Heading:setColorBG(oParent:sortColorUnselectedBG)
ELSEIF Valtype(oParent:sortColorNonselectableBG)=='N'
  ::Heading:setColorBG(oParent:sortColorNonselectableBG)
ENDIF

IF Valtype(oParent:sortFontUnselected)=='C' .AND. Valtype(::sortBlock)=='B'
  ::Heading:setFontCompoundName(oParent:sortFontUnselected)
ELSEIF Valtype(oParent:sortFontNonselectable)=='C'
  ::Heading:setFontCompoundName(oParent:sortFontNonSelectable)
ENDIF

IF Valtype(oParent:sortColorUnselectedFG)=='N' .AND. Valtype(::sortBlock)=='B'
  ::Heading:setColorFG(oParent:sortColorUnselectedFG)
ELSEIF Valtype(oParent:sortColorNonselectableFG)=='N'
  ::Heading:setColorFG(oParent:sortColorNonselectableFG)
ENDIF

aColor := aGetListItem[xGETLIST_OPTIONS2]
IF Valtype(aColor) = 'A' // header color
  ASize(aColor,4)
  IF Valtype(aColor[1])='B'
    aColor := Eval(aColor[1])
    ASize(aColor,4)
  ENDIF
  IF Valtype(aColor[1]) = 'A'
    aColor[1] := GraMakeRGBColor(aColor[1])
  ENDIF
  IF Valtype(aColor[2]) = 'A'
    aColor[2] := GraMakeRGBColor(aColor[2])
  ENDIF
  IF Valtype(aColor[1])='N'
    ::heading:setColorFG(aColor[1])
  ENDIF
  IF Valtype(aColor[2])='N'
    ::heading:setColorBG(aColor[2])
  ENDIF
  ::headerFont := aColor[3]
  ::headerHeight := aColor[4]
ENDIF

aColor := aGetListItem[xGETLIST_OPTIONS8]
IF Valtype(aColor) = 'A' // footer color
  ASize(aColor,4)
  IF Valtype(aColor[1])='B'
    aColor := Eval(aColor[1])
    ASize(aColor,4)
  ENDIF
  IF Valtype(aColor[1]) = 'A'
    aColor[1] := GraMakeRGBColor(aColor[1])
  ENDIF
  IF Valtype(aColor[2]) = 'A'
    aColor[2] := GraMakeRGBColor(aColor[2])
  ENDIF
  IF Valtype(aColor[1])='N'
//    DC_PresAdd( aArray, { XBP_PP_COL_FA_FGCLR, aColor[1] } )
  ENDIF
  IF Valtype(aColor[2])='N'
//    DC_PresAdd( aArray, { XBP_PP_COL_FA_BGCLR, aColor[2] } )
  ENDIF
  IF Valtype(aColor[3]) == 'C'
    ::footerFont := aColor[3]
  ENDIF
  IF Valtype(aColor[4]) == 'N'
    ::footerHeight := aColor[4]
  ENDIF
ENDIF

IF Valtype(aGetListItem[xGETLIST_OPTIONS2]) = 'A' .AND. ;
   Valtype(aGetListItem[xGETLIST_OPTIONS2,1]) = 'B'
  ::headerColorBlock := aGetListItem[xGETLIST_OPTIONS2,1]
  ::setHeaderColor()
ENDIF

IF Valtype(aGetListItem[xGETLIST_OPTIONS8]) = 'A' .AND. ;
   Valtype(aGetListItem[xGETLIST_OPTIONS8,1]) = 'B'
  ::footerColorBlock := aGetListItem[xGETLIST_OPTIONS8,1]
  ::setFooterColor()
ENDIF

IF !Empty(::headerFont)
  ::heading:setFontCompoundName(::headerFont)
ENDIF
IF !Empty(::headerHeight)
  ::heading:setSize({::heading:currentSize()[1],::headerHeight})
ENDIF

IF !Empty(::footerFont)
  ::footing:setFontCompoundName(::footerFont)
ENDIF
IF !Empty(::footerHeight)
  ::footing:setSize({::footing:currentSize()[1],::footerHeight})
ENDIF

bBlock := DC_GetBlock(aGetListItem[bGETLIST_GROUP],'O')
IF Valtype( bBlock ) = 'B'
  Eval( bBlock, self )
ENDIF

bBlock := DC_GetBlock(aGetListItem[bGETLIST_GOTFOCUS])
IF Valtype( bBlock ) = 'B'
  // ::setInputFocus := bBlock
  ::dataArea:setInputFocus := bBlock
ENDIF
bBlock := DC_GetBlock(aGetListItem[bGETLIST_LOSTFOCUS])
IF Valtype( bBlock ) = 'B'
  ::killInputFocus := bBlock
  ::dataArea:killInputFocus := bBlock
ENDIF

IF Valtype(oParent:rightButtonDown) = 'B'
  ::dataArea:RbDown  := DC_MergeBlocks(::dataArea:RbDown,oParent:rightButtonDown)
ENDIF

IF Valtype(::sortBlock) = 'B' .OR. ::sqlSort
  IF Valtype(::parent:sortButtonConfig) == 'O'
    ::buildSortButton()
  ELSE
    IF xOptions[6]
      ::heading:lbdown := _HeaderDown( oParent, self )
    ENDIF
  ::heading:rbdown := _HeaderDown( oParent, self )
  ENDIF
ELSEIF ::noResize
ENDIF

DEFAULT ::sortDefault := .t.
IF ::sortDefault
  ::sort( nil, oParent:colCount )
ENDIF

::getList:isDataToolTips := ::getList:isDataToolTips .OR. Valtype(aGetListItem[xGETLIST_OPTIONS9]) == 'A'
DEFAULT aGetListItem[xGETLIST_OPTIONS9] :=  {{||.f.},nil,'',0,0,0,nil,nil}
ASize(aGetListItem[xGETLIST_OPTIONS9],8)
aGetListItem[xGETLIST_OPTIONS9,7] := oParent // pointer to browse object
aGetListItem[xGETLIST_OPTIONS9,8] := oParent:dataSource
::dataarea:cargo := aGetListItem[xGETLIST_OPTIONS9]  // DataArea tooltip enable codeblock

// :cargo[1] = Enable Code Block
// :cargo[2] = Return Code Block
// :cargo[3] = Value to display in tooltip
// :cargo[4] = Last moused cell record offset
// :cargo[5] = Last selected record
// :cargo[6] = Last moused cell record
// :cargo[7] = Pointer to Browse object
// :cargo[8] = Alias(), Array or SQLExpress object

::DC_XbaseParts:setOrig(self)

IF IsMemberVar(oParent,'USEVISUALSTYLE')
  ::heading:useVisualStyle := oParent:useVisualStyle
  ::footing:useVisualStyle := oParent:useVisualStyle
ENDIF

::origFonts := ;
   { ::dataArea:setFontCompoundName(), ;
     ::heading:setFontCompoundName(), ;
     ::footing:setFontCompoundName() }

IF IsMemberVar(oParent,'USEVISUALSTYLE')
  ::visualStyleArray[1] := ::heading:useVisualStyle
  ::visualStyleArray[2] := ::dataArea:useVisualStyle
  ::visualStyleArray[3] := ::footing:useVisualStyle
ENDIF
#if XPPVER > 1900331
::drawModeArray[1] := ::heading:drawMode
::drawModeArray[2] := ::dataArea:drawMode
::drawModeArray[3] := ::footing:drawMode
#endif

RETURN self

* ------------

METHOD DC_XbpColumn:buildSortButton()

LOCAL GetList[0], nColPos := ::parent:colCount, ;
      aSize := ::heading:currentSize()

@ 4,5 DCPUSHBUTTONXP SIZE aSize[1]-25,aSize[2]-8 PIXEL ;
      PARENT ::heading ;
      TOOLTIP ::toolTip ;
      CAPTION ::heading:cargo ;
      OBJECT ::sortButtonObject ;
      CONFIG ::parent:sortButtonConfig ;
      ACTION {|a,n,o|::Sort(::parent,nColPos),SetAppFocus(::parent)} ;
      EVAL {|o|o:parent:resize := {|a,b,o|::sortButtonObject:setSize({o:currentSize()[1]-20,o:currentSize()[2]-8})}}

DCREAD GUI EXIT PARENT self

RETURN self

* ------------

METHOD DC_XbpColumn:resizeColumn( aOldSize, aNewSize )

::setSize({::currentSize()[1]+(aNewSize[1]-aOldSize[1]),::CurrentSize()[2]})
::parent:setSize(::parent:currentSize())

RETURN self

*-------------

METHOD DC_XbpColumn:refresh()

::setHeaderCaption()
::setHeaderColor()
::setFooterCaption()
::setFooterColor()

RETURN self

* --------- This is from build 227. It works for John Lau,
* --------- The commented out version was from 232.
* --------- It didn't work for him.  I don't know why.
* --------- I'd like to find a version that works for everyone
* --------- For now, wer'e going with this until somebody complains

METHOD DC_XbpColumn:WhenHideEval( lEnable, lShow, lProtect )

LOCAL oParent, nColumn, j, oColumn, nInsColumn, lOk

Grok(lShow)

IF Valtype(::hideBlock) = 'B'
  oParent := ::parent
  nColumn := ::columnPos
  IF Eval(::hideBlock,self)
    lOk := .f.
    FOR j := 1 TO oParent:colCount
      oColumn := oParent:getColumn(j)
      IF oColumn:columnPos == nColumn // column has not been deleted
        lOk := .t.
        EXIT
      ENDIF
    NEXT
    IF lOk
      oParent:lockUpdate(.t.)
      IF !Empty(oParent:freezeLeft)
        oParent:setLeftFrozen({})
      ENDIF
      oParent:delColumn(j)
    ENDIF
  ELSE
    lOk := .t.
    nInsColumn := 1
    FOR j := 1 TO oParent:colCount
      oColumn := oParent:getColumn(j)
      IF oColumn:columnPos == nColumn   // column is already in browse
        lOk := .f.
        EXIT
      ELSEIF oColumn:columnPos < nColumn
        nInsColumn := j
      ENDIF
    NEXT
    IF lOk
      IF nColumn > 1
        nInsColumn++
      ENDIF
      IF oParent:colCount == 0
        nInsColumn := 1
      ENDIF
      oParent:lockUpdate(.t.)
      IF !Empty(oParent:freezeLeft)
        oParent:setLeftFrozen({})
      ENDIF
      oParent:insColumn(nInsColumn, self)
      #if XPPVER > 1900331
      ::heading:useVisualStyle := ::visualStyleArray[1]
      ::dataArea:useVisualStyle := ::visualStyleArray[2]
      ::footing:useVisualStyle := ::visualStyleArray[3]
      ::heading:drawMode := ::drawModeArray[1]
      ::dataArea:drawMode := ::drawModeArray[2]
      ::footing:drawMode := ::drawModeArray[3]
      #endif
    ENDIF
  ENDIF
  IF lOk
    IF !Empty(oParent:freezeLeft)
      oParent:setLeftFrozen(oParent:freezeLeft)
    ENDIF
    oParent:lockUpdate(.f.)
    oParent:forceStable()
    oParent:refreshAll()
    oParent:invalidateRect()
  ENDIF
ENDIF

IF Valtype(::whenBlock) = 'B'
  lEnable := Eval(::whenBlock,self)
  IF lEnable
    ::enable()
  ELSE
    ::disable()
  ENDIF
ENDIF

IF Valtype(lProtect) == 'L' .AND. Valtype(::protectBlock) == 'B'
  lProtect := Eval(::protectBlock,self)
ENDIF

RETURN self

*----------------------

METHOD DC_XbpColumn:setHeaderCaption( bcHeader )

LOCAL i, aHeader

IF Valtype(bcHeader) = 'B'
  ::headerBlock := bcHeader
ELSEIF Valtype(bcHeader) $ 'CNO'
  ::headerCaption := bcHeader
ELSEIF !Empty(bcHeader)
  RETURN self
ENDIF

IF Valtype(::headerBlock) = 'B'
  bcHeader := DC_XtoC(Eval(::headerBlock))
ELSE
  bcHeader := ::headerCaption
ENDIF

::headerCaption := bcHeader

IF Valtype(::headerCaption) $ 'NO'
  ::headerCaption := DC_GetBitMap(::headerCaption,,,,,.t.)
  _LoadBitMap(::heading,::headerCaption)
ELSE
  aHeader := DC_TokenArray( bcHeader, ::parent:headerDelim )
  FOR i := 1 TO Len(aHeader)
    ::heading:setCell(i,aHeader[i])
  NEXT
  ::heading:invalidateRect()
ENDIF

RETURN self

* ---------------------

METHOD DC_XbpColumn:setHeaderColor( bColor )

LOCAL aColors

IF Empty(bColor) .AND. Valtype(::headerColorBlock) # 'B'
  RETURN self
ELSEIF Valtype(bColor) = 'B'
  ::headerColorBlock := bColor
ENDIF

aColors := Eval(::headerColorBlock)

IF Valtype(aColors) = 'A'
  ASize(aColors,2)
  IF Valtype(aColors[1]) == 'A'
    aColors[1] := GraMakeRGBColor(aColors[1])
  ENDIF
  IF Valtype(aColors[2]) == 'A'
    aColors[2] := GraMakeRGBColor(aColors[2])
  ENDIF
  ::heading:setColorFG(aColors[1])
  ::heading:setColorBG(aColors[2])
ENDIF

RETURN self

* ---------------------

METHOD DC_XbpColumn:setFooterColor( bColor )

LOCAL aColors

IF Empty(bColor) .AND. Valtype(::footerColorBlock) # 'B'
  RETURN self
ELSEIF Valtype(bColor) = 'B'
  ::footerColorBlock := bColor
ENDIF

aColors := Eval(::footerColorBlock)
IF Valtype(aColors) = 'A'
  ASize(aColors,2)
  IF Valtype(aColors[1]) == 'A'
    aColors[1] := GraMakeRGBColor(aColors[1])
  ENDIF
  IF Valtype(aColors[2]) == 'A'
    aColors[2] := GraMakeRGBColor(aColors[2])
  ENDIF
  ::footing:setColorFG(aColors[1])
  ::footing:setColorBG(aColors[2])
ENDIF

RETURN self

* ------------------

METHOD DC_XbpColumn:setFooterCaption( bcFooter )

LOCAL i, aFooter

IF Valtype(bcFooter) = 'B'
  ::footerBlock := bcFooter
ELSEIF Valtype(bcFooter) $ 'CNO'
  ::footerCaption := bcFooter
ELSEIF !Empty(bcFooter)
  RETURN self
ENDIF

IF Valtype(::footerBlock) = 'B'
  bcFooter := DC_XtoC(Eval(::footerBlock))
ELSE
  bcFooter := ::footerCaption
ENDIF

::footerCaption := bcFooter
IF Valtype(::footerCaption) $ 'NO'
  ::footerCaption := DC_GetBitMap(::footerCaption,,,,,.t.)
  _LoadBitMap(::footing,::footerCaption)
ELSE
  aFooter := DC_TokenArray( bcFooter, ::parent:footerDelim )
  FOR i := 1 TO Len(aFooter)
     ::footing:setCell(i,aFooter[i])
  NEXT
  ::footing:invalidateRect()
ENDIF

RETURN self

* ---------------------

METHOD DC_XbpColumn:MoveColumn( nMove )

LOCAL nPos, oBrowse

oBrowse := ::parent
oBrowse:lockUpdate(.t.)
nPos := oBrowse:colPos
oBrowse:delColumn(nPos)
oBrowse:insColumn(nPos+nMove,self)
IF nMove >= 1
  oBrowse:left()
ELSE
  oBrowse:right()
ENDIF
oBrowse:lockUpdate(.f.)
oBrowse:forceStable()
RETURN nil

* ----------------------

METHOD DC_XbpColumn:Sort( a, nColPos, nMode )

LOCAL i, j, oColumn, oXbp, aChildList, lStatus

DEFAULT nMode := 0

IF Valtype(::sortBlock) == 'B'
  IF !Empty(a)
    IF ::parent:dataSourceType # BROWSE_SQLEXPRESS .AND. ;
       (::displayIndex # ::parent:sortColumn .OR. !::parent:descendToggle)
      IF ::parent:descendToggle
        ::parent:descending := .f.
      ENDIF
      lStatus := Eval(::sortBlock,a,::displayIndex,self)
      IF Valtype(lStatus) = 'L' .AND. !lStatus
        RETURN self
      ENDIF
    ELSE
      IF nMode = 1
        ::parent:descending := .F.
      ELSEIF nMode = 2
        ::parent:descending := .T.
      ELSE
        ::parent:descending := !::parent:descending
        IF ::parent:descending
          // ::parent:goBottom()
          ::parent:forceStable()
        ENDIF
      ENDIF
      if ::parent:dataSourceType = BROWSE_DATAOBJECT              // 04-01-2004 J.Duijf
         ::parent:datasource:DbSetDescend(::parent:descending)    // 04-01-2004 J.Duijf
         ::parent:datasource:DbGoTop()                            // 04-01-2004 J.Duijf
      elseif ::parent:dataSourceType = BROWSE_SQLEXPRESS
         Eval(::sortBlock,a,nColPos,::parent:descending)
      endif
    ENDIF
    IF Valtype(::parent:dataSource) $ 'AO'
  //    ::parent:goTop()
    ENDIF
    ::parent:refreshAll()
  ENDIF
ELSE
  RETURN self
ENDIF

FOR i := 1 TO ::parent:colCount
  oColumn := ::parent:GetColumn(i)
  IF i == nColPos
    IF nColPos # ::parent:sortColumn
      IF Valtype(oColumn:parent:sortColorSelectedFG) = 'N'
        oColumn:heading:setColorFG(oColumn:parent:sortColorSelectedFG)
      ENDIF
      IF Valtype(oColumn:parent:sortColorSelectedBG) = 'N'
        oColumn:heading:setColorBG(oColumn:parent:sortColorSelectedBG)
      ENDIF
      IF Valtype(oColumn:parent:sortFontSelected) == 'C'
        oColumn:heading:setFontCompoundName(oColumn:parent:sortFontSelected)
      ENDIF
    ENDIF
    IF oColumn:heading:status() > 0
      oXbp := nil
      aChildList := oColumn:heading:childList()
      FOR j := 1 TO Len(aChildList)
        IF aChildList[j]:isDerivedFrom('XbpStatic') .AND. ;
           Valtype(aChildList[j]:cargo) = 'C' .AND. ;
           aChildList[j]:cargo == 'SortArrow'
          oXbp := aChildList[j]
          EXIT
        ENDIF
      NEXT
      IF Empty(oXbp) .AND. ::parent:descendToggle
        oXbp := XbpStatic():new(oColumn:heading,, ;
                           {oColumn:heading:currentSize()[1]-15,6}, ;
                           {12,12},,.t.)
        oXbp:type := XBPSTATIC_TYPE_TEXT
        oXbp:cargo := 'SortArrow'
        oXbp:create()
        oXbp:setFontCompoundName('12.Marlett')
      ENDIF
      IF Valtype(oXbp) = 'O' .AND. oXbp:isDerivedFrom('XbpStatic') .AND. oXbp:type == XBPSTATIC_TYPE_TEXT
        IF ::parent:descending
          oXbp:setCaption('u')
        ELSE
          oXbp:setCaption('t')
        ENDIF
      ENDIF
    ENDIF

  ELSEIF nColPos # ::parent:sortColumn

    IF !Empty(oColumn:sortBlock)
      IF Valtype(oColumn:parent:sortColorUnSelectedFG) = 'N'
        oColumn:heading:setColorFG(oColumn:parent:sortColorUnSelectedFG)
      ENDIF
      IF Valtype(oColumn:parent:sortColorUnSelectedBG) = 'N'
        oColumn:heading:setColorBG(oColumn:parent:sortColorUnSelectedBG)
      ENDIF
      IF Valtype(oColumn:parent:sortFontUnSelected) == 'C'
        oColumn:heading:setFontCompoundName(oColumn:parent:sortFontUnSelected)
      ENDIF
    ELSE
      /*
      // This is not needed anymore - never was needed
      IF Valtype(oColumn:parent:sortColorNonselectableFG) = 'N'
        oColumn:heading:setColorFG(oColumn:parent:sortColorNonselectableFG)
      ENDIF
      IF Valtype(oColumn:parent:sortColorNonselectableBG) = 'N'
        oColumn:heading:setColorBG(oColumn:parent:sortColorNonselectableBG)
      ENDIF
      IF Valtype(oColumn:parent:sortFontNonSelectable) == 'C'
        oColumn:heading:setFontCompoundName(oColumn:parent:sortFontNonSelectable)
      ENDIF
      */
    ENDIF
    IF oColumn:heading:status() > 0 .AND. !Empty(oColumn:heading:childList())
      aChildList := oColumn:heading:childList()
      FOR j := 1 TO Len(aChildList)
        oXbp := aChildList[j]
        IF oXbp:isDerivedFrom('XbpStatic') .AND. Valtype(oXbp:cargo) = 'C' .AND. ;
           oXbp:cargo == 'SortArrow'
          // destroy the arrow
          oXbp:destroy()
        ENDIF
      NEXT
    ENDIF
  ENDIF
NEXT

::parent:sortColumn := nColPos

RETURN self

//
// J. Duijf 7-03-2010
//
// The methods below take care of the property ::Evenrow  .T. or .F.
// This can be used to inplement a zebra striping of the browse,
// even if browsing records in a database that contains deleted records.
//
METHOD DC_XbpColumn:DrawRow(nP1,xP2)
::LastRow      := nP1 + ::nOffset
::parent:evenrow := (Abs(::LastRow) % 2) = 0

RETURN ::XbpColumn:DrawRow(nP1,xP2)

METHOD DC_XbpColumn:ScrollUp(nP1,lP)
::nOffset      -= nP1
RETURN ::XbpColumn:ScrollUp(nP1,lP)

METHOD DC_XbpColumn:ScrollDown(nP1,lP)
::nOffset      += nP1
RETURN ::XbpColumn:ScrollDown(nP1,lP)

* ---------------------

METHOD DC_XbpColumn:DeleteColumnFiltered()

LOCAL i, nFound, oBrowse := ::parent, aGetList

aGetList := oBrowse:cargo[2]
nFound := AScan( aGetList, {|a|a[oGETLIST_OBJECT]==self})
IF nFound > 0
  oBrowse:delColumn(::displayIndex)
  aGetList[nFound,oGETLIST_OBJECT] := nil
  oBrowse:stabilizeColumns()
ENDIF

RETURN nil

* ------------

METHOD DC_XbpColumn:MoveColumnFiltered(nWhichWay)

LOCAL oColumn, oGetList, cSubClass, nColumn, nLockedColumns, ;
      aGetList, nFound, oBrowse := ::parent, bEval

aGetList := ::parent:cargo[2]
nFound := AScan(aGetList,{|a|a[oGETLIST_OBJECT]==self})
oGetList := DC_GetDataStore(aGetList)
oGetList:getListPointer := nFound
cSubClass := aGetList[nFound,cGETLIST_SUBCLASS]
bEval := aGetList[nFound,bGETLIST_EVAL]
DEFAULT cSubClass := 'DC_XbpColumn()'
oColumn := &cSubClass
nColumn := ::displayIndex
oBrowse:delColumn(nColumn)
aGetList[nFound,oGETLIST_OBJECT] := nil
IF nWhichWay == 1 //left
  nColumn --
ELSE //right
  nColumn++
ENDIF
oColumn := oColumn:new( oBrowse, oGetList, nColumn )
oColumn:displayIndex := nColumn
oColumn:columnPos := nColumn
aGetList[oGetList:getListPointer,oGETLIST_OBJECT] := oColumn
IF Valtype(bEval) == 'B'
  Eval(bEval,oColumn)
ENDIF
IF nWhichWay == 1
  oBrowse:left()
ELSE
  oBrowse:right()
ENDIF
oBrowse:forceStable()
oBrowse:refreshAll()
oBrowse:stabilizeColumns()

RETURN nil

* -----------

METHOD DC_XbpColumn:AddColumnFiltered(lInsert)

LOCAL i, nFound, oBrowse := ::parent, aGetList, GetList[0], aColumns[0], ;
      nPointer := 1, nLockedColumns, oGetList, oXbp, cSubClass, nColumn, ;
      lStatus, oBrowse2, bEval

aGetList := oBrowse:cargo[2]
nFound := AScan( aGetList, {|a|a[oGETLIST_OBJECT]==self})
IF nFound > 0
  FOR i := 1 TO Len(aGetList)
    IF aGetList[i,nGETLIST_TYPE] == GETLIST_BROWSECOL .AND. ;
        !Empty(aGetList[i,bGETLIST_PARENT]) .AND. ;
        Eval(aGetList[i,bGETLIST_PARENT]) == oBrowse .AND. ;
        Valtype( oXbp := aGetList[i,oGETLIST_OBJECT] ) # 'O'
      AAdd(aColumns, {aGetlist[i,cGETLIST_CAPTION], ;
                      aGetList[i,cGETLIST_SUBCLASS], ;
                      aGetList[i,bGETLIST_EVAL], ;
                      i})
    ENDIF
  NEXT
ENDIF

ASort( aColumns,,,{|a,b|a[1]<b[1]} )

@ 0,0 DCBROWSE oBrowse2 DATA aColumns SIZE 45,20 FONT '10.Lucida Console' NOHSCROLL ;
      ITEMSELECTED {||DC_ReadGuiEvent(DCGUI_EXIT_OK,GetList)} ;
      POINTER nPointer ;
      EVAL {|o|o:keyboard := {|n,b,o|PickBrowseKeyboard(n,o)}} ;

DCBROWSECOL ELEMENT 1 HEADER 'Column Name' WIDTH 45 PARENT oBrowse2

DCREAD GUI FIT ADDBUTTONS TITLE 'Choose a Column to ' + IIF(lInsert,'Insert','Add') ;
    MODAL TO lStatus SETFOCUS @oBrowse2 NOAUTORESTORE

IF lStatus
  oGetList := DC_GetDataStore(aGetList)
  oGetList:getListPointer := aColumns[nPointer,4]
  cSubClass := aColumns[nPointer,2]
  bEval := aColumns[nPointer,3]
  DEFAULT cSubClass := 'DC_XbpColumn()'
  oXbp := &cSubClass
  IF lInsert
    nColumn := ::displayIndex
  ELSE
    nColumn := 0
  ENDIF
  oXbp := oXbp:new( oBrowse, oGetList, nColumn )
  aGetList[oGetList:getListPointer,oGETLIST_OBJECT] := oXbp
  IF Valtype(bEval) == 'B'
    Eval(bEval,oXbp)
  ENDIF
  oBrowse:refreshAll()
  oBrowse:stabilizeColumns()
ENDIF

RETURN lStatus

* ------------

STATIC FUNCTION _LoadBitmap( oParent, oCaption )

LOCAL oXbp

oXbp := XbpStatic():new(oParent,,{2,2},{oCaption:xSize,oCaption:ySize},,.t.)
oXbp:type := XBPSTATIC_TYPE_BITMAP
oXbp:caption := oCaption
oXbp:create()
DC_CenterObject(oXbp)

RETURN nil

* -------------

STATIC FUNCTION _HeaderDown( oBrowse, oColumn, nColPos )

DEFAULT nColPos := oBrowse:colCount

RETURN {|a,n,o|oColumn:Sort(a,nColPos)}

* -------------

FUNCTION DC_SkipArray( nWantSkip, oBrowse )

LOCAL nDidSkip, nPointer, nLastRec := Len(oBrowse:dataSource), ;
      nSavePointer, nRecords := 0

IF nLastRec = 0
  RETURN 0
ENDIF

nDidskip := nWantSkip

IF Empty(oBrowse:filter)

  IF oBrowse:descending
    IF oBrowse:arrayElement - nWantSkip < 1 // bof
      nDidSkip := -1 * (1 - oBrowse:arrayElement)
    ELSEIF oBrowse:arrayElement - nWantSkip > nLastRec // eof
      nDidSkip := -1 * (nLastRec - oBrowse:arrayElement)
    ENDIF
    oBrowse:arrayElement -= nDidSkip
  ELSE
    IF oBrowse:arrayElement + nWantSkip < 1 // bof
      nDidSkip := 1 - oBrowse:arrayElement
    ELSEIF oBrowse:arrayElement + nWantSkip > nLastRec // eof
      nDidSkip := nLastRec - oBrowse:arrayElement
    ENDIF
    oBrowse:arrayElement += nDidSkip
  ENDIF

ELSE

  nDidSkip := 0
  nPointer := oBrowse:arrayElement

  nSavePointer := nPointer
  IF oBrowse:descending
    nWantSkip *= -1
  ENDIF
  IF nWantSkip == 0
    nDidSkip := 0
  ELSEIF nWantSkip > 0
    DO WHILE nDidSkip < nWantSkip
      nPointer++
      IF nPointer > Len(oBrowse:dataSource)
        nPointer := Len(oBrowse:dataSource)
        DO WHILE nPointer > 0 .AND. !Eval(oBrowse:filter,oBrowse:dataSource[nPointer])
          nPointer --
        ENDDO
        EXIT
      ELSEIF Eval(oBrowse:filter,oBrowse:dataSource[nPointer])
        nDidSkip++
      ENDIF
    ENDDO
  ELSEIF nWantSkip < 0
    DO WHILE nDidSkip > nWantSkip
      nPointer--
      IF nPointer < 1
        nPointer := 1
        DO WHILE nPointer < Len(oBrowse:dataSource) .AND. !Eval(oBrowse:filter,oBrowse:dataSource[nPointer])
          nPointer ++
        ENDDO
        EXIT
      ELSEIF Eval(oBrowse:filter,oBrowse:dataSource[nPointer])
        nDidSkip--
      ENDIF
    ENDDO
  ENDIF
  IF oBrowse:descending
    nDidSkip := -1*nDidSkip
  ENDIF
  IF nDidSkip == 0
    nPointer := nSavePointer
  ENDIF
  oBrowse:arrayElement := nPointer

ENDIF

IF oBrowse:arrayElement > Len(oBrowse:dataSource)
  oBrowse:arrayElement := Len(oBrowse:dataSource)
ENDIF

oBrowse:cargo[4] := oBrowse:arrayElement

RETURN nDidSkip

* ------------------

STATIC FUNCTION ArrayPos( oBrowse )

RETURN Int(oBrowse:arrayElement/Len(oBrowse:dataSource))*100

* ------------------

STATIC FUNCTION ArrayGoPos( oBrowse, n )

RETURN Len(oBrowse:dataSource)*(n/100)


* ------------------

STATIC FUNCTION ArrayGoBottom( oBrowse )

LOCAL nElement := 0

IF Len(oBrowse:dataSource) = 0
  RETURN 0
ENDIF

IF Empty(oBrowse:filter)
  IF oBrowse:descending
    nElement := 1
  ELSE
    nElement := Len(oBrowse:dataSource)
  ENDIF
ELSE
  IF oBrowse:descending
    nElement := 1
    DO WHILE nElement <= Len(oBrowse:dataSource)
      IF Eval(oBrowse:filter,oBrowse:dataSource[nElement])
        EXIT
      ENDIF
      nElement++
    ENDDO
  ELSE
    nElement := Len(oBrowse:dataSource)
    DO WHILE nElement > 0
      IF Eval(oBrowse:filter,oBrowse:dataSource[nElement])
        EXIT
      ENDIF
      nElement--
    ENDDO
  ENDIF
ENDIF

RETURN nElement

* ------------------

STATIC FUNCTION ArrayGoTop( oBrowse )

LOCAL nElement := 0

IF Len(oBrowse:dataSource) = 0
  RETURN 0
ENDIF

IF Empty(oBrowse:filter)
  IF oBrowse:descending
    nElement := Len(oBrowse:dataSource)
  ELSE
    nElement := 1
  ENDIF
ELSE
  IF !oBrowse:descending
    nElement := 1
    DO WHILE nElement <= Len(oBrowse:dataSource)
      IF Eval(oBrowse:filter,oBrowse:dataSource[nElement])
        EXIT
      ENDIF
      nElement++
    ENDDO
    IF nElement > Len(oBrowse:dataSource)
      nElement := 0
    ENDIF
  ELSE
    nElement := Len(oBrowse:dataSource)
    DO WHILE nElement > 0
      IF Eval(oBrowse:filter,oBrowse:dataSource[nElement])
        EXIT
      ENDIF
      nElement--
    ENDDO
  ENDIF
ENDIF

RETURN nElement

* ------------------

STATIC FUNCTION _Pres( aPres )

IF Valtype( aPres ) = 'B'
  RETURN Eval(aPres)
ENDIF

RETURN aPres

* ----------------------

STATIC FUNCTION _ColPicture( bBlock, bcPicture, oColumn, oRecord )

LOCAL bDataLink

IF Valtype(bcPicture) = 'B'

  IF Valtype(oRecord) == 'O'
    bDataLink := {|x,y,z|x := Eval(bBlock,oRecord,oColumn), ;
                       z := Eval(bcPicture), ;
                       y:=IIF( x==NIL,Space(Len(z)),Transform(x,z)),y }
  ELSE
    bDataLink := {|x,y,z|x := Eval(bBlock,x,oColumn), ;
                         z := Eval(bcPicture), ;
                         y:=IIF( x==NIL,Space(Len(z)),Transform(x,z)),y }
  ENDIF

ELSE

  IF Valtype(oRecord) == 'O'
    bDataLink := {|x,y,z|x := Eval(bBlock,oRecord,oColumn), ;
                         z := bcPicture, ;
                         y:=IIF( x==NIL,Space(Len(z)),Transform(x,z)),y }
  ELSE
    bDataLink := {|x,y,z|x := Eval(bBlock,x,oColumn), ;
                       z := bcPicture, ;
                       y:=IIF( x==NIL,Space(Len(z)),Transform(x,z)),y }
  ENDIF

ENDIF

RETURN bDataLink

* ----------------------

STATIC FUNCTION _ColorBlock( aColor )

IF Valtype(aColor) = 'A' .AND. Len(aColor) = 2
  IF Valtype(aColor[1]) = 'B'
     RETURN aColor[1]
  ELSE
     RETURN {||aColor}
  ENDIF
ENDIF

RETURN {|a,o|o:TagColor()}

* ---------------------

STATIC PROCEDURE _Optimize( a, b, o )

LOCAL aArray, oColumn := o:setParent(), i, nWidth := 0, ;
      oPS, cCaption, oBrowse, aGraArray, nPointer, bDataLink, ;
      nColumn, cAlias, nSaveRec

Grok(a,b)
oBrowse  := oColumn:parent
nColumn := oBrowse:colPos

nPointer := oBrowse:arrayElement  // save pointer
bDataLink := oColumn:dataLink

oPS := o:lockPS()

IF Empty(oColumn:columnSize)
  oColumn:columnSize := oColumn:currentSize()[1]
  IF Valtype(oColumn:dataSource) = 'A'
    aArray   := oColumn:dataSource
    FOR i := 1 TO Len(aArray)
      oBrowse:arrayElement := i
      cCaption := DC_XtoC(Eval(bDataLink))
      aGraArray := GraQueryTextBox( oPS, cCaption )
      nWidth := Max( nWidth, ( aGraArray[3,1] - aGraArray[1,1]) + 2 )
    NEXT
  ELSEIF Valtype(oColumn:dataSource) = 'C'
    cAlias := oColumn:dataSource
    nSaveRec := (cAlias)->(RecNo())
    (cAlias)->(DC_DbGoTop())
    DO WHILE !(cAlias)->(DC_Eof())
      cCaption := (cAlias)->(DC_XtoC(Eval(bDataLink)))
      aGraArray := GraQueryTextBox( oPS, cCaption )
      nWidth := Max( nWidth, ( aGraArray[3,1] - aGraArray[1,1]) + 2 )
      (cAlias)->(DC_DbSkip())
    ENDDO
    (cAlias)->(dbGoTo(nSaveRec))
  ELSE
    o:unlockPS()
    RETURN
  ENDIF
ELSE
  nWidth := oColumn:columnSize
  oColumn:columnSize := 0
ENDIF

oBrowse:arrayElement := nPointer // restore pointer

IF Valtype(Eval(bDataLink)) = 'D'
  nWidth += 10
ELSEIF Valtype(Eval(bDataLink)) = 'N'
  nWidth += 30
ELSE
  nWidth += 10
ENDIF

o:unlockPS()
oColumn:setSize( {nWidth, oColumn:currentSize()[2] })
oBrowse:setSize(oBrowse:currentSize())
oBrowse:firstCol()
FOR i := 1 TO nColumn-1
  oBrowse:right()
NEXT
oBrowse:forceStable()

RETURN

* ---------------

FUNCTION DC_GetColArray( nElement, oBrowse, xVar )

LOCAL aArray, nPointer, xValue

IF Valtype(xVar) == 'A'
  RETURN xVar[nElement]
ENDIF

aArray   := oBrowse:dataSource
nPointer := oBrowse:arrayElement

IF nPointer > Len(aArray) .OR. nPointer = 0
  RETURN ''
ENDIF

IF nElement == NIL
  RETURN aArray[nPointer]
ELSEIF nElement <= 0
  RETURN nPointer
ELSEIF Empty(aArray) .OR. aArray[1] == NIL
  RETURN 0
ENDIF

xValue := aArray[nPointer,nElement]

IF Valtype(xVar) == Valtype(xValue)
  aArray[nPointer,nElement] := xVar
ENDIF

RETURN aArray[nPointer,nElement]

* -------------------

FUNCTION DC_BrowseRow( oBrowse, nElement )

LOCAL aArray, nOldElement

IF Valtype(oBrowse) # 'O'
   RETURN 0
ENDIF

aArray   := oBrowse:cargo[5]
nOldElement := oBrowse:cargo[4]

IF Valtype(nElement) = 'N' .AND. nElement > 0 .AND. nElement <= Len(aArray)
  oBrowse:cargo[4] := nElement
  oBrowse:arrayElement := nElement
ENDIF

RETURN nOldElement

* ----------------------

FUNCTION DC_GetBrowArray( oBrowse, aArray )

IF Valtype( aArray ) = 'A'
  IF oBrowse:isDerivedFrom('XbpBrowse')
    oBrowse:cargo[5] := aArray
  ENDIF
  oBrowse:dataSource := aArray
ENDIF

RETURN oBrowse:dataSource

* ----------------------

FUNCTION DC_BrowseSort( aSort )

STATIC saSort := ;
   { GRA_CLR_WHITE, ;    // Sort Selected Color (Foreground)
     GRA_CLR_RED, ;      // Sort Selected Color (Background)
     nil, ;              // Sort Nonselectable Color (Foreground)
     nil, ;              // Sort Nonselectable Color (Background)
     nil, ;              // Sort UP Bitmap (no longer used)
     nil, ;              // Sort DOWN Bitmap (no longer used)
     GRA_CLR_WHITE, ;    // Sort Unselected Color (Foreground)
     GRA_CLR_BLUE, ;     // Sort Unselected Color (Background)
     nil, ;              // Sort Selected Font
     nil, ;              // Sort Unselected Font
     nil, ;              // Sort Nonselectable Font
     nil }               // Sort Button Configuration object (DC_XbpPushButtonXPConfig())

LOCAL aOldSort := saSort

IF Valtype(aSort) = 'A'
  saSort := aSort
ENDIF

DEFAULT aOldSort := {}
ASize(aOldSort,12)

RETURN aOldSort

* ----------------------

FUNCTION DC_BrowseColor( aColor )

STATIC saColor := nil

LOCAL aOldColor := saColor

IF Valtype(aColor) = 'A'
  saColor := aColor
  ASize(saColor,4)
  IF Valtype(saColor[1]) == 'A'
    saColor[1] := GraMakeRGBColor(saColor[1])
  ENDIF
  IF Valtype(saColor[2]) == 'A'
    saColor[2] := GraMakeRGBColor(saColor[2])
  ENDIF
  IF Valtype(saColor[3]) == 'A'
    saColor[3] := GraMakeRGBColor(saColor[3])
  ENDIF
  IF Valtype(saColor[4]) == 'A'
    saColor[4] := GraMakeRGBColor(saColor[4])
  ENDIF
ELSEIF Pcount() == 1 .AND. aColor == NIL
  saColor := NIL
ENDIF

RETURN aOldColor


FUNCTION DC_MergePresentation(aTarget,aSource )
//
// J. Duijf    10-03-2011
//
// Merge presentation parameters.
// Both parameters are 2 dimensional arrays.
// Add item to target in not exist.
// Replace value is already exist.
//
// This function is used while creating Dc_XbpBrowse and Dc_XbpColumn
//
LOCAL nPos           := 0
LOCAL ni             := 0
if ValType(aTarget) = "A" .and. ;                                             // Make sure its a array
   ValType(aSource) = "A"
   For ni := 1 To Len(aSource)
      nPos           := Ascan(aTarget,{|a|a[1] = aSource[ni,1]})              // Locate property in target
      if nPos = 0                                                             // Not there,
         Aadd(aTarget,Aclone(aSource[ni]))                                    // Add to target
      else                                                                    // Ok, found
         aTarget[nPos,2] := aSource[ni,2]                                     // Replace propperty
      endif
   Next ni
endif
Return aTarget                                                                // Updated array presentation parameters

FUNCTION DC_GetPresentation(aPres,nId,uDefault)
//
// J. Duijf    10-03-2011
//
// Locate ID in presentation parameters.
// Return value if found, or default value if not found
//
// This function is used while creating Dc_XbpBrowse and Dc_XbpColumn
//
LOCAL nPos           := 0
LOCAL uRet           := uDefault
if ValType(aPres) = "A"                                                       // Make sure its a array
   nPos              := Ascan(aPres,{|a|a[1] == nId})
   if nPos > 0                                                                // ID found
      uRet           := aPres[nPos,2]                                         // Return value
   endif
endif
Return uRet

* ------------

CLASS DC_XbpCellGroupMultiLine FROM XbpCellGroup

PROTECTED:
   VAR aRows

EXPORTED:
   VAR column
   VAR browse
   VAR dataLinkMultiLine

   METHOD setCell
   METHOD drawCell
   METHOD hiliteCell
   METHOD paint
   METHOD setSize

   INLINE METHOD destroy
      ::XbpCellGroup:destroy()
//      ::aRows:= ::column:= ::browse := nil
   RETURN self

ENDCLASS

* -------------

METHOD DC_XbpCellGroupMultiLine:setCell( nRowPos, xValue, nType, lRepaint )

LOCAL aData

IF xValue <> nil
   aData:= {}
   AEVAL( ::dataLinkMultiLine, {|b| AADD( aData, EVAL( b ) ) } )
ENDIF

RETURN ::XbpCellGroup:setCell( nRowPos, aData, nType, lRepaint )

* --------------

METHOD DC_XbpCellGroupMultiLine:hiliteCell( nRowPos, lHilite, lFrame, lRepaint )

LOCAL oPS, aAttr, aAttrAntes, nColorHiliteFG, nFound, aPP
LOCAL aData := ::getCell( nRowPos )
LOCAL lExit := .T.
//
// J. Duijf 19-02-2012
// To prevent garbled data on the first line of a multiline browse:
// Clear the cell data just before highliting
//
if ValType(aData) = "A"                                                       // Only if multiline cell
   ::XbpCellGroup:setCell( nRowPos, "",,.F.)                                  // Clear cell (no paint)
   lExit := ::XbpCellGroup:hiliteCell( nRowPos, lHilite, lFrame, lRepaint )   // Hilite
   ::XbpCellGroup:setCell( nRowPos, aData,,.F.)                               // Restore cell data (no paint)
else                                                                          // Regular data
   lExit := ::XbpCellGroup:hiliteCell( nRowPos, lHilite, lFrame, lRepaint )
endif
IF aData <> nil

   oPS := ::lockPS()

   IF lHilite
     nColorHiliteFG := XBPSYSCLR_HILITEBACKGROUND
     aPP := ::browse:setPresParam()
     nFound := AScan(aPP,{|a|a[1]==XBP_PP_COL_DA_HILITE_FGCLR})
     IF nFound = 0
      nFound := AScan(aPP,{|a|a[1]==XBP_PP_HILITE_FGCLR})
     ENDIF
     IF nFound > 0
       nColorHiliteFG := aPP[nFound,2]
     ENDIF
     aAttr:= ARRAY( GRA_AS_COUNT )
     aAttr[ GRA_AS_COLOR ] := nColorHiliteFG
     aAttrAntes:= oPS:SetAttrString( aAttr )
     _DrawCell( self, oPS, nRowPos )
     oPS:SetAttrString( aAttrAntes )
   ELSE
     _DrawCell( self, oPS, nRowPos )
   ENDIF

   ::unLockPs( oPS )

ENDIF

RETURN lExit

* -------------

METHOD DC_XbpCellGroupMultiLine:drawCell( xRowPos, lRepaint )

LOCAL oPS, nCont, nFin, lExit:= ::XbpCellGroup:drawCell( xRowPos, lRepaint )

oPS:= ::lockPS()

IF ValType( xRowPos ) == 'A'
    nFin:= LEN( xRowPos )
    FOR nCont:= 1 TO nFin
      IF xRowPos[ nCont ] <> nil
        _DrawCell( self, oPS, xRowPos[ nCont ] )
      ENDIF
    NEXT
ELSEIF ValType( xRowPos ) == 'N'
    _DrawCell( self, oPS, xRowPos )
ENDIF

::unLockPs( oPS )

RETURN lExit

* --------------

METHOD DC_XbpCellGroupMultiLine:paint( aRect )

Grok(aRect)
IF NextAppEvent() <> xbeBRW_Navigate

   ::drawCell( ::aRows, .T. )
   IF ::browse:cursorMode == XBPBRW_CURSOR_ROW ;
         .OR. ( ::browse:cursorMode == XBPBRW_CURSOR_CELL ;
                .AND. ::browse:getColumn( ::browse:colPos ) == ::column )
      ::hiliteCell( ::browse:rowPos, .T., .F., .T. )
   ENDIF

ENDIF

RETURN self

* -----------

METHOD DC_XbpCellGroupMultiLine:setSize( aSize, lPaint )

LOCAL lExit := ::XbpCellGroup:setSize( aSize, lPaint )

::aRows:= AEVAL( Array( ::rowCount() ), {|x,i| x:= i },,, .T. )

RETURN lExit

*-------------

STATIC FUNCTION _DrawCell( oCellGroup, oPS, nRowPos )

LOCAL self := oCellGroup, aPos, nY, aArray, nSpace, ;
      nCellHeight, nFontHeight, nAlign, nFound, nOffset, aPP, ;
      aGetData[0], aData[0], i, j

aGetData := ::getCell(nRowPos)

IF aGetData <> nil

  FOR i := 1 TO Len(aGetData)
    IF Valtype(aGetData[i]) = 'B'
      AAdd(aData,Eval(aGetData[i]))
    ELSEIF Valtype(aGetData[i]) == 'A'
      FOR j := 1 TO Len(aGetData[i])
        AAdd(aData,aGetData[i,j])
      NEXT
    ELSE
      AAdd(aData,aGetData[i])
    ENDIF
  NEXT

  aPos:= ::cellRect( nRowPos, .T. )

  nCellHeight := aPos[4]-aPos[2]
  aArray := GraQueryTextBox( oPS, 'test' )
  nFontHeight := aArray[1,2] - aArray[2,2] // height

  aPP := ::column:presArray
  nFound := AScan(aPP,{|a|Valtype(a[1])=='N'.AND. ;
                   (a[1]==XBPCOL_DA_CELLALIGNMENT .OR. ;
                    a[1]==XBP_PP_COL_DA_CELLALIGNMENT) } )

  IF nFound == 0
    aPP := ::browse:setPresParam()
    nFound := AScan(aPP,{|a|Valtype(a[1])=='N'.AND. ;
                   a[1]==XBP_PP_COL_DA_CELLALIGNMENT} )
  ENDIF
  IF nFound > 0
    nAlign := aPP[nFound,2]
  ELSE
    nAlign := XBPALIGN_VCENTER
  ENDIF

  nSpace := nFontHeight
  nOffset := 3
  IF nSpace * Len(aData) >= nCellHeight + nOffset
    IF nAlign == XBPALIGN_VCENTER
      nOffset := 0
    ENDIF
    nSpace := nCellHeight/Len(aData)
  ENDIF

  IF nAlign = XBPALIGN_TOP
    nY := aPos[2] + nCellHeight - nFontHeight + nOffset
  ELSEIF DC_BitTest(nAlign,4) // align bottom
    nY := aPos[2] +  ( nSpace * Len(aData) ) - nSpace
  ELSEIF DC_BitTest(nAlign,3) // align vcenter
    nY := aPos[2] + ;
      ( nCellHeight + ( nSpace * Len(aData) ) ) / 2 - nSpace + nOffset
  ELSE // align top
    nY := aPos[2] + nCellHeight - nFontHeight + nOffset
  ENDIF

  AEVAL( aData, {|xDat| GraStringAt( oPS, ;
           { _AlignText(aPos,oPS,nAlign,xDat), nY }, Var2LChar( xDat ) ), nY -= nSpace } )
ENDIF

RETURN nil

* ------------

STATIC FUNCTION _AlignText(aPos,oPS,nAlign,xDat)

LOCAL nCellWidth, nTextWidth, aArray

IF DC_BitTest(nAlign,2) // align right
  nCellWidth := aPos[3] - aPos[1]
  aArray := GraQueryTextBox(oPS,xDat)
  nTextWidth := aArray[3,1] - aArray[1,1]
  RETURN nCellWidth - nTextWidth
ELSEIF DC_BitTest(nAlign,1) // align hcenter
  nCellWidth := aPos[3] - aPos[1]
  aArray := GraQueryTextBox(oPS,xDat)
  nTextWidth := aArray[3,1] - aArray[1,1]
  RETURN (nCellWidth - nTextWidth) / 2
ENDIF

RETURN aPos[1]

*-------------

STATIC FUNCTION ACEDbSkipper( nWantSkip, oServer )

   LOCAL nDidSkip := 0

   DO CASE
   CASE oServer:LastRec() == 0
   CASE nWantSkip == 0
      oServer:Skip(0)

   CASE nWantSkip > 0 .AND. oServer:Recno() <> oServer:LastRec() + 1

      // Abwrts skippen
      DO WHILE nDidSkip < nWantSkip
         oServer:Skip(1)
         IF oServer:Eof()
            IF .f. // lAppend
               // Append Modus: Ghost record
               nDidSkip ++
            ELSE
               oServer:Skip(-1)
            ENDIF
            EXIT
         ENDIF
         nDidSkip ++
      ENDDO

   CASE nWantSkip < 0

      // Aufwrts skippen
      DO WHILE nDidSkip > nWantSkip
         oServer:Skip(-1)
         IF oServer:Bof()
            EXIT
         ENDIF
         nDidSkip --
      ENDDO
   ENDCASE

RETURN  nDidSkip

* --------------

FUNCTION DC_AutoRestoreBrowse( aParams )

STATIC saParams := nil
LOCAL aOldParams := saParams

// [1] nHKey
// [2] Start Folder
// [3] Save Proc Name

IF Valtype(aParams) = 'A'
  saParams := aParams
  ASize(saParams,3)
  DEFAULT saParams[3] := .T.
ENDIF

RETURN aOldParams

* -------------

STATIC FUNCTION ApplyScaleFactor( aPres, oGetList )

LOCAL nScale, i

IF Valtype(aPres) = 'A'
  IF (nScale := oGetList:scaleFactor[4]) # 1
    aPres := AClone(aPres)
    FOR i := 1 TO Len(aPres)
      IF aPres[i,1] $ { XBP_PP_COL_DA_ROWHEIGHT, XBP_PP_COL_HA_HEIGHT, XBP_PP_COL_FA_HEIGHT }
         If aPres[i,2] <> -1
            aPres[i,2] *= nScale
         Endif
      ENDIF
    NEXT
  ENDIF
ENDIF

RETURN aPres

* --------------

FUNCTION DC_BrowseSQLExpress( oCursor )

LOCAL GetList[0], oBrowse, i, GetOptions

IF Valtype(oCursor) # 'O'
  RETURN .F.
ENDIF

@ 0,0 DCBROWSE oBrowse DATA oCursor ;
  SIZE 100,25 ;
  ID 'SQL_BROWSE' ;
  PRESENTATION DC_BrowPres() ;
  CURSORMODE XBPBRW_CURSOR_ROW

FOR i := 1 TO oCursor:fCount()

  DCBROWSECOL DATA SQLFieldBlock( @oCursor, i ) ;
     WIDTH 10 ;
     HEADER oCursor:fieldName(i) ;
     PARENT oBrowse ;
     SORT SQLSortBlock(@oBrowse,i)

NEXT

DCGETOPTIONS ;
   RESIZE ;
   RESIZEDEFAULT DCGUI_RESIZE_RESIZEONLY

DCREAD GUI ;
   FIT ;
   OPTIONS GetOptions ;
   TITLE 'SQL Browse'

RETURN .T.

* ---------------

STATIC FUNCTION SQLFieldBlock( oCursor, nField )

Return {||oCursor:fieldGet(nField)}

* ---------------

STATIC FUNCTION SQLSortBlock( oBrowse, nField, nColumn )

RETURN {|a,b,descend|;
   IIF(valtype(nField)=='C',nField:=oBrowse:dataSource:fieldPos(nField),nil), ;
       oBrowse:dataSource:sort(IIF(descend,;
       {|x,y|IIF(Valtype(x[nField])='C', ;
             Upper(x[nField]) > Upper(y[nField]), x[nField] > y[nField])}, ;
       {|x,y|IIF(Valtype(x[nField])='C', ;
             Upper(x[nField]) < Upper(y[nField]), x[nField] < y[nField])})), ;
       nColumn := nField}


* --------------

FUNCTION DC_XbpBrowseDefault( oBrowDefault )

STATIC soBrowDefault := nil
LOCAL oOldBrowDefault

oOldBrowDefault := soBrowDefault

IF Valtype(oBrowDefault) == 'O'
  soBrowDefault := oBrowDefault
ENDIF

IF Empty(oOldBrowDefault)
  oOldBrowDefault := DC_XbpBrowseConfig():new()
ENDIF

RETURN oOldBrowDefault

* ----------------

CLASS DC_XbpBrowseConfig

EXPORTED:

VAR font
VAR presParam
VAR useVisualStyleBrowse
VAR useVisualStyleScrollBars

METHOD Init

ENDCLASS

* ---------------

METHOD DC_XbpBrowseConfig:init()

::font := nil
::presParam := DC_BrowPres()
::useVisualStyleBrowse := .f.
::useVisualStyleScrollBars := .t.

RETURN self

* ---------------

GETSETFUNCTION DC_GetWhenOnBrowseItemMarked DEFAULT .f.

* ---------------

FUNCTION DC_Browse2Dbf( oBrowse, cDbfName, aStru )

LOCAL oColumn, aColumns, i, nRecord, nPreviousRecord, xValue, lStatus := .t., ;
      cFiblockpe, cFieldName, cFieldType, cAlias

IF Valtype(oBrowse:dataSource) = 'C'
  nPreviousRecord := (oBrowse:dataSource)->(RecNo())
  oBrowse:stableblock := {||nRecord := (oBrowse:dataSource)->(RecNo())}
ELSEIF Valtype(oBrowse:dataSource) = 'A'
  nPreviousRecord := oBrowse:arrayElement
  oBrowse:stableblock := {||nRecord := oBrowse:arrayElement}
ELSE
  RETURN .f.
ENDIF

aColumns := Array(oBrowse:colCount)

IF Len(aStru) # Len(aColumns)
  RETURN .f.
ENDIF

dbCreate( cDbfName, aStru )
USE (cDbfName) NEW
cAlias := Alias()

oBrowse:goTop()

FOR i := 1 TO Len(aColumns)
  oColumn := oBrowse:getColumn(i)
  aColumns[i]   := oColumn
NEXT

DO WHILE .t.

  (cAlias)->(DC_AddRec())

  FOR i := 1 TO Len(aColumns)
    xValue := Eval(aColumns[i]:dataLink)
    cFieldName := aStru[i,1]
    cFieldType := aStru[i,2]
    IF !(Valtype(xValue) == cFieldType)
      IF cFieldType == 'N' .AND. Valtype(xValue) == 'C'
        xValue := Val(xValue)
      ELSEIF cFieldType == 'D' .AND. Valtype(xValue) == 'C'
        xValue := Ctod(xValue)
      ELSEIF cFieldType = 'C'
        xValue := DC_XtoC(xValue)
      ENDIF
    ENDIF
    REPLACE (cAlias)->&(cFieldName) WITH xValue
  NEXT
  (cAlias)->(dbRUnlock())

  oBrowse:down()
  oBrowse:forcestable()
  IF nRecord == nPreviousrecord
    EXIT
  ENDIF
  nPreviousrecord := nRecord

ENDDO

RETURN lStatus

* -----------

FUNCTION DC_ObjectVarBlock( cVar, nElement )

LOCAL bAccess

bAccess := {|u,o|_ObjectAccess(u,o,cVar,nElement)}

RETURN bAccess

* ---------

STATIC FUNCTION _ObjectAccess( u,oColumn,cVar,nElement )

LOCAL oBrowse, aData, oObject, bError, lStatus := .f.
LOCAL oError
IF Valtype(oColumn) # 'O'
  RETURN ''
ENDIF

oBrowse := oColumn:parent
aData := oBrowse:dataSource

IF Empty(oColumn:objectVarName) .AND. Valtype(cVar) == 'C'
  oColumn:objectVarName := cVar
ENDIF

IF Empty(aData) .OR. oBrowse:arrayElement = 0 .OR. oBrowse:arrayElement > Len(aData)
   //
   // Return correct default value
   // Check the column type
   // J. Duijf 26-07-2014
   //
   // This prevents runtime error if browse is array, and contains images
   //
   Do Case
      Case oColumn:Type = XBPCOL_TYPE_TEXT
         RETURN ''
      Case oColumn:Type = XBPCOL_TYPE_MULTILINETEXT
         RETURN ''
      Otherwise
//         XBPCOL_TYPE_ICON
//         XBPCOL_TYPE_BITMAP
//         XBPCOL_TYPE_SYSICON
//         XBPCOL_TYPE_FILEICON
//         XBPCOL_TYPE_FILEMINIICON
         Return 0
   Endcase
ENDIF
//
// 12-12-2013 J. Duijf
// Prevent runtime error. Test array boundary
//
If oBrowse:arrayElement <= Len(aData) .and. oBrowse:arrayElement > 0
   IF Valtype(nElement) == 'N'
     If nElement <= Len(aData[oBrowse:arrayElement]) .and. nElement > 0
        oObject := aData[oBrowse:arrayElement,nElement]
     Endif
   ELSE
     oObject := aData[oBrowse:arrayElement]
   ENDIF
Endif

IF Valtype(oObject) # 'O'
  u := '<not an object>'
  lStatus := .t.
ELSE
  BEGIN SEQUENCE
  IF '{' $ cVar
    cVar := &cVar
    oColumn:varBlock := cVar
  ENDIF
  IF Valtype(u) == 'A'
    u := u[nElement]
  ENDIF
  IF Valtype(u) == 'O'
    oObject := u
    u := Nil
  ENDIF
  IF u == nil
    IF Valtype(cVar) = 'B'
      u := Eval(cVar,oObject)
    ELSE
      u := oObject:&(cVar)
    ENDIF
  ELSE
    IF Valtype(cVar) = 'B'
      u := Eval(cVar,oObject,u)
    Else
      u := oObject:&(cvar) := u
    Endif
  ENDIF
  lStatus := .t.
  RECOVER using oError
   oError := oError
  END SEQUENCE
ENDIF
IF !lStatus
  u := '<' + DC_XtoC(cVar) + ':undefined>'
ENDIF

RETURN u

* ---------------

CLASS DC_BrowseQuery

EXPORTED:

VAR alias
VAR queryArray
VAR browseObject
VAR browseObjectBlock
VAR parentObject
VAR queryGetList
VAR mainGetList
VAR buttonConfig
VAR startBlock
VAR doneBlock
VAR bgColor
VAR staticRow
VAR staticCol
VAR staticWidth
VAR staticHeight
VAR isPixel
VAR isSQL
VAR saySize
VAR isFloating
VAR isVertical
VAR dialogWindow
VAR maxGetWidth

METHOD init, buildGets, applyQuery, testQuery, clearFields, isVisible, destroy

ENDCLASS

* ----------

METHOD DC_BrowseQuery:init( GetList )

::isSQL := .f.
::isFloating := .f.
::isVertical := .f.
::alias := Alias()
::mainGetList := GetList
::saySize := 15
::staticRow := 0
::staticCol := 0
::maxGetWidth := 100

RETURN self

* ----------

METHOD DC_BrowseQuery:buildGets( oBrowse, oStaticBrowse )

LOCAL i, nRow := .5, nCol := 0, dPopUp, nSaveRec := (::alias)->(RecNo()), lData := .F., ;
      nColOffset, nButtonCol, GetList, oStaticQuery, oStaticText, nWidth, GetOptions, ;
      nGetSize

nButtonCol := 1

IF ::isFloating
  GetList := {}
  oStaticBrowse := nil
ELSE
  GetList := ::mainGetList
  ::isVertical := .f.
ENDIF
::queryGetList := GetList

::browseObjectBlock := DC_GetAnchorCb(@oBrowse)

(::alias)->(dbGoto(0))

IF !::isFloating

  @  ::staticRow,  ::staticCol DCSTATIC TYPE XBPSTATIC_TYPE_RECESSEDBOX OBJECT oStaticQuery ;
                SIZE ::staticWidth, ::staticHeight PARENT oStaticBrowse ;
                EVAL {|o|IIF(!::isFloating,o:hide(),nil)}

  @   2,  2 DCSTATIC TYPE XBPSTATIC_TYPE_TEXT OBJECT oStaticText ;
                SIZE DCGUI_PARENTWIDTH-4, DCGUI_PARENTHEIGHT-4 ;
                PARENT oStaticQuery COLOR GRA_CLR_BLACK, ::bgColor ;
                PIXEL

ENDIF

nRow := .5

@ nRow, nButtonCol DCPUSHBUTTON CAPTION 'Find' SIZE 8,1 ;
         ACTION {||::applyQuery(@oBrowse)} PARENT oStaticText GROUP 'QUERYBUTTONS' ;
         WHEN {|| ::testQuery() .OR. lData } ;
         TOOLTIP "Find Records Matching The Selected Search Criteria" ;
         CONFIG ::buttonConfig

IF !::isVertical
  nRow++
ELSE
  nButtonCol += 10
ENDIF

@ nRow, nButtonCol DCPUSHBUTTON CAPTION 'Clear' SIZE 8,1 ;
         ACTION {||::clearFields(@oBrowse)} PARENT oStaticText GROUP 'QUERYBUTTONS' ;
         WHEN {|| ::testQuery() .OR. lData } ;
         TOOLTIP "Reset Search Criteria" ;
         CONFIG ::buttonConfig

IF !::isVertical
  nRow++
ELSE
  nButtonCol += 10
ENDIF

@ nRow, nButtonCol DCPUSHBUTTON CAPTION 'Done' SIZE 8,1 ;
         ACTION {||::clearFields(@oBrowse), ;
                   Eval(::doneBlock)} PARENT oStaticText ;
         TOOLTIP "Close Search Window" ;
         CONFIG ::buttonConfig

IF !::isVertical
  nColOffset := nButtonCol + 10
  nRow := .5
ELSE
  nColOffset := 0
  nRow += 2
ENDIF

nCol := 0
nWidth := 0

DEFAULT ::staticHeight := (Len(::queryArray) / 3 + 1)

FOR i := 1 TO Len(::queryArray)
  ASize(::queryArray[i], QUERY_VAR)
  ::queryArray[i,QUERY_ROW] := nRow
  ::queryArray[i,QUERY_COL] := nCol
  ::queryArray[i,QUERY_GET] := (::alias)->(&(::queryArray[i,QUERY_FIELD]))

  nGetSize := Len(DC_XtoC(::queryArray[i,QUERY_GET]))
  nGetSize := Min(nGetSize,::maxGetWidth)

  nWidth := Max( nWidth, nGetSize )

  nRow++
  IF !::isVertical .AND. nRow > ::staticHeight-1
    nRow := .5
    nCol += ::saySize + nWidth + 2
    nWidth := 0
  ENDIF
NEXT

FOR i := 1 TO Len(::queryArray)

  dPopup := nil
  nGetSize := nil
  IF Valtype(::queryArray[i,QUERY_GET]) == 'D'
    dPopup := {|d|DC_PopDate(d,,,,,,2)}
  ELSEIF Valtype(::queryArray[i,QUERY_GET]) == 'C'
    nGetSize := Len(::queryArray[i,QUERY_GET])
    nGetSize := Min(nGetSize,::maxGetWidth)
  ENDIF

  nRow := IIF( ValType(::queryArray[i,QUERY_ROW]) == 'N', ::queryArray[i,QUERY_ROW], nRow++ )
  nCol := IIF( ValType(::queryArray[i,QUERY_COL]) == 'N', ::queryArray[i,QUERY_COL], 0 ) + nColOffset

  @ nRow++, nCol DCSAY ::queryArray[i,QUERY_SAY] GET ::queryArray[i,QUERY_GET] ;
                 KEYBLOCK {|a,b,o| IIF( Len(AllTrim(o:Get:buffer)) <= 2, ;
                                       ( lData := !Empty(AllTrim(o:Get:buffer)), DC_GetRefresh(::queryGetList,,,,{'QUERYBUTTONS'})), nil )} ;
                 SAYSIZE ::saySize SAYRIGHTBOTTOM TABSTOP PARENT oStaticText POPUP dPopUp ;
                 GETCOLOR GRA_CLR_BLACK, GRA_CLR_WHITE ;
                 GETID 'QUERY_' + Alltrim(Str(i)) ;
                 GETEVAL _GetEvalBlock(::queryArray[i],ATail(::queryGetList)) ;
                 PICTURE ::queryArray[i,QUERY_PICTURE] ;
                 GETSIZE nGetSize

NEXT

DCEVAL {||::browseObject := oBrowse, ;
          ::parentObject := oStaticQuery}

IF ::isFloating

  ::startBlock := {||::dialogWindow:show(), ;
                    ::dialogWindow:toFront(), ;
                    SetAppFocus(DC_GetObject(::queryGetList,'QUERY_1')), ;
                    DC_GetRefresh(::mainGetList)}

  ::doneBlock := {||::dialogWindow:hide(), ;
                    DC_GetRefresh(::mainGetList)}

ELSE

  ::startBlock := {|| IIF( oStaticQuery:isVisible(), nil, ;
                         (oBrowse:setSize( {oBrowse:currentSize()[1],oBrowse:currentSize()[2]-oStaticQuery:currentSize()[2]}), ;
                         oBrowse:setPos( {oBrowse:currentPos()[1],oStaticQuery:currentSize()[2]}),;
                         oStaticQuery:show(), ;
                         DC_GetRefresh(::queryGetList), ;
                         SetAppFocus(DC_GetObject(::queryGetList,'QUERY_1'))))}

  ::doneBlock := {|| oStaticQuery:hide(), ;
                   oBrowse:setSize( oStaticBrowse:currentSize() ), ;
                   oBrowse:setPos( {0,0}), DC_GetRefresh(::queryGetList) }
ENDIF

(::alias)->(dbGoTo(nSaveRec))

IF ::isFloating
  DCGETOPTIONS COLOR ::bgColor HIDE
  DCREAD GUI FIT TITLE 'Browse Query' EXIT ;
      EVAL {|o|::dialogWindow := o, ;
               ::browseObject := oBrowse} ;
      OPTIONS GetOptions SAVE
ENDIF

RETURN nil

* ------------

STATIC FUNCTION _GetEvalBlock( aQueryItem, aGetListItem )

RETURN {|o|aQueryItem[QUERY_VAR] := aGetListItem[bGETLIST_VAR]}

* ------------

METHOD DC_BrowseQuery:isVisible()

IF ::isFloating
  RETURN ::dialogWindow:isVisible()
ENDIF

RETURN ::parentObject:isVisible()

* ------------

METHOD DC_BrowseQuery:destroy()

IF Valtype(::dialogWindow) == 'O'
  ::dialogWindow:destroy()
ENDIF

RETURN self

* ------------

METHOD DC_BrowseQuery:applyQuery( oBrowse )

LOCAL i, oDlg, aRecords[0], lAdd, nCnt, nPct, nLastRec, oProgress, GetList[0], GetOptions, ;
      xScope, lCancelled := .f.

IF !::testQuery()
  DC_WinAlert('No Search Criteria Entered!!')
  RETURN nil
ENDIF

nLastRec := (::alias)->(LastRec())
nPct := Round(nLastRec/100,0)
nCnt := 0

@ 0,0 DCPROGRESS oProgress RADIUS 20 DYNAMIC OUTLINE COLOR GRA_CLR_CYAN, GRA_CLR_WHITE SIZE 50,1 PERCENT
@ DCGUI_ROW, DCGUI_COL + 10 DCPUSHBUTTON CAPTION 'Cancel' SIZE 10,1 CONFIG ::buttonConfig ;
   ACTION {||lCancelled := .t.}

DCGETOPTIONS COLOR GRA_CLR_WHITE
DCREAD GUI FIT TITLE 'Searching...' EXIT PARENT @oDlg OPTIONS GetOptions

DC_PointerWait()
(::alias)->(DC_ClrScope())
(::alias)->(DC_SetScopeArray(nil))

FOR i := 1 TO Len(::queryArray)
  IF !Empty(::queryArray[i,QUERY_GET]) .AND. Upper(DC_XtoC(::queryArray[i,QUERY_INDEX])) == (::alias)->(OrdSetFocus())
    IF Valtype(::queryArray[i,QUERY_GET]) = 'C'
      xScope := Trim(::queryArray[i,QUERY_GET])
    ELSEIF Valtype(::queryArray[i,QUERY_GET]) = 'D'
      xScope := Dtos(::queryArray[i,QUERY_GET])
    ELSE
      xScope := ::queryArray[i,QUERY_GET]
    ENDIF
    (::alias)->(DC_SetScope(0,xScope))
    (::alias)->(DC_SetScope(1,xScope))
    EXIT
  ENDIF
NEXT

(::alias)->(DC_dbGoTop())

DO WHILE !(::alias)->(Eof()) .AND. !lCancelled
  IF nCnt % 100 == 0
    DC_CompleteEvents()
  ENDIF
  IF ++nCnt%nPct == 0
    DC_GetProgress( oProgress, nCnt, nLastRec )
  ENDIF
  lAdd := .f.
  FOR i := 1 TO Len(::queryArray)
    IF !Empty(::queryArray[i,QUERY_GET])
      IF Valtype(::queryArray[i,QUERY_GET]) = 'C' .AND. Upper((::alias)->(&(::queryArray[i,QUERY_FIELD]))) = Upper(Trim(::queryArray[i,QUERY_GET]))
        lAdd := .t.
      ELSEIF Valtype(::queryArray[i,QUERY_GET]) = 'D' .AND. DTOS((::alias)->(&(::queryArray[i,QUERY_FIELD]))) = DTOS(::queryArray[i,QUERY_GET])
        lAdd := .t.
      ELSEIF (::alias)->(&(::queryArray[i,QUERY_FIELD])) == ::queryArray[i,QUERY_GET]
        lAdd := .t.
      ELSE
        lAdd := .f.
        EXIT
      ENDIF
    ENDIF
  NEXT
  IF lAdd
    AAdd(aRecords,(::alias)->(RecNo()))
  ENDIF
  (::alias)->(DC_DbSkip())
ENDDO
oDlg:destroy()

IF Empty(aRecords)
  aRecords := {-1 }
ENDIF

(::alias)->(DC_ClrScope())
IF lCancelled
  (::alias)->(DC_SetScopeArray(nil))
ELSE
  (::alias)->(DC_SetScopeArray(aRecords))
ENDIF
(::alias)->(DC_DbGoTop())

oBrowse:refreshAll()
DC_PointerArrow()

IF aRecords[1] = -1 .AND. !lCancelled
  DCMSGBOX "No Matching Records Were Found" TITLE "Data Alert" TIMEOUT 3
ENDIF

RETURN nil

* -----------

METHOD DC_BrowseQuery:testQuery()

LOCAL i, lStatus := .f.

FOR i := 1 TO Len(::queryArray)
  IF !Empty(::queryArray[i,QUERY_GET])
    lStatus := .t.
    EXIT
  ENDIF
NEXT

RETURN lStatus

* -----------

METHOD DC_BrowseQuery:clearFields( oBrowse )

LOCAL i

(::alias)->(dbGoTo(0))

FOR i := 1 TO Len(::queryArray)
  ::queryArray[i,QUERY_GET] := (::alias)->(&(::queryArray[i,QUERY_FIELD]))
NEXT

(::alias)->(DC_SetScopeArray(nil))
(::alias)->(DC_DbGoTop())

DC_GetRefresh(::queryGetList)
DC_GetRefresh(::mainGetList)
// oBrowse:refreshAll()

RETURN nil

* ------------

GETSETFUNCTION DC_BrowseTrapErrors DEFAULT .f.

* ------------

FUNCTION DC_BrowseConfigMessageBlock( bMsg )

STATIC sbMsg
LOCAL bMsgSave

bMsgSave := sbMsg

IF Valtype(bMsg) == 'B'
  sbMsg := bMsg
ENDIF

RETURN bMsgSave

/*

This is a GENERIC owner drawing system for DC_XbpBrowse
Each browse cell can draw text, icons or bitmaps anywhere
within the cell by declaring each within a multi-dimensional
array

  [1] = <ocData>
  [2] = <nCol>
  [3] = <nRow>
  [4] = <nWidth>
  [5] = <nHeight>
  [6] = <font>
  [7] = <selectedFont>
  [8] = <fgColor>
  [9] = <bgColor>
 [10] = <nAlign>
 [11] = <cargo>

*/

#if XPPVER > 1900331
CLASS DC_XbpOwnerDrawBrowseGeneric FROM DC_XbpBrowse

PROTECTED:
   VAR oIcon, oBitmap, aLineAttrs, aBoxAttrs, oStdFont, oBoldFont, oBoldItalicFont, ;
       oFont, oSelectedFont

EXPORTED:

VAR lOwnerDraw, nLineHeight, nCellHeight
VAR stdFontCompoundName, boldFontCompoundName, boldItalicFontCompoundName

METHOD CustomDrawCell

INLINE METHOD destroy
   ::DC_XbpBrowse:destroy()
RETURN self

INLINE METHOD init( oParent, oOwner, aPos, aSize, aPP, lVisible, oGetList )

::drawMode:= XBP_DRAW_OWNER
::oIcon := XbpIcon():New():Create()
::oBitmap := XbpBitmap():New():Create()
::nLineHeight := 17.5
::nCellHeight := 20
::lOwnerDraw := .t.
::aLineAttrs := ARRAY( GRA_AL_COUNT )
::aBoxAttrs  := ARRAY( GRA_AA_COUNT )
::aLineAttrs [ GRA_AL_COLOR ] := GRA_CLR_BLACK
::aLineAttrs [ GRA_AL_TYPE ]  := GRA_LINETYPE_SOLID
::stdFontCompoundName := '10.Tahoma'
::boldFontCompoundName := '10.Tahoma '+FONTBOLD
::boldItalicFontCompoundName := '10.Tahoma '+FONTBOLD+' '+FONTITALIC
::DC_XbpBrowse:init( oParent, oOwner, aPos, aSize, aPP, lVisible, oGetList )

RETURN self

INLINE METHOD create( oParent, oOwner, aPos, aSize, aPP, lVisible )

LOCAL nFound

::oStdFont := XbpFont():New():Create(::stdFontCompoundName)
::oBoldFont := XbpFont():New():Create(::boldFontCompoundName)
::oBoldItalicFont := XbpFont():New():Create(::boldItalicFontCompoundName)
::oSelectedFont := ::oBoldItalicFont
::oFont := XbpFont():New():Create()
::DC_XbpBrowse:create( oParent, oOwner, aPos, aSize, aPP, lVisible )

IF Empty(aPP)
  aPP := ::setPresParam()
ENDIF

nFound := AScan(aPP,{|a|Valtype(a[1])=='N'.AND. ;
                  a[1]==XBP_PP_COL_DA_CELLHEIGHT } )
IF nFound > 0
  ::nCellHeight := aPP[nFound,2]
ENDIF

RETURN self

ENDCLASS

* ----------

METHOD DC_XbpOwnerDrawBrowseGeneric:customDrawCell( oPS, aInfo )

LOCAL aData, aPP, nFound, nAlign, nLine, nFgColor := GRA_CLR_BLACK, ;
      nBgColor := GRA_CLR_WHITE, oBitmap, aSource, aTarget, ;
      aIcons, nIcons, nHeight, nWidth, nLineHeight, cCaption, ;
      nStartRow, nStartCol, nEndRow, nEndCol, i, aLine, ocFont, ;
      oOldFont, lDisabled := .f., lIcon := .f., lBitmap := .f., ;
      lSelected, cFont, cSelectedFont, nCellHeight, aChildList, ;
      bError, aPoints, mp1, mp2, oXbp, nEvent

IF !::lOwnerDraw // ownerdrawing is switched off, let the system draw
  RETURN .T.
ENDIF

BEGIN SEQUENCE

bError := ErrorBlock({|e|Break(e)})

* get column data
aData := aInfo[ XBP_DRAWINFO_AREA ]:getCell( aInfo[ XBP_DRAWINFO_ITEM ] )

lSelected := BAnd(aInfo[XBP_DRAWINFO_STATE],XBP_DRAWSTATE_SELECTED) # 0 // cell is selected

* get cell alignment
aPP := aInfo[ XBP_DRAWINFO_COLUMN ]:presArray
nFound := AScan(aPP,{|a|Valtype(a[1])=='N'.AND. ;
                 (a[1]==XBPCOL_DA_CELLALIGNMENT .OR. ;
                  a[1]==XBP_PP_COL_DA_CELLALIGNMENT) } )

IF nFound > 0
  nAlign := aPP[nFound,2]
ELSE
  nAlign := XBPALIGN_VCENTER
ENDIF

* get cell height
nLineHeight := ;
   Int((aInfo[ XBP_DRAWINFO_RECT, 4 ]-aInfo[ XBP_DRAWINFO_RECT, 2 ])/(::nCellHeight/::nLineHeight))

* get cell colors
aInfo[ XBP_DRAWINFO_AREA ]:getCellColor( aInfo[ XBP_DRAWINFO_ITEM ],@nFgColor, @nBgColor )

IF nFgColor # nil
  oPS:setColor( nFgColor, nBgColor )
ENDIF

IF Valtype(aData) == 'C'
  * no line break, just paint the text
  GraCaptionStr( oPS, aInfo[ XBP_DRAWINFO_RECT ], ;
                     { aInfo[ XBP_DRAWINFO_RECT, 3 ], ;
                       aInfo[ XBP_DRAWINFO_RECT, 4 ] }, aData, nAlign )
ENDIF

IF Valtype(aData) # 'A'
  BREAK
ENDIF

FOR i := 1 TO Len(aData)

  aLine := aData[i]
  IF Valtype(aLine) # 'A'
    LOOP
  ENDIF
  ASize(aLine,11)

  /*
  [1] = <ocData>
  [2] = <nCol>
  [3] = <nRow>
  [4] = <nWidth>
  [5] = <nHeight>
  [6] = <font>
  [7] = <selectedFont>
  [8] = <fgColor>
  [9] = <bgColor>
 [10] = <nAlign>
 [11] = <cargo>
  */

  lIcon := .f.
  lBitmap := .f.
  ocFont := nil
  cCaption := aLine[1]
  nStartCol := aLine[2]
  IF Valtype(nStartCol) # 'N'
    nStartCol := aInfo[ XBP_DRAWINFO_RECT,1 ] + 5
  ENDIF
  nStartRow := aLine[3]
  IF Valtype(nStartRow) # 'N'
    nStartRow := aInfo[ XBP_DRAWINFO_RECT, 4 ]-(nLineHeight*(i-1)) - 5
  ELSE
    nStartRow := aInfo[ XBP_DRAWINFO_RECT, 4 ]-nStartRow
  ENDIF
  cFont := aLine[6]
  cSelectedFont := aLine[7]
  IF Empty(cSelectedFont) .AND. !Empty(cFont)
    cSelectedFont := cFont
  ENDIF
  IF Empty(cFont)
    ocFont := ::oStdFont
  ELSEIF Valtype(cFont) == 'C' .AND. !lSelected
    ocFont := ::oFont
    ocFont:configure(cFont)
  ELSEIF lSelected .AND. Valtype(cSelectedFont) == 'C'
    ocFont := ::oFont
    ocFont:configure(cSelectedFont)
  ENDIF

  nFGColor := aLine[8]
  nBGColor := aLine[9]
  nAlign := aLine[10]

  IF Valtype(nFGColor) == 'A'
    nFGColor := GraMakeRGBColor(nFGColor)
  ENDIF

  IF Valtype(nBGColor) == 'A'
    nBGColor := GraMakeRGBColor(nBGColor)
  ENDIF

  IF Valtype(nFGColor) == 'N' .OR. nBGColor == 'N'
    GraSetColor( oPS, nFGColor, nBGColor )
  ENDIF

  IF Valtype(cCaption) == 'N'
    oBitmap := DC_GetBitmap(cCaption)
    IF Valtype(oBitmap) == 'O' .AND. oBitmap:isDerivedFrom('XbpBitmap')
      ::oBitmap := oBitmap
      lBitmap := .t.
    ELSEIF Valtype(oBitmap) == 'O' .AND. oBitmap:isDerivedFrom('XbpIcon')
      ::oIcon := oBitmap
      lIcon := .t.
    ENDIF
  ELSEIF Right(Upper(cCaption),4) == '.ICO'
    ::oIcon:loadFile(cCaption,nLineHeight,nLineHeight)
    lIcon := .t.
  ELSEIF Right(Upper(cCaption),4) $ {'.BMP','.JPG', '.PNG', '.GIF'}
    ::oBitmap:loadFile(cCaption)
    oBitmap := ::oBitmap
    lBitmap := .t.
  ENDIF

  nWidth := aLine[4]
  IF Valtype(nWidth) # 'N'
    IF !lIcon .AND. !lBitmap
      nEndCol := aInfo[ XBP_DRAWINFO_RECT, 3 ]
    ELSEIF lBitmap
      nEndCol := nStartCol + oBitmap:xSize
    ENDIF
  ELSE
    nEndCol := nStartCol + nWidth
  ENDIF

  nHeight := aLine[5]
  IF Valtype(nHeight) # 'N'
    IF !lIcon .AND. !lBitmap
      aPoints := GraQueryTextBox( oPS, 'Wxxxx' )
      nHeight :=  aPoints[1,2] - aPoints[2,2]
      nStartRow -= nHeight
      nEndRow := nStartRow + nHeight
    ELSEIF lBitmap
      nHeight := oBitmap:ySize
      nEndRow := nStartRow - nHeight
    ELSE
      nEndRow := aInfo[ XBP_DRAWINFO_RECT, 4 ]-(nLineHeight*i)
    ENDIF
  ELSE
    nStartRow -= nHeight
    nEndRow := nStartRow + nHeight
  ENDIF

  IF Valtype(ocFont) == 'O' .AND. !lIcon .AND. !lBitmap
    oOldFont := GraSetFont( oPS, ocFont )
  ENDIF

  IF lIcon

    aTarget := { nStartCol, nStartRow, nEndCol, nEndRow }
    ::oIcon:Draw( oPS, aTarget, XBP_STATE_NORMAL )

  ELSEIF lBitmap

    aSource := {0,0,oBitmap:xSize,oBitmap:ySize}

    aTarget := { nStartCol, nStartRow, nEndCol, nEndRow }

 //   ::oBitmap:draw( oPS, aTarget, aSource, GRA_BLT_ROP_SRCCOPY,, XBP_STATE_NORMAL )
    ::oBitmap:draw( oPS, aTarget, aSource, GRA_BLT_ROP_SRCCOPY,4,XBP_STATE_NORMAL ) // makes bitmaps look better


  ELSEIF Valtype( cCaption ) == 'C'

    GraCaptionStr( oPS, { nStartCol, nStartRow }, ;
                        { nEndCol, nEndRow }, cCaption, nAlign )

  ENDIF

  IF Valtype(oOldFont) = 'O'
    GraSetFont( oPS, oOldFont )
  ENDIF

  IF !lBitMap .AND. !lIcon
    GraSetColor( oPS, nFGColor )
  ENDIF

NEXT

END SEQUENCE

ErrorBlock(bError)

RETURN .F.

#endif

* ----------

CLASS DC_XbpBrowseFiltered FROM DC_XbpBrowse

EXPORTED:

VAR filterTimeOut
VAR filterStartTime
VAR filterSqlStatement
VAR filterRefreshBlock
VAR filterExpression
VAR filterScope
VAR orderColumn
VAR columnNames
VAR scopeWhere
VAR isPreserveScope
VAR AdsTableType
VAR AdsLockingType
VAR IndexTags
VAR AdsStatement
VAR nPrintOrientation
VAR nPrintFudge
VAR lPrintGrid
VAR sqlExecuteTime
VAR isQuoted

METHOD Init(), SetFilter(), ProcessFilterArray(), ProcessFilterWorkArea(), BuildSQL(), ;
       ExecuteSQL(), GetSqlStatement(), OrderBy(), BuildColumnNamesArray(), ;
       ScopeToWhere(), IsGoodIndex(), ResetFilter(), PrintBrowse(), Browse2Excel(), ;
       BuildFilter(), TestFilterValidity()

ENDCLASS

* --------

METHOD DC_XbpBrowseFiltered:init( oParent, oOwner, aPos, aSize, aPres, lVisible, oGetList )

::DC_XbaseParts:init( oGetList,,self )

::DC_XbpBrowse:Init( oParent, oOwner, aPos, aSize, aPres, lVisible, oGetList )

::footerLines := 2
::hasFooter := .t.
::filterTimeOut := 5
::filterStartTime := 0
::filterSqlStatement := ::GetSqlStatement()
::columnNames := ::buildColumnNamesArray()
::isPreserveScope := .f.
::AdsTableType := DC_AdsIndexModeDefault()
::AdsLockingType := DC_AdsLockModeDefault()
::nPrintOrientation := 2
::nPrintFudge := 3
::lPrintGrid := .f.
::filterExpression := ''
::SqlExecuteTime := 0
::isQuoted := .f.

RETURN self

* --------

METHOD DC_XbpBrowseFiltered:buildColumnNamesArray( cSql )

LOCAL i, nFound, aFields[0], cOrigField, cAsField

IF Empty(cSql)
  cSql := Alltrim(::filterSqlStatement)
ENDIF

cSql := Alltrim(Substr( cSql, 8 ))
IF Left(cSql,1) == '*'
  RETURN nil
ENDIF

cSql := Strtran(cSql,Chr(13),' ')
cSql := Strtran(cSql,Chr(10),' ')
nFound := At(' FROM ', Upper(cSql) )
cSql := Substr( cSql, 1, nFound )
aFields := DC_TokenArray( cSql, ',' )

FOR i := 1 TO Len(aFields)
  aFields[i] := Alltrim(aFields[i])
  IF ' ' $ aFields[i]
    cOrigField := Substr(aFields[i],1,At(' ',aFields[i])-1)
    cAsField := Substr(aFields[i],_AtLast(aFields[i],' ')+1)
  ELSE
    cOrigField := aFields[i]
    cAsField := aFields[i]
  ENDIF
  aFields[i] := { cOrigField, cAsField }
NEXT

RETURN aFields

* --------

STATIC FUNCTION _AtLast( cField, cChar )

LOCAL i

FOR i := Len(cField) TO 1 STEP -1
  IF cField[i] == cChar
    RETURN i
  ENDIF
NEXT

RETURN 0

* --------

METHOD DC_XbpBrowseFiltered:setFilter( nOrderColumn, cWhere )

LOCAL i, j, lError, lGoodRec, aColumns[0], cSaveSql, cSaveAlias := ::dataSource, ;
      cSQLStatement, cTagName, cFieldName, lStatus := .t., lBuildSql := .t., ;
      xScope1, xScope2, cScopeTagName, cOrder

cSaveSql := ::filterStatement
lError := .t.
DC_PointerWait()
::refresh()
FOR i := 1 TO ::colCount
  AAdd(aColumns,::getColumn(i))
NEXT

IF Valtype(::dataSource) == 'A'

  ::displayedRows := 0
  FOR i := 1 TO Len(::dataSource)
    ::displayedRows++
    FOR j := 1 TO Len(aColumns)
      IF Empty(aColumns[j]:filterBlock)
        LOOP
      ELSEIF !Eval(aColumns[j]:filterBlock,::datasource[i])
        ::displayedRows--
        EXIT
      ENDIF
    NEXT
  NEXT
  IF ::displayedRows == 0
    DC_PointerArrow()
    DCMSGBOX 'No Records match your search criteria' FONT '12.Arial Bold'
    RETURN .f.
  ENDIF
  ::filter := {|o|::ProcessFilterArray(o,aColumns)}
  lBuildSql := .f.

ELSEIF Valtype(::dataSource) == 'C' .AND. !Empty(nOrderColumn) // .AND. Empty(::sqlStatementOriginal) ;

  cFieldname := aColumns[nOrderColumn]:sqlField
  IF Empty(cFieldname)
    cFieldName := aColumns[nOrderColumn]:headerBlock
  ENDIF
  IF Valtype(cFieldName) == 'C'
    cTagName := ::findGoodIndex(Alltrim(Upper(cFieldName)))
    IF !Empty(cTagName)
      (::dataSource)->(OrdSetFocus(cTagName))
      lBuildSql := .f.
    ELSE
      lBuildSql := .t.
    ENDIF
  ENDIF

ENDIF

BEGIN SEQUENCE

IF lBuildSql

  IF (Empty(::sqlSessionDict) .AND. Empty(::sqlSessionFree)) // .OR. !::TestFilterValidity()
    ::filterExpression := ::BuildFilter(aColumns,@nOrderColumn,@xScope1,@xScope2,@cScopeTagName)

    IF !Empty(::filterExpression)
      (::DataSource)->(DC_SetScopeArray(nil))
      (::DataSource)->(dbSetFilter(&('{||' + ::filterExpression + '}')),::filterExpression)
      lStatus := .t.
    ELSE
      (::DataSource)->(dbClearFilter())
      IF Empty(::saveScope0)
        (::DataSource)->(DC_ClrScope())
      ELSE
        (::DataSource)->(DC_SetScope(0,::saveScope0))
        (::DataSource)->(DC_SetScope(1,::saveScope1))
      ENDIF
    ENDIF

    /*
    IF !Empty(xScope1) .AND. !Empty(cScopeTagName)
      ::orderColumn := nOrderColumn
      aColumns[nOrderColumn]:Sort(nil,nOrderColumn)
      (::DataSource)->(OrdSetFocus(cScopeTagName))
      (::DataSource)->(DC_SetScope(0,xScope1))
      (::DataSource)->(DC_SetScope(1,xScope2))
    ENDIF
    */
    lStatus := .t.
    BREAK
  ENDIF

  ::filterSqlStatement := ::BuildSQL(aColumns,nOrderColumn,cWhere)

  // cOrder := OrdSetFocus()
  IF !Empty(::filterSqlStatement)
    IF (lStatus := ::ExecuteSQL(cSaveAlias))
      ::dataSource := Alias()
    ELSEIF !Empty(cSaveSQL)
      ::filterSqlStatement := cSaveSQL
      lStatus := ::ExecuteSQL(cSaveAlias)
    ENDIF
  ENDIF
 // IF !Empty(cOrder)
 //   OrdSetFocus(cOrder)
 // ENDIF
  ::sortColumn := 0

  IF !lStatus .AND. Empty(::sqlSessionDict) .AND. Empty(::sqlSessionFree)
    DC_WinAlert('No SQL connection Available to sort or filter this column')
  ENDIF

ENDIF

END SEQUENCE

IF lStatus
  ::goTop()
  ::forceStable()
  ::refreshAll()
ENDIF

IF Valtype(::filterRefreshBlock) == 'B'
  Eval(::filterRefreshBlock,self)
ENDIF

DC_PointerArrow()

RETURN lStatus

* --------

METHOD DC_XbpBrowseFiltered:testFilterValidity()

LOCAL cSQL, lStatus := .t.

cSql := ::sqlStatementOriginal

IF 'ORDER BY' $ Upper(::sqlStatementOriginal)
  lStatus := .f.
ENDIF

RETURN lStatus

* --------

METHOD DC_XbpBrowseFiltered:processFilterArray(oRecord,aColumns)

LOCAL i

FOR i := 1 TO Len(aColumns)
  IF Empty(aColumns[i]:filterBlock)
    LOOP
  ENDIF

  IF !Eval(aColumns[i]:filterBlock,oRecord)
    RETURN .f.
  ENDIF
NEXT

RETURN .t.

* --------

METHOD DC_XbpBrowseFiltered:processFilterWorkArea(aColumns)

LOCAL i

IF .F.// ::filterTimeOut > 0 .AND. ::filterStartTime > 0 .AND. (Seconds() - ::filterStartTime) > ::filterTimeOut
  DCMSGBOX 'Filter has Timed Out! Showing All records.'
  dbClearFilter()
  ::filterStartTime := 0
  dbGoTop()
  ::refreshAll()
  RETURN .t.
ENDIF

FOR i := 1 TO Len(aColumns)
  IF Empty(aColumns[i]:filterBlock)
    LOOP
  ENDIF
  IF !Eval(aColumns[i]:filterBlock)
    RETURN .f.
  ENDIF
NEXT

RETURN .t.

* ----------

METHOD DC_XbpBrowseFiltered:resetFilter()

LOCAL cStatement := ::sqlStatementOriginal, oColumn, i, aData

FOR i := 1 TO ::colCount
  oColumn := ::getColumn(i)
  IF oColumn:isDerivedFrom('DC_XbpColumnFiltered')
    oColumn:resetFilter()
  ENDIF
NEXT

RETURN ::setFilter()

* ----------

METHOD DC_XbpBrowseFiltered:BuildSQL(aColumns,nOrderColumn,cWhere)

LOCAL cSQL, i, cFieldName, cFieldType, cDate, cValue, ;
      xFilterVal1, xFilterVal2, nFilterType, cDbfName, cLen1, cLen2, ;
      cPrefix := '', lNoAnd := .f., cOrderBy := '', nFound, lHasFields, ;
      nUnixTime1, nUnixTime2, cScopeWhere := '', cCaption, cSqlFieldName, ;
      lFilterExact

DEFAULT cWhere := ''

IF Empty(nOrderColumn)
  nOrderColumn := ::orderColumn
ENDIF

IF Empty(::sqlStatementOriginal)
  ::sqlStatementOriginal := (::dataSource)->(Upper(dbInfo( DBO_FILENAME )))
  IF !Upper(Substr(::sqlStatementOriginal,1,6)) == 'SELECT'
    ::sqlStatementOriginal := ''
  ENDIF
ENDIF

IF !Empty(::sqlStatementOriginal)
  cSQL := ::sqlStatementOriginal
  IF 'ORDER BY' $ Upper(cSQL)
    nFound := At('ORDER BY',Upper(cSQL))
    cOrderBy := Substr(cSQL,nFound)
    cSQL := Substr(cSQL,1,nFound-1)
  ENDIF
ELSE
  DEFAULT ::sqlTableName := Alias()
  cSQL := 'SELECT *'
ENDIF

/*
IF !Empty(::columnNames) .AND. ::columnNames[1,1] # '*'
  FOR i := 1 TO Len(aColumns)
    IF !IsMemberVar(aColumns[i],'struArray')
      aColumns[i]:struArray := ''
      AAdd(::columnNames,{'',''})
      Ains(::columnNames,i)
    ENDIF
  NEXT
ENDIF
*/

lHasFields := !Empty(cWhere)

FOR i := 1 TO Len(aColumns)
  IF !Empty(aColumns[i]:heading:childList())
    cCaption := aColumns[i]:heading:childList()[1]:caption
    /*
    IF !Empty(::columnNames)
      nFound := Ascan(::columnNames,{|a|Upper(a[2]) == Upper(cCaption)})
      IF nFound > 0
        cSqlFieldName := ::columnNames[nFound,1]
      ELSE
        cSqlFieldName := ''
      ENDIF
    ENDIF
    */
  ELSE
    cCaption := aColumns[i]:headerCaption
  ENDIF
  IF aColumns[i]:filterType#3 .AND. ;
      (!IsMemberVar(aColumns[i],'isdatefield') .OR. ;
      ( Empty(aColumns[i]:filterVal1) .AND. Empty(aColumns[i]:filterVal2) .AND. Empty(aColumns[i]:filterVal) .AND. Valtype(aColumns[i]:filterVal1)#'L'))
    LOOP
  ELSEIF !Empty(aColumns[i]:sqlField)
    cFieldName := aColumns[i]:sqlField
  ELSEIF !Empty(cSqlFieldname)
    cFieldName := cSqlFieldName
  ELSEIF !Empty(cCaption)
    cFieldName := cCaption
  ENDIF

  IF ::isQuoted .AND. !('"' $ cFieldName .AND. '"' $ cFieldName)
    cFieldName := '"' + cFieldName + '"'
  ENDIF

  xFilterVal1 := nil
  IF !Empty(aColumns[i]:filterVal)
    xFilterVal1 := aColumns[i]:filterVal
  ELSEIF !Empty(aColumns[i]:filterVal1)
    xFilterVal1 := aColumns[i]:filterVal1
  ELSEIF Valtype(aColumns[i]:filterVal1) == 'L'
    xFilterVal1 := aColumns[i]:filterVal1
  ENDIF
  xFilterVal2 := aColumns[i]:filterVal2
  lFilterExact := aColumns[i]:filterExact

  nFilterType := aColumns[i]:filterType

  IF nFilterType # 3 .AND. Empty(xFilterVal1) .AND. Empty(xFilterVal2) .AND. Valtype(xFilterVal1) # 'L'
    LOOP
  ENDIF

  IF nFilterType == 3
    lHasFields := .t.
  ENDIF

  cFieldType := Valtype(xFilterVal1)

  IF nFilterType == 3 .OR. (cFieldType # 'U' .AND. (!Empty(xFilterVal1) .OR. cFieldType = 'L'))
    lHasFields := .t.
    IF !Empty(cWhere)
      cWhere += ' AND '
    ENDIF
  ENDIF

  IF nFilterType = 3 // empty
     cWhere += " Empty(" + cFieldName + ")"
  ELSEIF cFieldType = 'C' .AND. !Empty(xFilterVal1) .AND. !Empty(xFilterVal2) .AND. !(xFilterVal1==xFilterVal2)
    IF nFilterType = 1
      cLen1 := Alltrim(Str(Len(Alltrim(xFilterVal1))))
      cLen2 := Alltrim(Str(Len(Alltrim(xFilterVal2))))
      IF Valtype(::SqlSessionDict) == 'O' .AND. 'PGDBE' $ ::SqlSessionDict:getConnectionString()
        cWhere += " Substring(Upper(" + cFieldName + "),1," + cLen1 + ") >= '" + Upper(Alltrim(xFilterVal1)) + "' AND " + ;
                  " Substring(Upper(" + cFieldName + "),1," + cLen2 + ") <= '" + Upper(Alltrim(xFilterVal2)) + "'"
      ELSE
        cWhere += " Left(Upper(" + cFieldName + ")," + cLen1 + ") >= '" + Upper(Alltrim(xFilterVal1)) + "' AND " + ;
                  " Left(Upper(" + cFieldName + ")," + cLen2 + ") <= '" + Upper(Alltrim(xFilterVal2)) + "'"
      ENDIF
    ELSE // contained in
      cWhere += " (Upper(" + cFieldName + ") LIKE '%" + Upper(Alltrim(xFilterVal1)) + "%' OR " + ;
                " Upper(" + cFieldName + ") LIKE '%" + Upper(Alltrim(xFilterVal2)) + "%')"
    ENDIF
  ELSEIF cFieldType = 'C' .AND. !Empty(xFilterVal1)
    IF nFilterType == 1
      cLen1 := Alltrim(Str(Len(Alltrim(xFilterVal1))))
      IF Valtype(::SqlSessionDict) == 'O' .AND. 'PGDBE' $ ::SqlSessionDict:getConnectionString()
        cWhere += " Substring(Upper(" + cFieldName + "),1," + cLen1 + ") = '" + Upper(Alltrim(xFilterVal1)) + "'"
      ELSEIF lFilterExact
        cWhere += " Upper(" + cFieldName + ") = '" + Upper(Alltrim(xFilterVal1)) + "'"
      ELSE
        cWhere += " Left(Upper(" + cFieldName + ")," + cLen1 + ") = '" + Upper(Alltrim(xFilterVal1)) + "'"
      ENDIF
    ELSE
      cWhere += " Upper(" + cFieldName + ") LIKE '%" + Upper(Alltrim(xFilterVal1)) + "%'"
    ENDIF
  ELSEIF aColumns[i]:isDateField .AND. !Empty(xFilterVal1) .AND. !Empty(xFilterVal2) .AND. !(xFilterVal1==xFilterVal2)
    nUnixTime1 := DC_DateTime2UnixTime( xFilterVal1 )
    nUnixTime2 := DC_DateTime2UnixTime( xFilterVal2 )
    cWhere += cFieldName + '>=' + DC_XtoC(nUnixTime1) + ' AND ' + ;
              cFieldName + '<=' + DC_XtoC(nUnixTime2)
  ELSEIF aColumns[i]:isDateField .AND. !Empty(xFilterVal1)
    nUnixTime1 := DC_DateTime2UnixTime( xFilterVal1 )
    cWhere += cFieldName + '=' + DC_XtoC(nUnixTime1)
  ELSEIF cFieldType = 'N' .AND. !Empty(xFilterVal1) .AND. !Empty(xFilterVal2) .AND. !(xFilterVal1==xFilterVal2)
    cWhere += cFieldName + '>=' + Alltrim(DC_XtoC(xFilterVal1)) + ' AND ' + ;
              cFieldName + '<=' + Alltrim(DC_XtoC(xFilterVal2))
  ELSEIF cFieldType = 'N' .AND. !Empty(xFilterVal1)
    cWhere += cFieldName + '=' + Alltrim(DC_XtoC(xFilterVal1))
  ELSEIF cFieldType = 'D' .AND. !Empty(xFilterVal1) .AND. !Empty(xFilterVal2)
    cDate := DC_Var2Sql(xFilterVal1)
    cWhere += cFieldname + '>=' +  cDate
    cDate := DC_Var2Sql(xFilterVal2)
    cWhere += ' AND ' + cFieldname + '<=' + cDate
  ELSEIF cFieldType = 'D' .AND. !Empty(xFilterVal1)
    cDate := DC_Var2Sql(xFilterVal1)
    cWhere += cFieldname + '=' + cDate
  ELSEIF cFieldType = 'L'
    IF xFilterVal1
      cValue := 'TRUE'
    ELSE
      cValue := 'FALSE'
    ENDIF
    cWhere += cFieldName + '=' + cValue
  ELSE
    LOOP
  ENDIF
NEXT

IF Empty(::sqlStatementOriginal)
  cSQL += ' FROM ' + ::sqlTableName
ENDIF

IF !lHasFields
  cWhere := ''
ENDIF

IF !Empty(nOrderColumn)
  IF !Empty(aColumns[nOrderColumn]:sqlField)
    cFieldName := aColumns[nOrderColumn]:sqlField
  ELSEIF !Empty(aColumns[nOrderColumn]:heading:childList())
    cFieldName := aColumns[nOrderColumn]:heading:childList()[1]:caption
  ENDIF
  cOrderBy := 'ORDER BY ' + IIF(::isQuoted,'"','') + cFieldName + IIF(::isQuoted,'"','')
  ::orderColumn := nOrderColumn
ENDIF

IF lHasFields
  IF 'WHERE ' $ Upper(::sqlStatementOriginal)
    cWhere := ' AND ' + cWhere
  ELSE
    cWhere := ' WHERE ' + cWhere
  ENDIF
ENDIF

// cScopeWhere := ::ScopeToWhere()

IF !Empty(cScopeWhere)
  IF Empty(cWhere)
    cWhere := 'WHERE ' + cScopeWhere
  ELSE
    cWhere += ' AND ' + cScopeWhere
  ENDIF
ENDIF

IF !lHasFields .AND. !Empty(::sqlStatementOriginal) .AND. Empty(nOrderColumn)
  cSQL := ::sqlStatementOriginal
ELSE
  cSQL += cWhere + ' ' + cOrderBy
ENDIF

RETURN cSQL

* ----------

METHOD DC_XbpBrowseFiltered:ExecuteSQL(cAlias)

LOCAL oStatement, cStatement, lAds, oAdsStatement, nSeconds := Seconds()

SELECT 0

::sqlExecuteTime := 0

cStatement := ::filterSqlStatement

dbSetNullValue(.f.)
SET NULLVALUE OFF

IF Valtype(::SqlSessionDict) == 'B'

  RETURN Eval(::SqlSessionDict,cStatement,cAlias)

ELSEIF  Valtype(::SqlSessionDict) # 'O'

  RETURN .f.

ENDIF

lAds := 'ADSDBE' $ ::SqlSessionDict:getConnectionString()

IF lAds

  dbeSetDefault('ADSDBE')
  oAdsStatement := DC_AdsStatement():New(cStatement,::SqlSessionDict,::AdsLockingType,::AdsTableType)

  IF oAdsStatement:LastError > 0
    RETURN .f.
  ENDIF

  IF DC_DbSel(cAlias)
    IF Valtype(::AdsStatement) == 'O'
      ::AdsStatement:close()
      ::AdsStatement := nil
    ELSE
      (cAlias)->(dbCloseArea())
    ENDIF
    Sleep(50)
  ENDIF

  ::AdsStatement := oAdsStatement
  ::AdsStatement:sessionFree := ::SqlSessionFree
  cAlias := ::AdsStatement:Execute( cAlias )

#if XPPVER > 2000000
ELSEIF 'ODBCDBE' $ ::SqlSessionDict:getConnectionString()

  IF DC_DbSel(cAlias)
    (cAlias)->(dbCloseArea())
  ENDIF

  dbeSetDefault('ODBCDBE')
  ODBC_SqlStmtExec(cStatement,::SqlSessionDict)
  cAlias := Alias()

ELSEIF 'PGDBE' $ ::SqlSessionDict:getConnectionString()

  IF DC_DbSel(cAlias)
    (cAlias)->(dbCloseArea())
  ENDIF

  dbeSetDefault('PGDBE')
  oStatement := DacSqlStatement(::SqlSessionDict):FromChar(cStatement)
  oStatement:Build():Query(,cAlias)
#endif
ENDIF

IF Empty(cAlias) .OR. Empty(Alias())
  RETURN .f.
ENDIF

::filterSqlStatement := cStatement
::sqlExecuteTime := Seconds() - nSeconds

RETURN .t.

* ----------

METHOD DC_XbpBrowseFiltered:getSqlStatement()

LOCAL cSqlTableName, cSqlStatement

IF Valtype(::dataSource) # 'C'
  RETURN ''
ENDIF

IF !Empty(::filterSqlStatement)
  RETURN ::filterSqlStatement
ENDIF

cSqlStatement := ''
cSqlTableName := (::dataSource)->(dbInfo( DBO_FILENAME ))

IF cSqlTableName = 'SELECT '
  cSqlStatement := cSqlTableName
ELSE
  cSqlStatement := (::dataSource)->(DC_AdsGetSqlStatement())
ENDIF

RETURN cSqlStatement

* -----------

METHOD DC_XbpBrowseFiltered:orderBy( nColumn, oColumn )

LOCAL lStatus

IF Empty(::sqlSessionDict) .AND. Empty(::sqlSessionFree)
  lStatus := ::DC_XbpBrowse:orderBy(nColumn,oColumn)
  IF lStatus
    IF Valtype(::filterRefreshBlock) == 'B'
      Eval(::filterRefreshBlock,self)
    ENDIF
  ENDIF
  RETURN lStatus
ENDIF

RETURN ::setFilter(nColumn)

* -----------

METHOD DC_XbpBrowseFiltered:BuildFilter( aColumns, nOrderColumn, xScope1, xScope2, cScopeTagName )

LOCAL cExpr, i, cFieldName, cFieldType, cDate, cValue, ;
      xFilterVal1, xFilterVal2, nFilterType, cDbfName, cLen1, cLen2, ;
      cPrefix := '', lNoAnd := .f., cOrderBy := '', nFound, lHasFields, ;
      cCaption, cSqlFieldName, cWhere, lFilterExact

cWhere := ''
lHasFields := .f.

IF Empty(nOrderColumn)
  nOrderColumn := ::orderColumn
ENDIF

IF !Empty(::filterStatementOriginal)
  cExpr := ::filterStatementOriginal
ELSE
  cExpr := ''
ENDIF

FOR i := 1 TO Len(aColumns)
  cCaption := NIL
  IF !Empty(aColumns[i]:heading:childList())
    cCaption := aColumns[i]:heading:childList()[1]:caption
    IF cCaption == 't'
      cCaption := NIL
    ENDIF
  ENDIF

    /*
    IF !Empty(::columnNames)
      nFound := Ascan(::columnNames,{|a|Upper(a[2]) == Upper(cCaption)})
       wtf nFound, cCaption, ::columnNames[nFound]
      IF nFound > 0
        cSqlFieldName := ::columnNames[nFound,1]
      ELSE
        cSqlFieldName := ''
      ENDIF
    ENDIF
    */

  IF Empty(cCaption)
    cCaption := aColumns[i]:headerCaption
  ENDIF

  nFilterType := aColumns[i]:filterType

  IF nFilterType # 3 .AND. ;
      (Empty(aColumns[i]:filterVal1) .AND. Empty(aColumns[i]:filterVal2) .AND. Empty(aColumns[i]:filterVal) .AND. Valtype(aColumns[i]:filterVal1)#'L')
    LOOP
  ELSEIF !Empty(aColumns[i]:sqlField)
    cFieldName := aColumns[i]:sqlField
  ELSEIF !Empty(cSqlFieldname)
    cFieldName := cSqlFieldName
  ELSEIF !Empty(cCaption)
    cFieldName := cCaption
  ENDIF

  xFilterVal1 := nil
  IF !Empty(aColumns[i]:filterVal)
    xFilterVal1 := aColumns[i]:filterVal
  ELSEIF !Empty(aColumns[i]:filterVal1)
    xFilterVal1 := aColumns[i]:filterVal1
  ELSEIF Valtype(aColumns[i]:filterVal1) == 'L'
    xFilterVal1 := aColumns[i]:filterVal1
  ENDIF
  xFilterVal2 := aColumns[i]:filterVal2
  lFilterExact := aColumns[i]:filterExact

  IF nFilterType # 3 .AND. Empty(xFilterVal1) .AND. Empty(xFilterVal2) .AND. Valtype(xFilterVal1) # 'L'
    LOOP
  ENDIF

  IF nFilterType == 3
    lHasFields := .t.
  ENDIF

  cFieldType := Valtype(xFilterVal1)
  IF nFilterType == 3 .OR. (cFieldType # 'U' .AND. (!Empty(xFilterVal1) .OR. cFieldType = 'L'))
    lHasFields := .t.
    IF !Empty(cWhere)
      cWhere += ' .AND. '
    ENDIF
  ENDIF

  IF nFilterType == 3 // empty
      cWhere += " Empty(" + cFieldName + ")"

  ELSEIF cFieldType = 'C' .AND. !Empty(xFilterVal1) .AND. !Empty(xFilterVal2) .AND. !(xFilterVal1==xFilterVal2)
    IF nFilterType = 1
      cLen1 := Alltrim(Str(Len(Alltrim(xFilterVal1))))
      cLen2 := Alltrim(Str(Len(Alltrim(xFilterVal2))))
      cWhere += " Left(Upper(" + cFieldName + ")," + cLen1 + ") >= '" + Upper(Alltrim(xFilterVal1)) + "' .AND. " + ;
                " Left(Upper(" + cFieldName + ")," + cLen2 + ") <= '" + Upper(Alltrim(xFilterVal2)) + "'"
    ELSE
      cWhere += "('" + Upper(Alltrim(xFilterVal1)) + "' $ Upper(" + cFieldName + ") .OR. '" + ;
                       Upper(Alltrim(xFilterVal2)) + "' $ Upper(" + cFieldName + "))"
    ENDIF
  ELSEIF cFieldType = 'C' .AND. !Empty(xFilterVal1)
    IF nFilterType == 1
      IF lFilterExact
        cWhere += " Trim(Upper(" + cFieldName + ")) = '" + Upper(Alltrim(xFilterVal1)) + "'"
      ELSE
        cLen1 := Alltrim(Str(Len(Alltrim(xFilterVal1))))
        cWhere += " Left(Upper(" + cFieldName + ")," + cLen1 + ") = '" + Upper(Alltrim(xFilterVal1)) + "'"
      ENDIF
      IF Empty(cScopeTagName) .AND. (!Empty(cScopeTagName := ::findGoodIndex(cFieldName)))
        xScope1 := Alltrim(xFilterVal1)
        xScope2 := Alltrim(xFilterVal2)
        nOrderColumn := i
      ENDIF
    ELSE
      cWhere += "'" + Upper(Alltrim(xFilterVal1)) + "' $ Upper(" + cFieldName + ")"
    ENDIF
  ELSEIF cFieldType = 'N' .AND. !Empty(xFilterVal1) .AND. !Empty(xFilterVal2) .AND. !(xFilterVal1==xFilterVal2)
    cWhere += cFieldName + '>=' + Alltrim(DC_XtoC(xFilterVal1)) + ' .AND. ' + ;
              cFieldName + '<=' + Alltrim(DC_XtoC(xFilterVal2))
    IF Empty(cScopeTagName) .AND. (!Empty(cScopeTagName := ::findGoodIndex(cFieldName)))
      xScope1 := xFilterVal1
      xScope2 := xFilterVal2
      nOrderColumn := i
    ENDIF
  ELSEIF cFieldType = 'N' .AND. !Empty(xFilterVal1)
    cWhere += cFieldName + '=' + Alltrim(DC_XtoC(xFilterVal1))
    IF Empty(cScopeTagName) .AND. (!Empty(cScopeTagName := ::findGoodIndex(cFieldName)))
      xScope1 := xFilterVal1
      xScope2 := xFilterVal1
      nOrderColumn := i
    ENDIF
  ELSEIF cFieldType = 'D' .AND. !Empty(xFilterVal1) .AND. !Empty(xFilterVal2)
    cDate := Dtos(xFilterVal1)
    cWhere += 'DtoS(' + cFieldname + " ) >= '" +  cDate + "'"
    cDate := Dtos(xFilterVal2)
    cWhere += " .AND. DtoS(" + cFieldname + ") <= '" + cDate + "'"
    IF Empty(cScopeTagName) .AND. (!Empty(cScopeTagName := ::findGoodIndex(cFieldName)))
      xScope1 := DtoS(xFilterVal1)
      xScope2 := DtoS(xFilterVal2)
      nOrderColumn := i
    ENDIF
  ELSEIF cFieldType = 'D' .AND. !Empty(xFilterVal1)
    cDate := Dtos(xFilterVal1)
    cWhere += "DtoS(" + cFieldname + "='" + cDate + "'"
    IF Empty(cScopeTagName) .AND. (!Empty(cScopeTagName := ::findGoodIndex(cFieldName)))
      xScope1 := DtoS(xFilterVal1)
      xScope2 := DtoS(xFilterVal1)
      nOrderColumn := i
    ENDIF
  ELSEIF cFieldType = 'L'
    IF xFilterVal1
      cValue := '.T.'
    ELSE
      cValue := '.F.'
    ENDIF
    cWhere += cFieldName + '=' + cValue
  ELSE
    LOOP
  ENDIF
NEXT

IF !lHasFields
  cWhere := ''
ENDIF

IF lHasFields
  IF !Empty(::filterStatementOriginal)
    cWhere := ' AND ' + cWhere
  ENDIF
ENDIF

IF !lHasFields .AND. !Empty(::filterStatementOriginal) .AND. Empty(nOrderColumn)
  cExpr := ::filterStatementOriginal
ELSE
  cExpr += cWhere
ENDIF

RETURN cExpr

* -----------

METHOD DC_XbpBrowseFiltered:ScopeToWhere()

LOCAL cWhere := '', xScopeTop, xScopeBot, cAlias, cLen1, cLen2, cIndex

IF !Empty(::scopeWhere)
  RETURN ::scopeWhere
ENDIF

cAlias := ::dataSource
IF Empty(cAlias) .OR. Valtype(cAlias) # 'C'
  RETURN cWhere
ENDIF

xScopeTop := (cAlias)->(DC_SetScope(0))
xScopeBot := (cAlias)->(DC_SetScope(1))

IF Empty(xScopeTop) .OR. Empty(xScopeBot)
  RETURN cWhere
ENDIF

cLen1 := Alltrim(Str(Len(DC_XtoC(xScopeTop))))
cLen2 := Alltrim(Str(Len(DC_XtoC(xScopeBot))))

cIndex := (cAlias)->(IndexKey(0))

IF Valtype(xScopeTop) == 'C'
  cWhere := 'Left(' + cIndex + ',' + cLen1 + ") >= '" + DC_XtoC(xScopeTop) + "'"
  cWhere += ' and '
  cWhere += 'Left(' + cIndex + ',' + cLen2 + ") <= '" + DC_XtoC(xScopeBot) + "'"
ELSEIF Valtype(xScopeTop) == 'N'
  cWhere := cIndex + " >= " + DC_XtoC(xScopeTop)
  cWhere += ' and '
  cWhere += cIndex + " <= " + DC_XtoC(xScopeBot)
ENDIF

::scopeWhere := cWhere

RETURN cWhere

* ----------

METHOD DC_XbpBrowseFiltered:IsGoodIndex( cFieldName, aStru, cIndexKey )

LOCAL lStatus := .f., aTokens, i

IF Upper(cFieldName) == Upper(cIndexKey)
  RETURN .t.
ELSEIF 'UPPER('+Upper(cFieldName)+')' == Upper(cIndexKey)
  RETURN .t.
ELSE
  aTokens := DC_TokenArray(Upper(cIndexKey),'+')
  FOR i := 1 TO Len(aTokens)
    aTokens[i] := Strtran(aTokens[i],'UPPER(','')
    aTokens[i] := Strtran(aTokens[i],'DTOS(','')
    aTokens[i] := Strtran(aTokens[i],')','')
    aTokens[i] := Alltrim(aTokens[i])
    IF AScan(aStru,{|a|Upper(a[1])==aTokens[i]}) = 0
      RETURN .f.
    ENDIF
  NEXT
ENDIF

RETURN .t.

* -----------

METHOD DC_XbpBrowseFiltered:PrintBrowse( cTitle, lLockUpdate, lDoneMessage, lCancelDialog )

LOCAL oPrinter, i, nStartcol, nEndCol, nCol, nColSize, nWidth
LOCAL nChoice := 1, nOrientation, lGrid, nFudge, nColWidth, nHeaderWidth
LOCAL lUnSelected := .F., aOptions, aColumns[0], aButtons, aMessage[0]
LOCAL aPrintJob[0], aRes := {}, lStatus, oAppWindow := SetAppWindow(self)
LOCAL nTags, nRecs, lPrintTaggedColumns := .f., nTaggedColumns := 0
LOCAL nCollapsedColumns := 0, oColumn, bSaveFilter := ::filter
LOCAL bItemMarked := ::itemMarked

nOrientation := ::nPrintOrientation
nChoice := nOrientation
lGrid := ::lPrintGrid
nFudge := ::nPrintFudge

nWidth := IIF(nOrientation=1, 700, 882 )

FOR i := 1 TO ::colCount
  oColumn := ::getColumn(i)
  IF oColumn:isTagged .AND. !oColumn:isCollapsed
    nTaggedColumns++
  ENDIF
  IF oColumn:isTagged .AND. oColumn:isCollapsed
    nCollapsedColumns++
  ENDIF
NEXT

IF nTaggedColumns > 0
  IF nCollapsedColumns = 0
    nChoice := 1
    DCMSGBOX 'Print Option:' ;
       BUTTONS {'Print only the ' + Alltrim(Str(nTaggedColumns)) + ' Tagged Columns', ;
                'Print ALL Columns','Cancel'} CHOICE @nChoice ;
                 FONT '12.Lucida Console'

    IF nChoice == 1
      lPrintTaggedColumns := .t.
    ELSEIF nChoice # 2
      RETURN .f.
    ENDIF
  ENDIF
ENDIF

aMessage := {'Columns to print:',''}

nHeaderWidth := 0
FOR i := 1 TO ::colCount
  nHeaderWidth := Max(nHeaderWidth,Len(::getColumn(i):heading:cargo))
NEXT
nHeaderWidth++

nColWidth := 0
FOR i := 1 TO ::colCount
  IF lPrintTaggedColumns .AND. !::getColumn(i):isTagged
    LOOP
  ELSEIF ::getColumn(i):isCollapsed
    LOOP
  ENDIF
  nColWidth += ::getColumn(i):currentSize()[1]
  IF nColWidth <= nWidth
    AAdd(aColumns,::getColumn(i))
    AAdd(aMessage,Pad(::getColumn(i):heading:cargo,nHeaderwidth) + ' : ' + IIF(nColWidth<700,'Portrait or Landscape','Landscape Only'))
  ELSE
    EXIT
  ENDIF
NEXT

aButtons := {'Portrait','Landscape','Cancel' }

DC_MsgBox(,,aMessage,'Print Confirmation',,,,@nChoice, aButtons,,,,'12.Lucida Console')

aRes := AppDeskTop():currentSize()

IF nChoice $ {1,2}

   nTags := 0

   IF Valtype(::dataSource) == 'C'
     nTags := Len((::datasource)->(DC_RecTagArray()))
   ELSE
     AEval(::dataSource,{|a,x|x := a[2],IIF(x,nTags++,nil)})
   ENDIF

   IF nTags > 0
     nChoice := 1
     DCMSGBOX 'Print Option:' ;
      BUTTONS {'Print only the Tagged (Yellow) Records', ;
               'Print ALL Records','Cancel'} CHOICE @nChoice ;
               FONT '12.Lucida Console'

     IF nChoice == 1
       IF Valtype(::dataSource) == 'C'
         (::dataSource)->(DC_SetScopeArray(DC_RecTagArray()))
       ELSE
         IF Empty(::filter)
           ::filter := {|a|a[2]}
         ELSE
           ::filter := {|a|Eval(bSaveFilter) .AND. a[2]}
         ENDIF
       ENDIF
     ELSEIF nChoice # 2
       RETURN .f.
     ENDIF
   ENDIF

   nRecs := ::displayedRows()

   nOrientation := nChoice

   DCPRINT ON ;
      TO oPrinter ;
      OPTIONS aOptions ;
      PIXEL ;
      ORIENTATION nOrientation

   IF Valtype(oPrinter) # 'O' .OR. !oPrinter:lActive
     SetAppWindow(oAppWindow)
     RETURN .f.
   ENDIF

   ::itemMarked := nil

   lStatus := PrintBrowse2( self, aOptions, aColumns, oPrinter, ;
                 nOrientation, nFudge, lGrid, nTags, nRecs, ;
                 lPrintTaggedColumns, bSaveFilter, cTitle, ;
                 lLockUpdate, lDoneMessage, lCancelDialog )

   ::itemMarked := bItemMarked

   IF lStatus
     DCPRINT OFF PRINTER oPrinter
   ELSE
     DCPRINT ABORT PRINTER oPrinter
   ENDIF

   DC_ClearEvents()

ENDIF

::filter := bSaveFilter
DC_SetScopeArray(nil)
SetAppWindow(oAppWindow)

RETURN lStatus

* -----------------

STATIC FUNCTION PrintBrowse2( self, aOptions, aColumns, oPrinter, nOrientation, ;
                        nFudge, lGrid, nTags, nRecs, lPrintTaggedColumns, ;
                        bSaveFilter, cTitle, lLockUpdate, lDoneMessage, ;
                        lCancelDialog )

LOCAL nSaveRec, aHeaders, aFooters, nSaveArea, ;
      i, j, nRow, nCol, oColumn, nRecord, nPreviousRecord, nColSize, ;
      xValue, oFont, cFont, nPointSize, nPage := 1, nHeaderHeight, ;
      nFooterHeight, nDataHeight, nHFRowHeight, nDataRowHeight, ;
      nStartDataRow, lStatus, nDataRows, nPages, nCenter, nRight, nLeft, nBoxBottom, nPageLine,;
      aAttr := Array( GRA_AS_COUNT ), aAreaAttr := Array(GRA_AA_COUNT), ;
      cLastFont, aHeader, aFooter, nRows := 58, nExit := 0, lCancelled := .f., ;
      oProgress, oDlg, nEvery, nDisplayedRows

LOCAL nCount, GetList[0], nProgCount := 0

LOCAL nSeconds := Seconds()

nCenter := IIF(nOrientation==1,40,51)
nRight := IIF(nOrientation==1,79,102)
nLeft := 10

DEFAULT cTitle := Upper(Alias()) + " CUSTOM LIST", ;
        lLockUpdate := .t., ;
        lDoneMessage := .t., ;
        lCancelDialog := .t.

aHeader := { ;
   { cTitle, nCenter*8.75, DCPRINT_ALIGN_HCENTER, '14.Arial Bold' },;
   { "As Of "+DTOC(Date())+" - "+FormatDate(Date())+" At "+SecAmPm(), ;
      nCenter*8.75, DCPRINT_ALIGN_HCENTER, '10.Arial Bold' }}

aFooter := {{"",""}} /// we'll define the footer at the time of printing

aHeaders := Array(Len(aColumns))
aFooters := Array(Len(aColumns))

FOR i := 1 TO Len(aColumns)
  oColumn := aColumns[i]
  IF !Empty(oColumn:heading:cargo)
    aHeaders[i] := oColumn:heading:cargo
  ELSE
    aHeaders[i] := oColumn:heading:referenceString
  ENDIF
  IF !Empty(oColumn:footing:cargo)
    aFooters[i] := oColumn:footing:cargo
  ELSE
    aFooters[i] := oColumn:footing:referenceString
  ENDIF
NEXT

nHeaderHeight := (Len(aHeader)+1) * oPrinter:nRowSize // header length plus 1 line below for shading

nFooterHeight := (Len(aFooter)+1) * oPrinter:nRowSize
nHFRowHeight := oPrinter:nRowSize

DCPRINT SIZE nRows, 80 PRINTER oPrinter
nDataRowHeight := oPrinter:nRowSize

nDataHeight := oPrinter:offset() - nHeaderHeight - nFooterHeight - (nHFRowHeight*2)
nDataRows := nDataHeight / nDataRowHeight

nDataHeight := (Int( nDataRows) * nDataRowHeight)

nRow := 1

nCount := 0

::hitBottomBlock := {||nExit++}
::goTop()
::forceStable()
IF lLockUpdate
  ::setParent():lockUpdate(.t.)
ENDIF

nRecord := 0

IF lCancelDialog
  @ 0,0 DCPROGRESS oProgress SIZE 50,1 PERCENT COLOR GRA_CLR_DARKCYAN
  @ 2,0 DCPUSHBUTTON CAPTION 'Cancel' SIZE 10 ;
        ACTION {||lCancelled := .t.}
  DCREAD GUI FIT TITLE 'Printing ' + cTitle PARENT @oDlg MODAL EXIT
ENDIF

nDisplayedRows := ::displayedRows()

nEvery := Int(nDisplayedRows/100)

DO WHILE lStatus := DC_PrinterOk(oPrinter) .AND. !lCancelled

  nCol := 10*nFudge
  nProgCount ++

  IF lCancelDialog .AND. (nEvery == 0 .OR. nProgCount % nEvery == 0)
    DC_CompleteEvents()
    DC_GetProgress( oProgress, nProgCount, nDisplayedRows )
  ENDIF

  /// Print Title and Column Headers
  IF nRow <= 1
    DCPRINT SIZE nRows, IIF( nOrientation=1,80,132) PRINTER oPrinter
    nRow += oPrinter:nRowSize

    /// if there's a header draw a shaded area around it
    IF !Empty(aHeader)
      aAreaAttr[GRA_AA_COLOR] := GRA_CLR_PALEGRAY
      aAreaAttr[GRA_AA_SYMBOL] := GRA_SYM_SOLID
      @ nRow-20, (10*nFudge), ;
        nRow+((oPrinter:nRowSize*Len(aHeader))+10), oPrinter:nCols*oPrinter:nColSize - ;
             (10 * nFudge) ;
      DCPRINT BOX LINEWIDTH 1 FILL GRA_FILL AREAATTR aAreaAttr
    ENDIF

    // Print Title area
    nStartDataRow := nRow
    nRow*=2
    FOR i := 1 TO Len(aHeader)
       IF Valtype(aHeader[i,1]) = 'A'
         FOR j := 1 TO Len(aHeader[i])
            @ nRow, aHeader[i,j,2]*nFudge ;
              DCPRINT SAY aHeader[i,j,1] ALIGN aHeader[i,j,3] FONT aHeader[i,j,4]
         NEXT j
       ELSE
         @ nRow, aHeader[i,2]*nFudge ;
            DCPRINT SAY aHeader[i,1] ALIGN aHeader[i,3] FONT aHeader[i,4]
       ENDIF
       nRow += oPrinter:nRowSize
    NEXT i

    nRow := nHeaderHeight + oPrinter:nRowSize + 1
    nStartDataRow := nRow
    IF lGrid

       // Print double line (horizontal) under Column Headers
       @ nRow+nHFRowHeight,nCol,nRow+nHFRowHeight, ;
           oPrinter:nCols*oPrinter:nColSize - (10*nFudge) ;
           DCPRINT LINE PRINTER oPrinter

       @ nRow+nHFRowHeight+5,nCol,nRow+nHFRowHeight+5, oPrinter:nCols*oPrinter:nColSize - (10*nFudge) ;
           DCPRINT LINE PRINTER oPrinter

    ENDIF

    // Print Column headers
    FOR i := 1 TO Len(aColumns)
      nColSize := aColumns[i]:currentSize()[1]*nFudge
      IF nCol + nColSize > oPrinter:nCols*oPrinter:nColSize - (10*nFudge)
        EXIT
      ENDIF

      cFont := '12.Cambria Bold'

      @ nRow+oPrinter:nRowSize-10,nCol+10 DCPRINT SAY aHeaders[i] ;
                  PRINTER oPrinter ;
                  FONT cFont ;
                  WIDTH nColSize
      nCol += nColSize
    NEXT
    nRow += ( nHFRowHeight + nDataRowHeight )
    DCPRINT SIZE nRows, 80 PRINTER oPrinter

  ENDIF
  nCol := 10*nFudge

  IF lGrid
    // Print Horizontal line between rows
    @ nRow,nCol,nRow,oPrinter:nCols*oPrinter:nColSize - ( 10 * nFudge) ;
       DCPRINT LINE PRINTER oPrinter
  ENDIF

  nRow += nDataRowHeight

  // Print Data
  FOR i := 1 TO Len(aColumns)
    nColSize := aColumns[i]:currentSize()[1]*nFudge
    IF nCol + nColSize > oPrinter:nCols*oPrinter:nColSize - ( 10 * nFudge )
      EXIT
    ENDIF
    xValue := Eval(aColumns[i]:dataLink)
    IF !Empty(xValue)
      oFont := aColumns[i]:dataArea:setFont()
      IF Valtype(oFont) = 'O'
        cFont := oFont:compoundName
        nPointSize := oFont:nominalPointSize
        cFont := Alltrim(Str(nPointSize)) + '.' + cFont
      ELSEIF Valtype(oFont) = 'C'
        cFont := oFont
      ELSE
        cFont := aColumns[i]:dataArea:setFontCompoundName()
      ENDIF
      IF Empty(cFont)
        cFont := ::setFontCompoundName()
      ENDIF

      IF !Empty(cFont) .AND. Upper(Alltrim(cFont)) == 'COURIER'
        cFont := nil
      ENDIF

      @ nRow-10,nCol+10 DCPRINT SAY Trim(DC_XtoC(Eval(aColumns[i]:dataLink))) ;
                PRINTER oPrinter ;
                FONT cFont ;
                WIDTH nColSize
    ENDIF
    nCol += nColSize
  NEXT

  nCount++

  IF nRow >= nHeaderHeight + nDataHeight .OR. nExit >= 2

     aFooter := {{AllTrim(Str(nCount))+" Records Are Listed On This Page."+Space(5)+;
                  AllTrim(Str(::displayedRows()))+" Total Records Were Printed.",;
                  nCenter*8.75, DCPRINT_ALIGN_HCENTER, '10.Arial Bold' }}

     nRow += nDataRowHeight

    // Print Vertical Lines between columns
    IF lGrid
      nCol := 10*nFudge

      // Print Box around Column Headers and Data
      @ nStartDataRow,nCol,nRow - nDataRowHeight,oPrinter:nCols*oPrinter:nColSize - (9.9 * nFudge);
        DCPRINT BOX PRINTER oPrinter /// normally it would be 10*nFudge but to make preview fit we need 9.9 instead

      FOR i := 1 TO Len(aColumns)
        nColSize := aColumns[i]:currentSize()[1]*nFudge
        IF nCol + nColSize > oPrinter:nCols*oPrinter:nColSize - (10 * nFudge)
          EXIT
        ENDIF
        @ nStartDataRow,nCol,nRow - nDataRowHeight,nCol DCPRINT LINE PRINTER oPrinter
        nCol += nColSize
      NEXT
    ENDIF
    nBoxBottom := nRow - nDataRowHeight // we're saving it for positioning comparisons

 //   DCPRINT SIZE nRows, 80 PRINTER oPrinter

    // if there's a footer draw a shaded area around it
    IF !Empty(aFooter)
      aAreaAttr[GRA_AA_COLOR] := GRA_CLR_PALEGRAY
      aAreaAttr[GRA_AA_SYMBOL] := GRA_SYM_SOLID
      nRow := IIF( (nRow-oPrinter:nRowSize)-5 <= nBoxBottom, nBoxBottom+oPrinter:nRowSize+8, nRow )
      @ (nRow-oPrinter:nRowSize)-5, (10*nFudge), ;
        nRow+((oPrinter:nRowSize*(Len(aFooter)-1))+10), ;
        oPrinter:nCols*oPrinter:nColSize - (10 * nFudge) ;
          DCPRINT BOX LINEWIDTH 1 FILL GRA_FILL AREAATTR aAreaAttr
    ENDIF

    FOR i := 1 TO Len(aFooter)
      @ nRow, aFooter[i,2]*nFudge DCPRINT SAY aFooter[i,1] ALIGN aFooter[i,3] FONT aFooter[i,4]
      nRow += oPrinter:nRowSize
    NEXT

    IF ValType(nPages)#'N'
      #if XPPVER > 2000000
        nPages := Ceiling(nRecs/nCount)
      #else
        nPages := Int(nRecs/nCount)
      #endif
    ENDIF

    nPageLine := (oPrinter:nRows-1.5)*oPrinter:nRowSize
    nPageLine := IIF( nPageLine <= nRow, nRow+=oPrinter:nRowSize, nPageLine )

    @ nPageLine,(IIF( nOrientation = 1, 700, 903 )*nFudge)-(10*nFudge) ;
      DCPRINT SAY 'Page ' + Alltrim(Str(nPage++)) + ' Of ' + Alltrim(Str(Int(nPages))) ;
      ALIGN DCPRINT_ALIGN_RIGHT ;
      FONT '8.Arial'
    IF .t. // !Eof()
      DCPRINT EJECT PRINTER oPrinter
    ENDIF
    nRow := 1
    nCount := 0
  ENDIF

  ::down()
  ::forcestable()

  IF nExit >= 2
    EXIT
  ELSEIF Valtype(::dataSource) == 'C'
    nRecord := (::dataSource)->(RecNo())
    IF nRecord == nPreviousrecord
      EXIT
    ENDIF
    nPreviousrecord := nRecord
    IF (::dataSource)->(Eof())
      EXIT
    ENDIF
  ENDIF

ENDDO

IF lCancelDialog
  oDlg:destroy()
ENDIF

IF Valtype(::dataSource) == 'C'
  (::dataSource)->(DC_SetScopeArray(nil))
ENDIF

::filter := bSaveFilter
IF lLockUpdate
  ::setParent():lockUpdate(.f.)
ENDIF
::hitBottomBlock := nil
::goTop()
::forceStable()
::refreshAll()
IF lLockUpdate
  ::setParent():lockUpdate(.f.)
ENDIF

IF lCancelled
  lStatus := .f.
ENDIF

IF lDoneMessage .AND. !lCancelled
  DCMSGBOX 'Done' TIMEOUT 2 OWNER self
ENDIF

RETURN lStatus

* --------

FUNCTION FormatDate(dDate, nType, lBrief )

LOCAL cNum, cFmtDate, cDay, cMonth
LOCAL nDay

dDate  := IIF(ValType(dDate) == 'D', dDate, Date() )
nType  := IIF(ValType(nType) == 'N', nType, 1 )
lBrief := IIF(ValType(lBrief) == 'L', lBrief, .F. )

nDay := Val(SubStr(DTOC(dDate), 4, 2 ) )

IF nType == 9 .OR. nType == 11
   cNum := AllTrim(Str(nDay))
ELSE
  DO CASE
     CASE nDay == 1
          cNum :="1st"
     CASE nDay == 2
          cNum :="2nd"
     CASE nDay == 3
          cNum :="3rd"
     CASE ( nDay >= 4 .AND. nDay <= 9 )
          cNum := SubStr(DTOC(dDate),5,1) + "th"
     CASE nDay == 21 .OR. nDay == 31
          cNum := SubStr(DTOC(dDate),4,2) + "st"
     CASE nDay == 22
          cNum := SubStr(DTOC(dDate),4,2) + "nd"
     CASE nDay == 23
          cNum := SubStr(DTOC(dDate),4,2) + "rd"
     CASE ( nDay >= 10 .AND. nDay <= 20 ) .OR. ( nDay >= 24 .AND. nDay <= 30 )
          cNum := SubStr(DTOC(dDate),4,2) + "th"
     OTHERWISE
          cNum := " "
  ENDCASE
ENDIF

cDay := IIF(!lBrief,CDow(dDate),;
           IIF(Dow(dDate)=5,SubStr(CDow(dDate),1,4),SubStr(CDow(dDate),1,3) ) )

cMonth := IIF(!lBrief,CMonth(dDate),;
           IIF(Month(dDate) = 6  .OR. Month(dDate) = 7 .OR. Month(dDate) = 9,;
               SubStr(CMonth(dDate),1,4),SubStr(CMonth(dDate),1,3) ) )

DO CASE

   // Wednesday, September 31st, 1991   31 chars max
   CASE nType == 1
        cFmtDate := cDay+", "+cMonth+" "+cNum+", "+Str(Year(dDate),4)

   // Wednesday, the 31st of September, 1991  38 chars max
   CASE nType == 2
        cFmtDate := cDay+", the "+cNum+" of "+cMonth+", "+Str(Year(dDate),4)

   // Wednesday, September 31st   25 chars max
   CASE nType == 3
        cFmtDate := cDay+", "+cMonth+" "+cNum

   // Wednesday, the 31st of September  32 chars max
   CASE nType == 4
        cFmtDate := cDay+", the "+cNum+" of "+cMonth

   // September 31st  14 chars max
   CASE nType == 5
        cFmtDate := cMonth+" "+cNum

   // September the 31st  18 chars max
   CASE nType == 6
        cFmtDate := cMonth+" the "+cNum

   // September 31st, 1991 20 chars max
   CASE nType == 7
        cFmtDate := cMonth+" "+cNum+", "+Str(Year(dDate),4)

   // September the 31st, 1991 24 chars max
   CASE nType == 8
        cFmtDate := cMonth+" the "+cNum+", "+Str(Year(dDate),4)

   // September the 31st, 1991 24 chars max
   CASE nType == 9
        cFmtDate := cMonth+" "+cNum+", "+Str(Year(dDate),4)

   // September 1991 14 chars max
   CASE nType == 10
        cFmtDate := cMonth+" "+Str(Year(dDate),4)

   // 31 September, 1991 18 chars max
   CASE nType == 11
        cFmtDate := cNum+" "+cMonth+", "+Str(Year(dDate),4)


   // 31st, September 1991 20 chars max
   CASE nType == 12
        cFmtDate := cNum+" "+cMonth+", "+Str(Year(dDate),4)

   // 31st of September 1991 22 chars max
   CASE nType == 13
        cFmtDate := cNum+" of "+cMonth+" "+Str(Year(dDate),4)

ENDCASE

RETURN( IIF(Len(AllTrim(cFmtDate))<=2,Space(14),cFmtDate) )

* ---------

FUNCTION SecAmPm(nSec, l24 )

LOCAL cTime := ""

l24 := IIF( ValType(l24) = 'L', l24, .F. )  // 12 or 24 hour format

IF ValType(nSec) = 'C'
   nSec := Val(nSec)
ENDIF
IF ValType(nSec) != 'N'
   nSec := Seconds()
ENDIF

cTime := StrZero( Int(Mod(nSec/3600, 24)), 2, 0 ) +':'+;
         StrZero( Int(Mod(nSec/  60, 60)), 2, 0 )

cTime += IIF( Val(SubStr(cTime, 1, 2 )) >= 12, " PM", " AM" )

RETURN cTime

*------------

METHOD DC_XbpBrowseFiltered:Browse2Excel(oOwner, lCsv, cExcelFile)

LOCAL GetList[0], lStatus, nChoice := 0, i, ;
      nTaggedColumns := 0, nCollapsedColumns := 0, lExportTaggedColumns := .f., ;
      oColumn, nTags := 0, aData[0], lExit := .f., cPath, lAborted := .f., ;
      aColData, aColumns[0], aHeaders[0], xValue, bSaveFilter := ::filter, ;
      oDlg, oSay, oProgress, nRows, nCount, nRecno

cPath := DC_Path(AppName(.t.))
DEFAULT lCsv := .f.

DEFAULT cExcelFile := Pad(cPath + Alias() + IIF( lCsv,'.csv','.xls') ,70)

@ 0,0 DCSAY 'Export to ' + IIF(lCsv,'CSV File', 'Excel File') SAYSIZE 0 FONT '12.Arial'

@ 2,0 DCSAY IIF( lCsv, 'CSV','Excel') ' File Name to Create' GET cExcelFile  SAYSIZE 0 SAYBOTTOM

DCREAD GUI FIT ADDBUTTONS TITLE 'Export to ' + IIF(lCsv,'CSV','Excel') ;
   MODAL TO lStatus NOAUTORESTORE OWNER oOwner // EVAL {|o|::SetWindowPos(o)}

IF !lStatus
  RETURN .f.
ENDIF

cExcelFile := Alltrim(cExcelFile)

FOR i := 1 TO ::colCount
  oColumn := ::getColumn(i)
  AAdd(aColumns,oColumn)
  IF oColumn:isTagged .AND. !oColumn:isCollapsed
    nTaggedColumns++
  ENDIF
  IF oColumn:isTagged .AND. oColumn:isCollapsed
    nCollapsedColumns++
  ENDIF
NEXT
IF nTaggedColumns > 0
  IF nCollapsedColumns = 0
    nChoice := 1
    DCMSGBOX 'Export Option:' ;
       BUTTONS {'Export only the ' + Alltrim(Str(nTaggedColumns)) + ' Tagged Columns', ;
                'Export ALL Columns','Cancel'} CHOICE @nChoice ;
                 FONT '12.Lucida Console' OWNER oOwner

    IF nChoice == 1
      lExportTaggedColumns := .t.
    ELSEIF nChoice # 2
      RETURN .f.
    ENDIF
  ENDIF
ENDIF

nTags := 0

IF Valtype(::dataSource) == 'C'
  nTags := Len((::datasource)->(DC_RecTagArray()))
ELSE
  AEval(::dataSource,{|a,x|x := a[2],IIF(x,nTags++,nil)})
ENDIF

IF nTags > 0
   nChoice := 1
   DCMSGBOX 'Export Option:' ;
    BUTTONS {'Export only the ' + Alltrim(Str(nTags)) + ' Tagged Rows', ;
             'Export ALL Rows','Cancel'} CHOICE @nChoice ;
             FONT '12.Lucida Console' OWNER oOwner

   IF nChoice == 1
     IF Valtype(::dataSource) == 'C'
       (::dataSource)->(DC_SetScopeArray(DC_RecTagArray()))
     ELSE
       IF Empty(::filter)
         ::filter := {|a|a[2]}
       ELSE
         ::filter := {|a|Eval(bSaveFilter) .AND. a[2]}
       ENDIF
     ENDIF
   ELSEIF nChoice # 2
     RETURN .f.
   ENDIF
ENDIF

DC_PointerWait()

::hitBottomBlock := {||lExit:=.t.}
::goTop()
::forceStable()
::setParent():lockUpdate(.t.)
lExit := .f.

FOR i := 1 TO Len(aColumns)
  IF Valtype(aColumns[i]) # 'O'
    LOOP
  ELSEIF lExportTaggedColumns .AND. !aColumns[i]:isTagged
    LOOP
  ELSEIF aColumns[i]:isCollapsed
    LOOP
  ENDIF
  AAdd(aHeaders,aColumns[i]:heading:cargo)
NEXT

AAdd(aData,aHeaders)

GetList := {}
@ 0,0 DCSAY 'Exporting (Pass 1)...' SAYSIZE 0 FONT '10.Lucida Console' OBJECT oSay
@ 1,0 DCPROGRESS oProgress SIZE 80,1 PERCENT COLOR GRA_CLR_CYAN, GRA_CLR_WHITE
@ 2,0 DCPUSHBUTTON CAPTION 'Cancel' SIZE 10,1.2 ;
      ACTION {||lAborted := .t.}
DCREAD GUI FIT EXIT SAVE PARENT @oDlg TITLE 'Exporting to ' + cExcelFile

nRows := ::displayedRows()
nCount := 0

DO WHILE !lAborted

  nCount++
  IF nRows > 100
    IF nCount%(Int(nRows/100)) == 0
      DC_GetProgress(oProgress,nCount,nRows)
      DC_CompleteEvents()
      IF lAborted
        EXIT
      ENDIF
    ENDIF
  ENDIF

  aColData := {}
  FOR i := 1 TO Len(aColumns)
    IF Valtype(aColumns[i]) # 'O'
      LOOP
    ELSEIF lExportTaggedColumns .AND. !aColumns[i]:isTagged
      LOOP
    ELSEIF aColumns[i]:isCollapsed
      LOOP
    ENDIF
    xValue := Eval(aColumns[i]:dataLink)
    AAdd(aColData,xValue)
  NEXT
  AAdd(aData,aColData)

  IF Valtype(::dataSource) == 'C'
    nRecno := (::dataSource)->(RecNo())
  ELSEIF Valtype(::dataSource) == 'A'
    nRecno := ::arrayElement
  ELSE
    nRecno := 0
  ENDIF

  ::down()
  ::forceStable()

  IF lExit
    IF Valtype(::dataSource) == 'C'
      IF (::dataSource)->(RecNo()) == nRecno
        EXIT
      ENDIF
    ELSEIF Valtype(::dataSource) == 'A'
      IF ::arrayElement == nRecno
        EXIT
      ENDIF
    ELSE
      EXIT
    ENDIF
  ENDIF

ENDDO

::filter := bSaveFilter
::goTop()
::forceStable()
::refreshAll()
::setParent():lockUpdate(.f.)

DC_PointerArrow()

oDlg:destroy()

IF lCsv
  lStatus := DC_Array2Csv( cExcelFile, aData )
ELSE
  lStatus := DC_Array2Excel( cExcelFile, {aData},,,,,,,,.t.,.t.,)
ENDIF
IF lStatus
  DC_SpawnUrl( cExcelFile )
ENDIF

RETURN lStatus


* ================

CLASS DC_XbpColumnFiltered FROM DC_XbpColumn

EXPORTED:

VAR filterVal
VAR filterVal1
VAR filterVal2
VAR filterBlock
VAR filterType
VAR filterExact
VAR isDateField
VAR fieldName
VAR struArray
VAR isColumnConfigEnabled

METHOD Init(), GetFilter(), ResetFilter(), SetFilterBlock(), LocateRecord(), ;
       TotalColumn(), TagColumn(), SetFooter()

ENDCLASS

* ----------

METHOD DC_XbpColumnFiltered:init( oParent, oGetList, nColumn )

LOCAL aGetListitem := {}, aOptions, bSort

::DC_XbaseParts:init( oGetList,,self )

aGetListItem := ::getList:getListArray[::getListPointer]
aOptions     := aGetListItem[xGETLIST_OPTIONS]

::DC_XbpColumn:Init( oParent, oGetList, nColumn )

::footerBlock := {||Chr(255)}
::filterType := 1

IF !Empty(::heading:childList())
  ::heading:childlist()[1]:rbDown := {||::getFilter()}
  bSort := ::heading:childlist()[1]:lbDown
ENDIF
::heading:rbDown := {||::getFilter()}
::isDateField := .f.
::filterExact := .f.
::isColumnConfigEnabled := .f.
::filterArray := {}
IF ::sqlSort
  ::sortBlock := {|nColumn,oColumn|oParent:orderBy(nColumn,oColumn)}
ENDIF

RETURN self

* ----------

METHOD DC_XbpColumnFiltered:ResetFilter( lSetRange, bFilter, nFilterType )

::filterVal := nil
::filterVal1 := nil
::filterVal2 := nil
::filterType := 1
::filterExact := .f.
::filterBlock := nil
::footerBlock := {||Chr(255)+';'+Chr(255)}

RETURN self

* ----------

METHOD DC_XbpColumnFiltered:GetFilter( lSetRange, bFilter, nFilterType )

LOCAL GetList[0], GetOptions, cSay, cPicture, lStatus, xDefaultVal, xValue, ;
      oOwner := ::parent:setParent():setParent(), oGet1, oGet2, lFlag2, ;
      cSpace := '', bVarBlock, aStru, cFieldName, cType, nLength, nDecimals, ;
      cSqlField, bClear, bLocate, bReset, nRow, aStruArray, nFound, lClear, ;
      bTotal, bTotalWhen, nGetSize, bTag, bDelete, bAdd, bInsert, ;
      oToolBar1, oToolBar2, oDlg

DEFAULT lSetRange := .t.

nLength := 0
lFlag2 := .f.

IF !Empty(::heading:cargo)
  cSay := ::heading:cargo
ELSEIF !Empty(::heading:childList())
  cSay := ::heading:childList()[1]:caption
ENDIF
cSay := StrTran(cSay, "...", "")

IF ::struArray == Nil .AND. Valtype(::parent:dataSource) == 'C'
  aStruArray := (::parent:dataSource)->(dbStruct())
  nFound := AScan(aStruArray,{|a|Upper(a[1])==Upper(cSay)})
  IF nFound > 0
    ::struArray := aStruArray[nFound]
  ENDIF
ENDIF

::fieldName := cSay

aStru := ::struArray

cSqlField := ::sqlField
DEFAULT cSqlField := '???'

IF Valtype(aStru)=='A' .AND. Len(aStru) == 4
  cFieldName := aStru[1]
  cType := Substr(aStru[2],1)
  nLength := aStru[3]
  nDecimals := aStru[4]
  IF cType $ 'MQ'
    xValue := Space(30)
  ELSEIF cType $ 'CR'
    xValue := Space(aStru[3])
  ELSEIF cType $ 'DQT'
    xValue := Ctod('')
  ELSEIF cType $ 'NI'
    xValue := 0
    cPicture := Repl('9',nLength+1)
    IF nDecimals > 0
      cPicture += '.' + Repl('9',nDecimals)
    ENDIF
  ELSEIF cType == 'L'
    xValue := .f.
    lSetRange := .f.
  ENDIF
ELSEIF !Empty(::sqlField) .AND. Valtype(::parent:dataSource) = 'C'
  cSqlField := Strtran(::sqlField,'[','')
  cSqlField := Strtran(cSqlField,']','')
  cSqlField := Strtran(::sqlField,'"','')
  cSqlField := Strtran(cSqlField,'"','')
  xValue := (::parent:dataSource)->(&(cSqlField))
ELSE
  xValue := Eval(::dataLink)
ENDIF

nGetSize := nil

IF Valtype(xValue) $ 'CM'
  xDefaultVal := Space(Len(xValue))
  ::filterVal1 := Pad(DC_XtoC(::filterVal1),Len(xValue))
  ::filterVal2 := Pad(DC_XtoC(::filterVal2),Len(xValue))
  IF Valtype(xValue) = 'M'
    nGetSize := 30
  ELSEIF Len(::filterVal1) > 60
    nGetSize := 60
  ENDIF
ELSEIF Valtype(xValue) == 'N' .AND. !::isDateField
  IF nLength == 10 .AND. nDecimals == 0 .AND. !::isDateField
    DCMSGBOX 'Is ' + cFieldName + ' a DATE field?' ;
       YESNO TO ::isDateField OWNER ::parent
  ENDIF
  IF ::isDateField
    xDefaultVal := CtoD('')
    ::filterVal1 := DC_UnixTime2DateTime(::filterVal1)[1]
    ::filterVal2 := DC_UnixTime2DateTime(::filterVal2)[1]
  ELSE
    xDefaultVal := 0
    ::filterVal1 := Val(DC_XtoC(::filterVal1))
    ::filterVal2 := Val(DC_XtoC(::filterVal2))
  ENDIF
  IF Empty(cPicture)
    cPicture := '999999999.99'
  ENDIF
ELSEIF Valtype(xValue) == 'D' .OR. ::isDateField
  xDefaultVal := CtoD('')
  ::filterVal1 := CtoD(DC_XtoC(::filterVal1))
  ::filterVal2 := CtoD(DC_XtoC(::filterVal2))
ELSEIF Valtype(xValue) == 'L'
  ::filterVal1 := .f.
  ::filterVal2 := .f.
  xDefaultVal := .f.
  cPicture := 'Y'
ENDIF

@ 0,0 DCSAY cSay + ' Filter' + IIF( lSetRange,' (from)','') ;
      GET ::filterVal1 PICTURE cPicture;
      GETSIZE nGetSize ;
      SAYSIZE 0 SAYBOTTOM ;
      PROPER PROPOPTIONS {.F., Space(1) } ;
      POPUP {|a,oGet| ::filterVal1 := xDefaultVal};
      POPCAPTION "X" POPFONT '10.Arial' POPSTYLE DCGUI_POPUPSTYLE_IMBEDDED ;
      POPTOOLTIP "Clear Filter";
      SAYTOOLTIP "Enter " + cSay + Chr(13) + ;
                 '( SQL: ' + cSqlField + ' )' ;
      KEYBLOCK {|a,b,o|IIF(!lFlag2 .AND. lSetRange .AND. a#xbeK_TAB .AND. a#xbeK_ENTER, ;
        (PostAppevent(xbeP_Keyboard,a,,oGet2),DC_CompleteEvents(),oGet2:getData()),nil)} ;
      GETOBJECT oGet1 ;
      TABSTOP

IF Valtype(xValue) = 'D' .OR. ::isDateField
  @ 0, DCGUI_COL DCGET cSpace ;
      TABSTOP ;
      POPUP {|d|d:=DC_PopDate(oGet1:getData(),,,,,,2),oGet1:setData(d),cSpace} ;
      POPSTYLE DCGUI_POPUPSTYLE_IMBEDDED
ENDIF

IF lSetRange
  @ 1,0 DCSAY cSay + " Filter ( to )";
      GET ::filterVal2 PICTURE cPicture;
      GETSIZE nGetSize ;
      GOTFOCUS {|a,b,o|IIF(!Empty(o:get:buffer),lFlag2:=.t.,lFlag2:=.f.)} ;
      PROPER PROPOPTIONS {.F., Space(1) } ;
      POPUP {|a,oGet| ::filterVal2 := xDefaultVal};
      POPCAPTION "X" POPFONT '10.Arial' POPSTYLE DCGUI_POPUPSTYLE_IMBEDDED ;
      POPTOOLTIP "Clear Filter";
      SAYTOOLTIP "Enter " + cSay + Chr(13) + ;
                 '( SQL: ' + cSqlField + ' )' ;
      GETOBJECT oGet2 ;
      TABSTOP

  IF Valtype(xValue) = 'D' .OR. ::isDateField
    @ 1.0, DCGUI_COL DCGET cSpace ;
      TABSTOP ;
      POPUP {|d|d:=DC_PopDate(oGet2:getdata(),,,,,,2),oGet2:setData(d),cSpace} ;
      POPSTYLE DCGUI_POPUPSTYLE_IMBEDDED
  ENDIF
ENDIF

nRow := 2

IF Valtype(xValue) $ 'CM' .AND. Empty(bFilter) .AND. Empty(nFilterType)
  @ nRow++,0 DCRADIO ::filterType VALUE 1 PROMPT 'Value Starts with' ;
        EVAL {|o|o:keyboard := ;
         {|a,b,o|IIF(a==xbeK_ENTER,SetAppFocus(DC_GetObject(GetList,'DCGUI_BUTTON_OK')),nil)}}

  @ DCGUI_ROW, DCGUI_COL + 30 DCCHECKBOX ::filterExact PROMPT 'Exact Match' ;
       WHEN {||::filterType == 1 .AND. Valtype(xValue)='C'}

  @ nRow++,0 DCRADIO ::filterType VALUE 2 PROMPT 'Value contained in' ;
        EVAL {|o|o:keyboard := ;
         {|a,b,o|IIF(a==xbeK_ENTER,SetAppFocus(DC_GetObject(GetList,'DCGUI_BUTTON_OK')),nil)}}

  @ nRow++,0 DCRADIO ::filterType VALUE 3 PROMPT 'Value is Empty' ;
        EVAL {|o|o:keyboard := ;
         {|a,b,o|IIF(a==xbeK_ENTER,SetAppFocus(DC_GetObject(GetList,'DCGUI_BUTTON_OK')),nil)}} ;
        WHEN {||(::parent:datasource)->(DbInfo( DBO_DBENAME )) $ ;
                 { 'ADSDBE','FOXCDX','DBFCDX','DBFNTX'}}

ELSEIF Valtype(xValue) $ 'N' .AND. Empty(bFilter) .AND. Empty(nFilterType)
  ::filterType := 0
  @ nRow++,0 DCRADIO ::filterType VALUE 3 PROMPT 'Value is Zero' ;
        EVAL {|o|o:keyboard := ;
         {|a,b,o|IIF(a==xbeK_ENTER,SetAppFocus(DC_GetObject(GetList,'DCGUI_BUTTON_OK')),nil)}} ;
        WHEN {||(::parent:datasource)->(DbInfo( DBO_DBENAME )) $ ;
                 { 'ADSDBE','FOXCDX','DBFCDX','DBFNTX'}}

ELSEIF Valtype(xValue) $ 'D' .AND. Empty(bFilter) .AND. Empty(nFilterType)
  ::filterType := 0
  @ nRow++,0 DCRADIO ::filterType VALUE 3 PROMPT 'Value is Empty' ;
        EVAL {|o|o:keyboard := ;
         {|a,b,o|IIF(a==xbeK_ENTER,SetAppFocus(DC_GetObject(GetList,'DCGUI_BUTTON_OK')),nil)}} ;
        WHEN {||(::parent:datasource)->(DbInfo( DBO_DBENAME )) $ ;
                 { 'ADSDBE','FOXCDX','DBFCDX','DBFNTX'}}


ELSE
  ::filterType := nFilterType


ENDIF

lClear := .f.
bClear := {||::filterVal1:=xDefaultVal, ;
             ::filterVal2:=xDefaultVal, ;
             ::filterType := 0, ;
             lFlag2:=.f., ;
             DC_GetRefresh(GetList),;
             SetAppFocus(oGet1), ;
             IIF(Valtype(::filterVal1)=='L',lClear := .t.,nil), ;
             DC_ReadGuiEvent(DCGUI_EXIT_OK,GetList)}

bLocate := {||::LocateRecord(bFilter,cFieldName,bClear,GetList)}
bTotal := {||::TotalColumn(cSay)}
bTotalWhen := {||ValType(::filterVal1)='N'}
bReset := {||::parent:resetAll()}
bTag := {||::TagColumn(GetList)}
bDelete := {||::DeleteColumnFiltered(),DC_ReadGuiEvent(DCGUI_EXIT_ABORT,GetList)}
bAdd := {||::AddColumnFiltered(.f.),DC_ReadGuiEvent(DCGUI_EXIT_ABORT,GetList)}
bInsert := {||::AddColumnFiltered(.t.),DC_ReadGuiEvent(DCGUI_EXIT_ABORT,GetList)}

nRow++
@ nRow, 0 DCTOOLBAR oToolBar1 SIZE 60 BUTTONSIZE 60 SPACE 2
DCSETPARENT TO oToolBar1
DCADDBUTTON CAPTION 'Ok' ACTION {||DC_ReadGuiEvent(DCGUI_EXIT_OK,GetList)} ID 'DCGUI_BUTTON_OK'
DCADDBUTTON CAPTION 'Cancel' ACTION {||DC_ReadGuiEvent(DCGUI_EXIT_ABORT,GetList)}
DCADDBUTTON CAPTION 'Clear' ACTION bClear TOOLTIP 'Clear this Column'
DCADDBUTTON CAPTION 'Locate' ACTION bLocate TOOLTIP 'Locate next record that matches entered value'
DCADDBUTTON CAPTION 'Total' ACTION bTotal TOOLTIP 'Total this Column' WHEN bTotalWhen
DCADDBUTTON CAPTION {||IIF(::isTagged,'UnTag','Tag')} ACTION bTag ;
            TOOLTIP {||IIF(::isTagged,'Remove the Column Tag','Tag this Column')}

nRow++
DCSETPARENT TO

IF ::isColumnConfigEnabled
  @ nRow, 0 DCTOOLBAR oToolBar2 SIZE 60 BUTTONSIZE 60 SPACE 2
  DCSETPARENT TO oToolBar2
  DCADDBUTTON CAPTION 'Delete' ACTION bDelete TOOLTIP 'Delete this Column'
  DCADDBUTTON CAPTION 'Add' ACTION bAdd TOOLTIP 'Add a Column'
  DCADDBUTTON CAPTION 'Insert' ACTION bInsert TOOLTIP 'Insert a Column before this Column'
  DCADDBUTTON CAPTION '< Move' ACTION {||::MoveColumnFiltered(1), ;
                                         DC_ReadGuiEvent(DCGUI_EXIT_ABORT,GetList)} ;
            TOOLTIP 'Move Column LEFT' ;
            WHEN {||::displayIndex > Len(::parent:setLeftFrozen())+1}
  DCADDBUTTON CAPTION 'Move >' ACTION {||::MoveColumnFiltered(2), ;
                                         DC_ReadGuiEvent(DCGUI_EXIT_ABORT,GetList)} ;
            TOOLTIP 'Move Column RIGHT' ;
            WHEN {||::displayIndex < ::parent:colCount}
  DCADDBUTTON CAPTION 'Choose' ACTION {||::parent:ChooseCachedColumn(), ;
                                         DC_ReadGuiEvent(DCGUI_EXIT_ABORT,GetList)};
                    TOOLTIP 'Choose Column List'

  DCSETPARENT TO
ENDIF

DCGETOPTIONS ;
      TABSTOP ;
      ENTERTAB ;
      NOMINBUTTON ;
      NOMAXBUTTON ;
      NORESIZE ;
      SAYRIGHTBOTTOM ;
      SAYWIDTH 0 ;
     ;// ORIGIN XBPDLG_ORIGIN_MOUSE ;
     ;// HIDE ;
      SAYFONT '10.Lucida Console' ;
      GETFONT '10.Lucida Console' ;

DCREAD GUI OPTIONS GetOptions ;
     FIT ENTEREXIT NOAUTORESTORE ;
     MODAL SETAPPWINDOW TO lStatus ;
     OWNER oOwner ;
     PARENT @oDlg ;
     TITLE 'Choose Filter for ' + cSay ;
     EVAL {|o| Pos_WinAtHeader(o, ::parent, ::columnPos)}


/*
      EVAL {|o|oDlg := o, ;
               o:setPos({o:currentPos()[1]-130,o:currentPos()[2]-o:currentSize()[2]-10}), ;
               o:show()}
*/

::filterVal := ::filterVal1

IF !lSetRange
  ::filterVal2 := ::filterVal1
ENDIF

IF lStatus
  ::setFooter(lClear)
  IF lClear
    ::filterBlock := nil
    ::filterVal1 := nil
    ::filterVal2 := nil
  ELSE
    ::setFilterBlock(bFilter)
  ENDIF
  ::parent:setFilter()
ENDIF

RETURN nil

* -----------

METHOD DC_XbpColumnFiltered:SetFooter( lClear, lSetRange )

DEFAULT lClear := .f., ;
        lSetRange := .f.

IF ::filterType == 3
  ::footerBlock := {||Chr(255)+'*Empty*'}
ELSEIF (Empty(::filterVal1) .AND. Empty(::filterVal2)) .OR. lClear
  ::footerBlock := {||Chr(255)+';'+Chr(255)}
ELSEIF ::filterType == 2
  ::footerBlock := {||Chr(255)+;
                      IIF(Empty(::filterVal1),'','%'+Alltrim(DC_XtoC(::filterVal1))+'%;') + ;
                      Chr(255)+ ;
                      IIF(Empty(::filterVal2) .OR. !lSetRange,'','%'+Alltrim(DC_XtoC(::filterVal2))+'%')}
ELSE
  ::footerBlock := {||Chr(255)+Alltrim(DC_XtoC(::filterVal1))+';'+Chr(255)+Alltrim(DC_XtoC(::filterVal2))}
ENDIF

RETURN nil

* -----------

METHOD DC_XbpColumnFiltered:TagColumn(GetList)

::isTagged := !::isTagged

::parent:refreshAll()
DC_GetRefresh(GetList)
DC_ReadGuiEvent(DCGUI_EXIT_ABORT,GetList)

RETURN nil

* -----------

METHOD DC_XbpColumn:TagColor()

LOCAL aColor

IF Valtype(::parent:columnTagColor) = 'B'
  aColor := Eval(::parent:columnTagColor,nil,self)
ELSEIF ::isTagged
  aColor := ::parent:columnTagColor
ELSEIF Valtype(::parent:rowColorBlock) == 'B'
  aColor := Eval(::parent:rowColorBlock,::parent)
ELSE
  aColor := {nil,nil}
ENDIF

RETURN aColor

* -----------

METHOD DC_XbpColumnFiltered:TotalColumn( cSay )

LOCAL cAlias, nSaveRec, nTotal := 0, i, bFilter, GetList[0], oProgress, ;
      nRows := ::parent:displayedRows(), lCancelled := .f., oDlg, ;
      nCount := 0, nSavePointer, lProgress := .f.

IF nRows > 500
  @ 0,0 DCSAY 'Totalling...' SAYSIZE 0
  @ 1,0 DCPROGRESS oProgress SIZE 50,1 PERCENT COLOR GRA_CLR_CYAN, GRA_CLR_WHITE
  @ 3,0 DCPUSHBUTTON CAPTION 'Cancel' SIZE 10,1.2 ACTION {||lCancelled := .t.}
  DCREAD GUI FIT TITLE 'Totalling column' EXIT PARENT @oDlg
  lProgress := .t.
ENDIF

DEFAULT cSay := ''

DC_PointerWait()
IF Valtype(::parent:dataSource) == 'C'
  cAlias := ::parent:dataSource
  nSaveRec := (cAlias)->(RecNo())
  (cAlias)->(dbGoTop())
  DO WHILE !(cAlias)->(Eof()) .AND. !lCancelled
    nCount ++
    IF lProgress
      IF nCount%(Int(nRows/20)) == 0
        DC_CompleteEvents()
        DC_GetProgress(oProgress,nCount,nRows)
      ENDIF
    ENDIF
    IF !(cAlias)->(Deleted())
      nTotal += (cAlias)->(Eval(::dataLink))
    ENDIF
    (cAlias)->(dbSkip())
  ENDDO
  (cAlias)->(dbGoTo(nSaveRec))
ELSEIF Valtype(::parent:dataSource) == 'A' .AND. Len(::parent:dataSource) > 0
  IF Valtype(::filterBlock) == 'B'
    bFilter := ::filterBlock
  ELSE
    bFilter := {||.t.}
  ENDIF
  nSavePointer := ::parent:arrayElement
  nRows := Len(::parent:dataSource)
  FOR i := 1 TO Len(::parent:dataSource)
    nCount ++
    IF lProgress
      IF nCount%(Int(nRows/20)) == 0
        DC_CompleteEvents()
        DC_GetProgress(oProgress,nCount,nRows)
      ENDIF
    ENDIF
    IF lCancelled
      EXIT
    ENDIF
    IF Eval(bFilter,::parent:dataSource[i])
      ::parent:arrayElement := i
      nTotal += Eval(::dataLink)
    ENDIF
  NEXT
  ::parent:arrayElement := nSavePointer
ENDIF

DC_PointerArrow()
IF lProgress
  oDlg:destroy()
ENDIF

IF !lCancelled
  DCMSGBOX 'Total for column ' + cSay + ' = ' + AllTrim(Str(nTotal)) ;
    FONT '10.Lucida Console'
ENDIF

RETURN nil

* -----------

METHOD DC_XbpColumnFiltered:SetFilterBlock(bFilter)

LOCAL bVarBlock

IF ValType(::varBlock) == 'B'
  bVarBlock := ::varBlock
ELSE
  bVarBlock := ::dataLink
ENDIF

IF Empty(::filterVal1) .AND. Valtype(::filterVal1) # 'L' .AND. ::filterType # 3
  ::filterBlock := nil
ELSEIF Valtype(bFilter) == 'B'
  ::filterBlock := bFilter
ELSEIF Valtype(::filterVal1) $ 'C'
  IF ::filterType == 1 // starts with
    ::filterBlock := {|o,x|x := Upper(Eval(bVarBlock,o)), ;
                           x >= Upper(Alltrim(::filterVal1)) .AND. ;
                           x <= Upper(Alltrim(::filterVal2))}
  ELSEIF ::filterType == 2 // contained in
    ::filterBlock := {|o,x|x := Upper(Eval(bVarBlock,o)), ;
                           AT(Upper(Alltrim(::filterVal1)), x) > 0 .OR. ;
                           AT(Upper(Alltrim(::filterVal2)), x) > 0  }
  ELSEIF ::filterType == 3 // empty
    ::filterBlock := {|o,x|Empty(Eval(bVarBlock,o))}
  ENDIF
ELSEIF Valtype(::filterVal1) $ 'NDL'
  IF ::filterType == 3 // empty
    ::filterBlock := {|o,x|Empty(Eval(bVarBlock,o))}
  ELSE
    ::filterBlock := {|o,x|x := Eval(bVarBlock,o), x >= ::filterVal1 .AND. x <= ::filterVal2 }
  ENDIF
ENDIF

RETURN nil

* -----------

METHOD DC_XbpColumnFiltered:LocateRecord(bFilter,cFieldName,bClear,GetList)

LOCAL nSaveRec := RecNo(), cAlias, i, lLocate := .t., nSaveOrder, cIndexKey, ;
      cIndexTag

IF !Empty(::filterVal1)
  ::setFilterBlock(bFilter)
  DC_PointerWait()
  IF Valtype(::parent:dataSource) == 'C'
    cAlias := ::parent:dataSource
    IF Empty(::parent:sqlStatementOriginal) .AND. Valtype(::filterVal1) $ 'CD'
      cIndexTag := ::parent:FindGoodIndex(cFieldName)
      IF !Empty(cIndexTag)
        IF Valtype(::filterVal1) == 'D'
          IF (cAlias)->(dbSeek(Dtos(::filterVal1),,cIndexTag))
            lLocate := .f.
          ENDIF
        ELSEIF Valtype(::filterVal1) $ 'CM'
          IF (cAlias)->(dbSeek(Trim(::filterVal1),,cIndexTag))
            lLocate := .f.
          ELSEIF (cAlias)->(dbSeek(Trim(Upper(::filterVal1)),cIndexTag))
            lLocate := .f.
          ENDIF
        ENDIF
      ENDIF
    ENDIF
    IF lLocate
      (cAlias)->(dbLocate(::filterBlock))
    ENDIF
    IF (cAlias)->(Eof())
      DCMSGBOX Trim(DC_Xtoc(::filterVal1)) + ' not found!'
      (cAlias)->(dbGoTo(nSaveRec))
    ELSE
      ::parent:refreshAll()
      DC_ReadGuiEvent(DCGUI_EXIT_ABORT,GetList)
    ENDIF


/*
  ELSEIF Valtype(::parent:dataSource) == 'A' .AND. Len(::parent:dataSource) > 0 ;
         .AND. Valtype(::parent:dataSource[1]) == 'O'
    FOR i := 1 TO Len(::parent:dataSource)
      IF Eval(::filterBlock,::parent:dataSource[i])
        EXIT
      ENDIF
    NEXT
    IF i > Len(::parent:dataSource)
      DCMSGBOX Trim(DC_Xtoc(::filterVal1)) + ' not found!'
    ELSE
      ::parent:arrayElement := i
      ::parent:refreshAll()
    ENDIF
*/
  ELSEIF Valtype(::parent:dataSource) == 'A' .AND. Len(::parent:dataSource) > 0
    FOR i := 1 TO Len(::parent:dataSource)
      IF Eval(::filterBlock,::parent:dataSource[i])
        EXIT
      ENDIF
    NEXT
    IF i > Len(::parent:dataSource)
      DCMSGBOX Trim(DC_Xtoc(::filterVal1)) + ' not found!'
    ELSE
      ::parent:arrayElement := i
      ::parent:refreshAll()
    ENDIF
  ENDIF
  DC_PointerArrow()
ENDIF

RETURN nil

* -----------

STATIC FUNCTION Pos_WinAtHeader( oDialog, oBrowse, nColPos, nRowOffSet, nColOffSet )

LOCAL nHeight, nWidth, nCol, nRow, oCell, aCoords

IF ValType(oBrowse) == 'O'
  IF ValType(nColPos) == 'N'
    IF nColPos == 0
      nColPos := 1
    ENDIF
  ELSE
    nColPos := oBrowse:colPos
  ENDIF

  oCell := oBrowse:getColumn(nColPos):heading
  aCoords := DC_CalcAbsolutePosition(oCell:cellRect(oBrowse:rowPos),oCell)
ENDIF

IF Valtype(aCoords) = 'A'

   nRowOffSet := IIF( ValType(nRowOffSet) == 'N', nRowOffSet, 5 )
   nColOffSet := IIF( ValType(nColOffSet) == 'N', nColOffSet, 0 )

   nHeight := oDialog:currentSize()[2]
   nWidth  := oDialog:currentSize()[1]

   nCol := aCoords[1] - (nWidth/2) + (oCell:currentSize()[1]/2)  // x coordinate
   nRow := aCoords[2]-nHeight+nRowOffSet                         // y coordinate

   oDialog:setPos({nCol,nRow})
ENDIF

oDialog:show()

RETURN nil

* -----------

STATIC FUNCTION ODBC_SqlStmtExec(cSql,oSession)

LOCAL oError, lStatus := .f., GetList[0], GetOptions, ;
      bError := ErrorBlock({|oError|break(oError)}), ;
      aStack[0], aStackList[0], i := 1, cSqlError

BEGIN SEQUENCE

  // Execute a statement
  lStatus := &('SqlStmtExec')(cSql,oSession,,,.F.)

RECOVER USING oError

  DO WHILE .t.

    AAdd( aStackList, "Called from " + Trim(ProcName(i)) + ;
       "(" + Alltrim(Str(ProcLine(i))) + ")" )
    i++
    IF Empty(ProcName(i))
      EXIT
    ENDIF

  ENDDO

  cSqlError := &('OdbcDebug')(oSession)[15,2]

  @ 0,0 DCSAY 'Error in SQL statement:' SAYSIZE 0 FONT '10.Lucida Console' ;
        RESIZE DCGUI_RESIZE_REPOSONLY_Y

  @ 1,0 DCMULTILINE cSql SIZE 120,18 FONT '10.Lucida Console' NOHSCROLL ;
        RESIZE DCGUI_RESIZE_RESIZEONLY

  @ 20,0 DCMULTILINE cSqlError SIZE 120,3 FONT '8.Lucida Console' NOHSCROLL ;
        RESIZE DCGUI_RESIZE_REPOSY_RESIZEX

  FOR i := 1 TO Len(aStackList)
    @ 23+i,0 DCSAY aStackList[i] FONT '10.Lucida Console' SAYSIZE 0 ;
      RESIZE DCGUI_RESIZE_REPOSONLY_Y
  NEXT

  DCGETOPTIONS RESIZE

  DCREAD GUI FIT TITLE 'SQL Error' BUTTONS DCGUI_BUTTON_OK ;
     MODAL OPTIONS GetOptions

END SEQUENCE

ErrorBlock(bError)

RETURN lStatus

* ------------

FUNCTION DC_BrowseSqlQuery( oSqlSession, oOwner )

LOCAL nDll, oSqlQuery, oSqlQueryConnection

IF Empty(Alias())
  DC_WinAlert('No Alias in Workarea to Browse')
  RETURN .f.
ELSEIF !IsFunction('SqlQuery')
  nDll := DllLoad('SQLQUERY.DLL')
  IF nDll <= 0
    DC_WinAlert('Cannot load SQLQUERY.DLL')
    RETURN .f.
  ENDIF
ENDIF
oSqlQuery := &('SqlQuery()')
oSqlQuery := oSqlQuery:new()
oSqlQueryConnection := &('SqlQueryConnection()')
oSqlQueryConnection := oSqlQueryConnection:new()
IF Valtype(oSqlSession) == 'O'
  oSqlQueryConnection:sessionDict := oSqlSession
ELSEIF IsMemVar('SRV1')
  oSqlQueryConnection:sessionDict := M->srv1
ELSE
  oSqlQueryConnection:sessionDict := DacSession():getDefault()
ENDIF

DEFAULT oOwner := SetAppWindow()

oSqlQuery:dataBrowse( .t.,, oSqlQueryConnection,,,,,,,,,,,oOwner)

RETURN .t.


